	.module lanc111.c
	.area text(rom, con, rel)
;            msk -> R22
;              i -> R22
;             rs -> R20
;             we -> R18
;            reg -> R16
	.even
_NicPhyRegSelect:
	xcall push_xgsetF000
; /*
;  * Copyright (C) 2003-2005 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  */
; 
; /*
;  * $Log: lanc111.c,v $
;  * Revision 1.3  2005/10/24 18:02:34  haraldkipp
;  * Fixes for ATmega103.
;  *
;  * Revision 1.2  2005/08/02 17:46:45  haraldkipp
;  * Major API documentation update.
;  *
;  * Revision 1.1  2005/07/26 18:02:40  haraldkipp
;  * Moved from dev.
;  *
;  * Revision 1.13  2005/04/30 16:42:41  chaac
;  * Fixed bug in handling of NUTDEBUG. Added include for cfg/os.h. If NUTDEBUG
;  * is defined in NutConf, it will make effect where it is used.
;  *
;  * Revision 1.12  2005/02/02 19:55:34  haraldkipp
;  * If no Ethernet link was available on the LAN91C111, each outgoing packet
;  * took 15 seconds and, even worse, the ouput routine doesn't return an error.
;  * Now the first attempt to send a packet without Ethernet link will wait for
;  * 5 seconds and subsequent attempts take 0.5 seconds only, always returning
;  * an error.
;  *
;  * Revision 1.11  2005/01/24 21:11:49  freckle
;  * renamed NutEventPostFromIRQ into NutEventPostFromIrq
;  *
;  * Revision 1.10  2005/01/22 19:24:11  haraldkipp
;  * Changed AVR port configuration names from PORTx to AVRPORTx.
;  *
;  * Revision 1.9  2005/01/21 16:49:45  freckle
;  * Seperated calls to NutEventPostAsync between Threads and IRQs
;  *
;  * Revision 1.8  2004/09/22 08:14:48  haraldkipp
;  * Made configurable
;  *
;  * Revision 1.7  2004/03/08 11:14:17  haraldkipp
;  * Added quick hack for fixed mode.
;  *
;  * Revision 1.6  2004/02/25 16:22:33  haraldkipp
;  * Do not initialize MAC with all zeros
;  *
;  * Revision 1.5  2004/01/14 19:31:43  drsung
;  * Speed improvement to NicWrite applied. Thanks to Kolja Waschk
;  *
;  * Revision 1.4  2003/11/06 09:26:50  haraldkipp
;  * Removed silly line with hardcoded MAC, left over from testing
;  *
;  * Revision 1.3  2003/11/04 17:54:47  haraldkipp
;  * PHY configuration timing changed again for reliable linking
;  *
;  * Revision 1.2  2003/11/03 17:12:53  haraldkipp
;  * Allow linking with RTL8019 driver.
;  * Links more reliable to 10 MBit networks now.
;  * Reset MMU on allocation failures.
;  * Some optimizations.
;  *
;  * Revision 1.1  2003/10/13 10:13:49  haraldkipp
;  * First release
;  *
;  */
; 
; #include <cfg/os.h>
; #include <cfg/arch/avr.h>
; 
; #include <string.h>
; 
; #include <sys/atom.h>
; #include <sys/heap.h>
; #include <sys/thread.h>
; #include <sys/event.h>
; #include <sys/timer.h>
; #include <sys/confnet.h>
; 
; #include <netinet/if_ether.h>
; #include <net/ether.h>
; #include <net/if_var.h>
; 
; #include <dev/irqreg.h>
; #include <dev/lanc111.h>
; 
; #ifdef NUTDEBUG
; #include <stdio.h>
; #endif
; 
; /*
;  * Determine ports, which had not been explicitely configured.
;  */
; #ifndef LANC111_BASE_ADDR
; #define LANC111_BASE_ADDR   0xC000
; #endif
; 
; #ifndef LANC111_SIGNAL_IRQ
; #define LANC111_SIGNAL_IRQ  INT5
; #endif
; 
; #ifdef LANC111_RESET_BIT
; 
; #if (LANC111_RESET_AVRPORT == AVRPORTB)
; #define LANC111_RESET_PORT   PORTB
; #define LANC111_RESET_DDR    DDRB
; 
; #elif (LANC111_RESET_AVRPORT == AVRPORTD)
; #define LANC111_RESET_PORT   PORTD
; #define LANC111_RESET_DDR    DDRD
; 
; #elif (LANC111_RESET_AVRPORT == AVRPORTE)
; #define LANC111_RESET_PORT   PORTE
; #define LANC111_RESET_DDR    DDRE
; 
; #elif (LANC111_RESET_AVRPORT == AVRPORTF)
; #define LANC111_RESET_PORT   PORTF
; #define LANC111_RESET_DDR    DDRF
; 
; #endif /* LANC111_RESET_AVRPORT */
; 
; #endif /* LANC111_RESET_BIT */
; 
; /*
;  * Determine interrupt settings.
;  */
; #if (LANC111_SIGNAL_IRQ == INT0)
; #define LANC111_SIGNAL          sig_INTERRUPT0
; #define LANC111_SIGNAL_MODE()   sbi(EICRA, ISC00); sbi(EICRA, ISC01)
; 
; #elif (LANC111_SIGNAL_IRQ == INT1)
; #define LANC111_SIGNAL          sig_INTERRUPT1
; #define LANC111_SIGNAL_MODE()   sbi(EICRA, ISC10); sbi(EICRA, ISC11)
; 
; #elif (LANC111_SIGNAL_IRQ == INT2)
; #define LANC111_SIGNAL          sig_INTERRUPT2
; #define LANC111_SIGNAL_MODE()   sbi(EICRA, ISC20); sbi(EICRA, ISC21)
; 
; #elif (LANC111_SIGNAL_IRQ == INT3)
; #define LANC111_SIGNAL          sig_INTERRUPT3
; #define LANC111_SIGNAL_MODE()   sbi(EICRA, ISC30); sbi(EICRA, ISC31)
; 
; #elif (LANC111_SIGNAL_IRQ == INT4)
; #define LANC111_SIGNAL          sig_INTERRUPT4
; #define LANC111_SIGNAL_MODE()   sbi(EICR, ISC40); sbi(EICR, ISC41)
; 
; #elif (LANC111_SIGNAL_IRQ == INT6)
; #define LANC111_SIGNAL          sig_INTERRUPT6
; #define LANC111_SIGNAL_MODE()   sbi(EICR, ISC60); sbi(EICR, ISC61)
; 
; #elif (LANC111_SIGNAL_IRQ == INT7)
; #define LANC111_SIGNAL          sig_INTERRUPT7
; #define LANC111_SIGNAL_MODE()   sbi(EICR, ISC70); sbi(EICR, ISC71)
; 
; #else
; #define LANC111_SIGNAL          sig_INTERRUPT5
; #define LANC111_SIGNAL_MODE()   sbi(EICR, ISC50); sbi(EICR, ISC51)
; 
; #endif
; 
; /*!
;  * \addtogroup xgSmscRegs
;  */
; /*@{*/
; 
; /*! 
;  * \brief Bank select register. 
;  */
; #define NIC_BSR         (LANC111_BASE_ADDR + 0x0E)
; 
; /*! 
;  * \brief Bank 0 - Transmit control register. 
;  */
; #define NIC_TCR         (LANC111_BASE_ADDR + 0x00)
; 
; #define TCR_SWFDUP      0x8000  /*!< \ref NIC_TCR bit mask, enables full duplex.  */
; #define TCR_EPH_LOOP    0x2000  /*!< \ref NIC_TCR bit mask, enables internal loopback. */
; #define TCR_STP_SQET    0x1000  /*!< \ref NIC_TCR bit mask, enables transmission stop on SQET error. */
; #define TCR_FDUPLX      0x0800  /*!< \ref NIC_TCR bit mask, enables receiving own frames. */
; #define TCR_MON_CSN     0x0400  /*!< \ref NIC_TCR bit mask, enables carrier monitoring. */
; #define TCR_NOCRC       0x0100  /*!< \ref NIC_TCR bit mask, disables CRC transmission. */
; #define TCR_PAD_EN      0x0080  /*!< \ref NIC_TCR bit mask, enables automatic padding. */
; #define TCR_FORCOL      0x0004  /*!< \ref NIC_TCR bit mask, forces collision. */
; #define TCR_LOOP        0x0002  /*!< \ref NIC_TCR bit mask, enables PHY loopback. */
; #define TCR_TXENA       0x0001  /*!< \ref NIC_TCR bit mask, enables transmitter. */
; 
; 
; /*! 
;  * \brief Bank 0 - EPH status register. 
;  */
; #define NIC_EPHSR       (LANC111_BASE_ADDR + 0x02)
; 
; /*! 
;  * \brief Bank 0 - Receive control register. 
;  */
; #define NIC_RCR         (LANC111_BASE_ADDR + 0x04)
; 
; #define RCR_SOFT_RST    0x8000  /*!< \ref NIC_RCR bit mask, activates software reset. */
; #define RCR_FILT_CAR    0x4000  /*!< \ref NIC_RCR bit mask, enables carrier filter. */
; #define RCR_ABORT_ENB   0x2000  /*!< \ref NIC_RCR bit mask, enables receive abort on collision. */
; #define RCR_STRIP_CRC   0x0200  /*!< \ref NIC_RCR bit mask, strips CRC. */
; #define RCR_RXEN        0x0100  /*!< \ref NIC_RCR bit mask, enables receiver. */
; #define RCR_ALMUL       0x0004  /*!< \ref NIC_RCR bit mask, multicast frames accepted when set. */
; #define RCR_PRMS        0x0002  /*!< \ref NIC_RCR bit mask, enables promiscuous mode. */
; #define RCR_RX_ABORT    0x0001  /*!< \ref NIC_RCR bit mask, set when receive was aborted. */
; 
; /*! 
;  * \brief Bank 0 - Counter register.
;  */
; #define NIC_ECR         (LANC111_BASE_ADDR + 0x06)
; 
; /*! 
;  * \brief Bank 0 - Memory information register.
;  */
; #define NIC_MIR         (LANC111_BASE_ADDR + 0x08)
; 
; /*! 
;  * \brief Bank 0 - Receive / PHY control register.
;  */
; #define NIC_RPCR        (LANC111_BASE_ADDR + 0x0A)
; 
; #define RPCR_SPEED      0x2000  /*!< \ref NIC_RPCR bit mask, PHY operates at 100 Mbps. */
; #define RPCR_DPLX       0x1000  /*!< \ref NIC_RPCR bit mask, PHY operates at full duplex mode. */
; #define RPCR_ANEG       0x0800  /*!< \ref NIC_RPCR bit mask, sets PHY in auto-negotiation mode. */
; #define RPCR_LEDA_PAT   0x0000  /*!< \ref NIC_RPCR bit mask for LEDA mode. */
; #define RPCR_LEDB_PAT   0x0010  /*!< \ref NIC_RPCR bit mask for LEDB mode. */
; 
; /*! 
;  * \brief Bank 1 - Configuration register.
;  */
; #define NIC_CR          (LANC111_BASE_ADDR + 0x00)
; 
; #define CR_EPH_EN       0x8000  /*!< \ref NIC_CR bit mask, . */
; 
; /*! 
;  * \brief Bank 1 - Base address register.
;  */
; #define NIC_BAR         (LANC111_BASE_ADDR + 0x02)
; 
; /*! 
;  * \brief Bank 1 - Individual address register.
;  */
; #define NIC_IAR         (LANC111_BASE_ADDR + 0x04)
; 
; /*! 
;  * \brief Bank 1 - General purpose register.
;  */
; #define NIC_GPR         (LANC111_BASE_ADDR + 0x0A)
; 
; /*! 
;  * \brief Bank 1 - Control register.
;  */
; #define NIC_CTR         (LANC111_BASE_ADDR + 0x0C)
; 
; #define CTR_RCV_BAD     0x4000  /*!< \ref NIC_CTR bit mask. */
; #define CTR_AUTO_RELEASE 0x0800 /*!< \ref NIC_CTR bit mask, transmit packets automatically released. */
; 
; /*!
;  * \brief Bank 2 - MMU command register.
;  */
; #define NIC_MMUCR       (LANC111_BASE_ADDR + 0x00)
; 
; #define MMUCR_BUSY      0x0001
; 
; #define MMU_NOP         0
; #define MMU_ALO         (1<<5)
; #define MMU_RST         (2<<5)
; #define MMU_REM         (3<<5)
; #define MMU_TOP         (4<<5)
; #define MMU_PKT         (5<<5)
; #define MMU_ENQ         (6<<5)
; #define MMU_RTX         (7<<5)
; 
; /*!
;  * \brief Bank 2 - Packet number register.
;  *
;  * This byte register specifies the accessible transmit packet number.
;  */
; #define NIC_PNR         (LANC111_BASE_ADDR + 0x02)
; 
; /*!
;  * \brief Bank 2 - Allocation result register.
;  *
;  * This byte register is updated upon a \ref MMU_ALO command.
;  */
; #define NIC_ARR         (LANC111_BASE_ADDR + 0x03)
; 
; #define ARR_FAILED      0x80
; 
; /*!
;  * \brief Bank 2 - FIFO ports register.
;  */
; #define NIC_FIFO        (LANC111_BASE_ADDR + 0x04)
; 
; /*!
;  * \brief Bank 2 - Pointer register.
;  */
; #define NIC_PTR         (LANC111_BASE_ADDR + 0x06)
; 
; #define PTR_RCV         0x8000  /*! \ref NIC_PTR bit mask, specifies receive or transmit buffer. */
; #define PTR_AUTO_INCR   0x4000  /*! \ref NIC_PTR bit mask, enables automatic pointer increment. */
; #define PTR_READ        0x2000  /*! \ref NIC_PTR bit mask, indicates type of access. */
; #define PTR_ETEN        0x1000  /*! \ref NIC_PTR bit mask, enables early transmit underrun detection. */
; #define PTR_NOT_EMPTY   0x0800  /*! \ref NIC_PTR bit mask, set when write data fifo is not empty. */
; 
; /*!
;  * \brief Bank 2 - Data register.
;  */
; #define NIC_DATA        (LANC111_BASE_ADDR + 0x08)
; 
; /*!
;  * \brief Bank 2 - Interrupt status register.
;  */
; #define NIC_IST         (LANC111_BASE_ADDR + 0x0C)
; 
; /*!
;  * \brief Bank 2 - Interrupt acknowledge register.
;  */
; #define NIC_ACK         (LANC111_BASE_ADDR + 0x0C)
; 
; /*!
;  * \brief Bank 2 - Interrupt mask register.
;  */
; #define NIC_MSK         (LANC111_BASE_ADDR + 0x0D)
; 
; #define INT_MD          0x80    /*!< \ref PHY state change interrupt bit mask. */
; #define INT_ERCV        0x40    /*!< \ref Early receive interrupt bit mask. */
; #define INT_EPH         0x20    /*!< \ref Ethernet protocol interrupt bit mask. */
; #define INT_RX_OVRN     0x10    /*!< \ref Receive overrun interrupt bit mask. */
; #define INT_ALLOC       0x08    /*!< \ref Transmit allocation interrupt bit mask. */
; #define INT_TX_EMPTY    0x04    /*!< \ref Transmitter empty interrupt bit mask. */
; #define INT_TX          0x02    /*!< \ref Transmit complete interrupt bit mask. */
; #define INT_RCV         0x01    /*!< \ref Receive interrupt bit mask. */
; 
; /*!
;  * \brief Bank 3 - Multicast table register.
;  */
; #define NIC_MT          (LANC111_BASE_ADDR + 0x00)
; 
; /*!
;  * \brief Bank 3 - Management interface register.
;  */
; #define NIC_MGMT        (LANC111_BASE_ADDR + 0x08)
; 
; #define MGMT_MDOE       0x08    /*!< \ref NIC_MGMT bit mask, enables MDO pin. */
; #define MGMT_MCLK       0x04    /*!< \ref NIC_MGMT bit mask, drives MDCLK pin. */
; #define MGMT_MDI        0x02    /*!< \ref NIC_MGMT bit mask, reflects MDI pin status. */
; #define MGMT_MDO        0x01    /*!< \ref NIC_MGMT bit mask, drives MDO pin. */
; 
; /*!
;  * \brief Bank 3 - Revision register.
;  */
; #define NIC_REV         (LANC111_BASE_ADDR + 0x0A)
; 
; /*!
;  * \brief Bank 3 - Early RCV register.
;  */
; #define NIC_ERCV        (LANC111_BASE_ADDR + 0x0C)
; 
; /*!
;  * \brief PHY control register.
;  */
; #define NIC_PHYCR       0
; 
; #define PHYCR_RST       0x8000  /*!< \ref NIC_PHYCR bit mask, resets PHY. */
; #define PHYCR_LPBK      0x4000  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_SPEED     0x2000  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_ANEG_EN   0x1000  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_PDN       0x0800  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_MII_DIS   0x0400  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_ANEG_RST  0x0200  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_DPLX      0x0100  /*!< \ref NIC_PHYCR bit mask, . */
; #define PHYCR_COLST     0x0080  /*!< \ref NIC_PHYCR bit mask, . */
; 
; 
; /*!
;  * \brief PHY status register.
;  */
; #define NIC_PHYSR       1
; 
; #define PHYSR_CAP_T4    0x8000  /*!< \ref NIC_PHYSR bit mask, indicates 100BASE-T4 capability. */
; #define PHYSR_CAP_TXF   0x4000  /*!< \ref NIC_PHYSR bit mask, indicates 100BASE-TX full duplex capability. */
; #define PHYSR_CAP_TXH   0x2000  /*!< \ref NIC_PHYSR bit mask, indicates 100BASE-TX half duplex capability. */
; #define PHYSR_CAP_TF    0x1000  /*!< \ref NIC_PHYSR bit mask, indicates 10BASE-T full duplex capability. */
; #define PHYSR_CAP_TH    0x0800  /*!< \ref NIC_PHYSR bit mask, indicates 10BASE-T half duplex capability. */
; #define PHYSR_CAP_SUPR  0x0040  /*!< \ref NIC_PHYSR bit mask, indicates preamble suppression capability. */
; #define PHYSR_ANEG_ACK  0x0020  /*!< \ref NIC_PHYSR bit mask, auto-negotiation completed. */
; #define PHYSR_REM_FLT   0x0010  /*!< \ref NIC_PHYSR bit mask, remote fault detected. */
; #define PHYSR_CAP_ANEG  0x0008  /*!< \ref NIC_PHYSR bit mask, indicates auto-negotiation capability. */
; #define PHYSR_LINK      0x0004  /*!< \ref NIC_PHYSR bit mask, valid link status. */
; #define PHYSR_JAB       0x0002  /*!< \ref NIC_PHYSR bit mask, jabber collision detected. */
; #define PHYSR_EXREG     0x0001  /*!< \ref NIC_PHYSR bit mask, extended capabilities available. */
; 
; 
; /*!
;  * \brief PHY identifier register 1.
;  */
; #define NIC_PHYID1      2
; 
; /*!
;  * \brief PHY identifier register 1.
;  */
; #define NIC_PHYID2      3
; 
; /*!
;  * \brief PHY auto-negotiation advertisement register.
;  */
; #define NIC_PHYANAD     4
; 
; #define PHYANAD_NP      0x8000  /*!< \ref NIC_PHYANAD bit mask, exchanging next page information. */
; #define PHYANAD_ACK     0x4000  /*!< \ref NIC_PHYANAD bit mask, acknowledged. */
; #define PHYANAD_RF      0x2000  /*!< \ref NIC_PHYANAD bit mask, remote fault. */
; #define PHYANAD_T4      0x0200  /*!< \ref NIC_PHYANAD bit mask, indicates 100BASE-T4 capability. */
; #define PHYANAD_TX_FDX  0x0100  /*!< \ref NIC_PHYANAD bit mask, indicates 100BASE-TX full duplex capability. */
; #define PHYANAD_TX_HDX  0x0080  /*!< \ref NIC_PHYANAD bit mask, indicates 100BASE-TX half duplex capability. */
; #define PHYANAD_10FDX   0x0040  /*!< \ref NIC_PHYANAD bit mask, indicates 10BASE-T full duplex capability. */
; #define PHYANAD_10_HDX  0x0020  /*!< \ref NIC_PHYANAD bit mask, indicates 10BASE-T half duplex capability. */
; #define PHYANAD_CSMA    0x0001  /*!< \ref NIC_PHYANAD bit mask, indicates 802.3 CSMA capability. */
; 
; /*!
;  * \brief PHY auto-negotiation remote end capability register.
;  */
; #define NIC_PHYANRC     5
; 
; /*!
;  * \brief PHY configuration register 1.
;  */
; #define NIC_PHYCFR1     16
; 
; /*!
;  * \brief PHY configuration register 2.
;  */
; #define NIC_PHYCFR2     17
; 
; /*!
;  * \brief PHY status output register.
;  */
; #define NIC_PHYSOR      18
; 
; #define PHYSOR_INT      0x8000  /*!< \ref NIC_PHYSOR bit mask, interrupt bits changed. */
; #define PHYSOR_LNKFAIL  0x4000  /*!< \ref NIC_PHYSOR bit mask, link failure detected. */
; #define PHYSOR_LOSSSYNC 0x2000  /*!< \ref NIC_PHYSOR bit mask, descrambler sync lost detected. */
; #define PHYSOR_CWRD     0x1000  /*!< \ref NIC_PHYSOR bit mask, code word error detected. */
; #define PHYSOR_SSD      0x0800  /*!< \ref NIC_PHYSOR bit mask, start of stream error detected. */
; #define PHYSOR_ESD      0x0400  /*!< \ref NIC_PHYSOR bit mask, end of stream error detected. */
; #define PHYSOR_RPOL     0x0200  /*!< \ref NIC_PHYSOR bit mask, reverse polarity detected. */
; #define PHYSOR_JAB      0x0100  /*!< \ref NIC_PHYSOR bit mask, jabber detected. */
; #define PHYSOR_SPDDET   0x0080  /*!< \ref NIC_PHYSOR bit mask, 100/10 speed detected. */
; #define PHYSOR_DPLXDET  0x0040  /*!< \ref NIC_PHYSOR bit mask, duplex detected. */
; 
; /*!
;  * \brief PHY mask register.
;  */
; #define NIC_PHYMSK      19
; 
; #define PHYMSK_MINT     0x8000  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_INT interrupt. */
; #define PHYMSK_MLNKFAIL 0x4000  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_LNKFAIL interrupt. */
; #define PHYMSK_MLOSSSYN 0x2000  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_LOSSSYNC interrupt. */
; #define PHYMSK_MCWRD    0x1000  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_CWRD interrupt. */
; #define PHYMSK_MSSD     0x0800  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_SSD interrupt. */
; #define PHYMSK_MESD     0x0400  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_ESD interrupt. */
; #define PHYMSK_MRPOL    0x0200  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_RPOL interrupt. */
; #define PHYMSK_MJAB     0x0100  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_JAB interrupt. */
; #define PHYMSK_MSPDDT   0x0080  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_SPDDET interrupt. */
; #define PHYMSK_MDPLDT   0x0040  /*!< \ref NIC_PHYMSK bit mask, enables \ref PHYSOR_DPLXDET interrupt. */
; 
; 
; 
; #define MSBV(bit)       (1 << ((bit) - 8))
; 
; #define nic_outlb(addr, val) (*(volatile u_char *)(addr) = (val))
; #define nic_outhb(addr, val) (*(volatile u_char *)((addr) + 1) = (val))
; #define nic_outwx(addr, val) (*(volatile u_short *)(addr) = (val))
; #define nic_outw(addr, val) { \
;     *(volatile u_char *)(addr) = (u_char)(val); \
;     *((volatile u_char *)(addr) + 1) = (u_char)((val) >> 8); \
; }
; 
; #define nic_inlb(addr) (*(volatile u_char *)(addr))
; #define nic_inhb(addr) (*(volatile u_char *)((addr) + 1))
; #define nic_inw(addr) (*(volatile u_short *)(addr))
; 
; #define nic_bs(bank)    nic_outlb(NIC_BSR, bank)
; 
; /*!
;  * \struct _NICINFO lanc111.h dev/lanc111.h
;  * \brief Network interface controller information structure.
;  */
; struct _NICINFO {
;     HANDLE volatile ni_rx_rdy;  /*!< Receiver event queue. */
;     HANDLE volatile ni_tx_rdy;  /*!< Transmitter event queue. */
;     u_short ni_tx_cnt;          /*!< Number of bytes in transmission queue. */
;     u_long ni_rx_packets;       /*!< Number of packets received. */
;     u_long ni_tx_packets;       /*!< Number of packets sent. */
;     u_long ni_interrupts;       /*!< Number of interrupts. */
;     u_long ni_overruns;         /*!< Number of packet overruns. */
;     u_long ni_rx_frame_errors;  /*!< Number of frame errors. */
;     u_long ni_rx_crc_errors;    /*!< Number of CRC errors. */
;     u_long ni_rx_missed_errors; /*!< Number of missed packets. */
; };
; 
; /*!
;  * \brief Network interface controller information type.
;  */
; typedef struct _NICINFO NICINFO;
; 
; /*@}*/
; 
; /*!
;  * \addtogroup xgNicLanc111
;  */
; /*@{*/
; 
; 
; static HANDLE mutex;
; static HANDLE maq;
; 
; /*!
;  * \brief Select specified PHY register for reading or writing.
;  *
;  * \note NIC interrupts must have been disabled before calling this routine.
;  *
;  * \param reg PHY register number.
;  * \param we  Indicates type of access, 1 for write and 0 for read.
;  *
;  * \return Contents of the PHY interface rgister.
;  */
; static u_char NicPhyRegSelect(u_char reg, u_char we)
; {
;     u_char rs;
;     u_char msk;
;     u_char i;
; 
;     nic_bs(3);
	ldi R24,3
	sts 49166,R24
;     rs = (nic_inlb(NIC_MGMT) & ~(MGMT_MCLK | MGMT_MDO)) | MGMT_MDOE;
	lds R20,49160
	andi R20,250
	ori R20,8
; 
;     /* Send idle pattern. */
;     for (i = 0; i < 33; i++) {
	clr R22
	xjmp L6
L3:
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;     }
L4:
	inc R22
L6:
	cpi R22,33
	brlo L3
X0:
; 
;     /* Send start sequence. */
;     nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;     nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;     nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;     nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
; 
;     /* Write or read mode. */
;     if (we) {
	tst R18
	breq L7
X1:
;         nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;         nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;     } else {
	xjmp L8
L7:
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;         nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;         nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;         nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;     }
L8:
; 
;     /* Send PHY address. Zero is used for the internal PHY. */
;     for (i = 0; i < 5; i++) {
	clr R22
	xjmp L12
L9:
;         nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;         nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;     }
L10:
	inc R22
L12:
	cpi R22,5
	brlo L9
X2:
; 
;     /* Send PHY register number. */
;     for (msk = 0x10; msk; msk >>= 1) {
	ldi R22,16
	xjmp L16
L13:
;         if (reg & msk) {
	mov R2,R16
	and R2,R22
	breq L17
X3:
;             nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;             nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;         } else {
	xjmp L18
L17:
;             nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;             nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;         }
L18:
;     }
L14:
	lsr R22
L16:
	tst R22
	brne L13
X4:
;     nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
; 
;     return rs;
	mov R16,R20
L2:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;              i -> R22
;             rc -> R10,R11
;             rs -> R20
;            reg -> R12
	.even
_NicPhyRead:
	xcall push_xgsetF03C
	mov R12,R16
; }
; 
; /*!
;  * \brief Read contents of PHY register.
;  *
;  * \note NIC interrupts must have been disabled before calling this routine.
;  *
;  * \param reg PHY register number.
;  *
;  * \return Contents of the specified register.
;  */
; static u_short NicPhyRead(u_char reg)
; {
;     u_short rc = 0;
	clr R10
	clr R11
;     u_char rs;
;     u_char i;
; 
;     /* Select register for reading. */
;     rs = NicPhyRegSelect(reg, 0);
	clr R18
	mov R16,R12
	xcall _NicPhyRegSelect
	mov R20,R16
; 
;     /* Switch data direction. */
;     rs &= ~MGMT_MDOE;
	andi R20,247
;     nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;     nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
; 
;     /* Clock data in. */
;     for (i = 0; i < 16; i++) {
	clr R22
	xjmp L23
L20:
;         nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;         nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;         rc <<= 1;
	lsl R10
	rol R11
;         rc |= (nic_inlb(NIC_MGMT) & MGMT_MDI) != 0;
	lds R2,49160
	sbrs R2,1
	rjmp L24
X5:
	ldi R24,1
	ldi R25,0
	movw R12,R24
	xjmp L25
L24:
	clr R12
	clr R13
L25:
	or R10,R12
	or R11,R13
;     }
L21:
	inc R22
L23:
	cpi R22,16
	brlo L20
X6:
; 
;     /* This will set the clock line to low. */
;     nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
; 
;     return rc;
	movw R16,R10
L19:
	.dbline 0 ; func end
	xjmp pop_xgsetF03C
;             rs -> R20
;            msk -> R22,R23
;            val -> R10,R11
;            reg -> R12
	.even
_NicPhyWrite:
	xcall push_xgsetF03C
	movw R10,R18
	mov R12,R16
; }
; 
; /*!
;  * \brief Write value to PHY register.
;  *
;  * \note NIC interrupts must have been disabled before calling this routine.
;  *
;  * \param reg PHY register number.
;  * \param val Value to write.
;  */
; static void NicPhyWrite(u_char reg, u_short val)
; {
;     u_short msk;
;     u_char rs;
; 
;     /* Select register for writing. */
;     rs = NicPhyRegSelect(reg, 1);
	ldi R18,1
	mov R16,R12
	xcall _NicPhyRegSelect
	mov R20,R16
; 
;     /* Switch data direction dummy. */
;     nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;     nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;     nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;     nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
; 
;     /* Clock data out. */
;     for (msk = 0x8000; msk; msk >>= 1) {
	ldi R22,32768
	ldi R23,128
	xjmp L30
L27:
;         if (val & msk) {
	movw R2,R10
	and R2,R22
	and R3,R23
	tst R2
	brne X7
	tst R3
	breq L31
X7:
;             nic_outlb(NIC_MGMT, rs | MGMT_MDO);
	mov R24,R20
	ori R24,1
	sts 49160,R24
;             nic_outlb(NIC_MGMT, rs | MGMT_MDO | MGMT_MCLK);
	mov R24,R20
	ori R24,5
	sts 49160,R24
;         } else {
	xjmp L32
L31:
;             nic_outlb(NIC_MGMT, rs);
	sts 49160,R20
;             nic_outlb(NIC_MGMT, rs | MGMT_MCLK);
	mov R24,R20
	ori R24,4
	sts 49160,R24
;         }
L32:
;     }
L28:
	lsr R23
	ror R22
L30:
	cpi R22,0
	cpc R22,R23
	brne L27
X8:
; 
;     /* Set clock line low and output line int z-state. */
;     nic_outlb(NIC_MGMT, rs & ~MGMT_MDOE);
	mov R24,R20
	andi R24,247
	sts 49160,R24
L26:
	.dbline 0 ; func end
	xjmp pop_xgsetF03C
;        phy_sor -> R10,R11
;           mode -> R22,R23
;         phy_sr -> R10,R11
;         phy_to -> R20,R21
	.even
_NicPhyConfig:
	xcall push_xgsetF03C
; }
; 
; /*!
;  * \brief Configure the internal PHY.
;  *
;  * Reset the PHY and initiate auto-negotiation.
;  */
; static int NicPhyConfig(void)
; {
;     u_short phy_sor;
;     u_short phy_sr;
;     u_short phy_to;
;     u_short mode;
; 
;     /* 
;      * Reset the PHY and wait until this self clearing bit
;      * becomes zero. We sleep 63 ms before each poll and
;      * give up after 3 retries. 
;      */
;     //printf("Reset PHY..");
;     NicPhyWrite(NIC_PHYCR, PHYCR_RST);
	ldi R18,32768
	ldi R19,128
	clr R16
	xcall _NicPhyWrite
;     for (phy_to = 0;; phy_to++) {
	clr R20
	clr R21
L34:
;         NutSleep(63);
	ldi R16,63
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;         if ((NicPhyRead(NIC_PHYCR) & PHYCR_RST) == 0)
	clr R16
	xcall _NicPhyRead
	sbrc R17,7
	rjmp L38
X9:
;             break;
	xjmp L36
L38:
;         if (phy_to > 3)
	ldi R24,3
	ldi R25,0
	cp R24,R20
	cpc R25,R21
	brsh L40
X10:
;             return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L33
L40:
;     }
L35:
	subi R20,255  ; offset = 1
	sbci R21,255
	xjmp L34
L36:
;     //printf("OK\n");
; 
;     /* Store PHY status output. */
;     phy_sor = NicPhyRead(NIC_PHYSOR);
	ldi R16,18
	xcall _NicPhyRead
; 
;     /* Enable PHY interrupts. */
;     NicPhyWrite(NIC_PHYMSK, PHYMSK_MLOSSSYN | PHYMSK_MCWRD | PHYMSK_MSSD |
	ldi R18,16320
	ldi R19,63
	ldi R16,19
	xcall _NicPhyWrite
;                 PHYMSK_MESD | PHYMSK_MRPOL | PHYMSK_MJAB | PHYMSK_MSPDDT | PHYMSK_MDPLDT);
; 
;     /* Set RPC register. */
;     mode = RPCR_ANEG | RPCR_LEDA_PAT | RPCR_LEDB_PAT;
	ldi R22,2064
	ldi R23,8
;     nic_bs(0);
	clr R2
	sts 49166,R2
;     nic_outw(NIC_RPCR, mode);
	sts 49162,R22
	movw R2,R22
	mov R2,R3
	clr R3
	sts 49163,R2
; 
; #ifdef NIC_FIXED
;     /* Disable link. */
;     phy_sr = NicPhyRead(NIC_PHYCFR1);
;     NicPhyWrite(NIC_PHYCFR1, phy_sr | 0x8000);
;     NutSleep(63);
; 
;     /* Set fixed capabilities. */
;     NicPhyWrite(NIC_PHYCR, NIC_FIXED);
;     nic_bs(0);
;     nic_outw(NIC_RPCR, mode);
; 
;     /* Enable link. */
;     phy_sr = NicPhyRead(NIC_PHYCFR1);
;     NicPhyWrite(NIC_PHYCFR1, phy_sr & ~0x8000);
;     phy_sr = NicPhyRead(NIC_PHYCFR1);
; 
; #else
;     /*
;      * Advertise our capabilities, initiate auto negotiation
;      * and wait until this has been completed.
;      */
;     //printf("Negotiate..");
;     NicPhyWrite(NIC_PHYANAD, PHYANAD_TX_FDX | PHYANAD_TX_HDX | PHYANAD_10FDX | PHYANAD_10_HDX | PHYANAD_CSMA);
	ldi R18,481
	ldi R19,1
	ldi R16,4
	xcall _NicPhyWrite
;     NutSleep(63);
	ldi R16,63
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;     for (phy_to = 0, phy_sr = 0;; phy_to++) {
	clr R20
	clr R21
	clr R10
	clr R11
L42:
;         /* Give up after 10 seconds. */
;         if (phy_to >= 1024)
	cpi R20,0
	ldi R30,4
	cpc R21,R30
	brlo L46
X11:
;             return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L33
L46:
;         /* Restart auto negotiation every 4 seconds or on failures. */
;         if ((phy_to & 127) == 0 /* || (phy_sr & PHYSR_REM_FLT) != 0 */ ) {
	movw R24,R20
	andi R24,127
	andi R25,0
	cpi R24,0
	cpc R24,R25
	brne L48
X12:
;             NicPhyWrite(NIC_PHYCR, PHYCR_ANEG_EN | PHYCR_ANEG_RST);
	ldi R18,4608
	ldi R19,18
	clr R16
	xcall _NicPhyWrite
;             //printf("Restart..");
;             NutSleep(63);
	ldi R16,63
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;         }
L48:
;         /* Check if we are done. */
;         phy_sr = NicPhyRead(NIC_PHYSR);
	ldi R16,1
	xcall _NicPhyRead
	movw R12,R16
	movw R10,R12
;         //printf("[SR %04X]", phy_sr);
;         if (phy_sr & PHYSR_ANEG_ACK)
	sbrs R16,5
	rjmp L50
X13:
;             break;
	xjmp L44
L50:
;         NutSleep(63);
	ldi R16,63
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;     }
L43:
	subi R20,255  ; offset = 1
	sbci R21,255
	xjmp L42
L44:
;     //printf("OK\n");
; #endif
; 
;     return 0;
	clr R16
	clr R17
L33:
	.dbline 0 ; func end
	xjmp pop_xgsetF03C
;            tmo -> R20,R21
	.even
_NicMmuWait:
	st -y,R20
	st -y,R21
	movw R20,R16
; }
; 
; /*!
;  * \brief Wait until MMU is ready.
;  *
;  * Poll the MMU command register until \ref MMUCR_BUSY
;  * is cleared.
;  *
;  * \param tmo Timeout in milliseconds.
;  *
;  * \return 0 on success or -1 on timeout.
;  */
; static INLINE int NicMmuWait(u_short tmo)
; {
	xjmp L54
L53:
;     while (tmo--) {
;         if ((nic_inlb(NIC_MMUCR) & MMUCR_BUSY) == 0)
	lds R2,49152
	sbrc R2,0
	rjmp L56
X14:
;             break;
	xjmp L55
L56:
;         NutDelay(1);
	ldi R16,1
	xcall _NutDelay
;     }
L54:
	movw R2,R20
	subi R20,1
	sbci R21,0
	tst R2
	brne L53
	tst R3
	brne L53
X15:
L55:
;     return tmo ? 0 : -1;
	cpi R20,0
	cpc R20,R21
	breq L58
X16:
	clr R20
	clr R21
	xjmp L59
L58:
	ldi R20,-1
	ldi R21,-1
L59:
	movw R16,R20
L52:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
	.even
_NicReset:
; }
; 
; /*!
;  * \brief Reset the Ethernet controller.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; static int NicReset(void)
; {
; #ifdef LANC111_RESET_BIT
;     sbi(LANC111_RESET_DDR, LANC111_RESET_BIT);
;     sbi(LANC111_RESET_PORT, LANC111_RESET_BIT);
;     NutDelay(WAIT100);
;     cbi(LANC111_RESET_PORT, LANC111_RESET_BIT);
;     NutDelay(WAIT250);
;     NutDelay(WAIT250);
; #endif
; 
;     /* Disable all interrupts. */
;     nic_outlb(NIC_MSK, 0);
	clr R2
	sts 49165,R2
; 
;     /* MAC and PHY software reset. */
;     nic_bs(0);
	sts 49166,R2
;     nic_outw(NIC_RCR, RCR_SOFT_RST);
	ldi R24,32768
	ldi R25,128
	sts 49156,R24
	ldi R24,128
	sts 49157,R24
; 
;     /* Enable Ethernet protocol handler. */
;     nic_bs(1);
	ldi R24,1
	sts 49166,R24
;     nic_outw(NIC_CR, CR_EPH_EN);
	clr R24
	sts 49152,R24
	ldi R24,128
	sts 49153,R24
; 
;     NutDelay(10);
	ldi R16,10
	xcall _NutDelay
; 
;     /* Disable transmit and receive. */
;     nic_bs(0);
	clr R2
	sts 49166,R2
;     nic_outw(NIC_RCR, 0);
	sts 49156,R2
	sts 49157,R2
;     nic_outw(NIC_TCR, 0);
	sts 49152,R2
	sts 49153,R2
; 
;     /* Enable auto release. */
;     nic_bs(1);
	ldi R24,1
	sts 49166,R24
;     nic_outw(NIC_CTR, CTR_AUTO_RELEASE);
	ldi R24,2048
	ldi R25,8
	sts 49164,R24
	ldi R24,8
	sts 49165,R24
; 
;     /* Reset MMU. */
;     nic_bs(2);
	ldi R24,2
	sts 49166,R24
;     nic_outlb(NIC_MMUCR, MMU_RST);
	ldi R24,64
	sts 49152,R24
;     if (NicMmuWait(1000))
	ldi R16,1000
	ldi R17,3
	xcall _NicMmuWait
	cpi R16,0
	cpc R16,R17
	breq L61
X17:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L60
L61:
; 
;     return 0;
	clr R16
	clr R17
L60:
	.dbline 0 ; func end
	ret
;              i -> R20
;            mac -> R10,R11
	.even
_NicStart:
	xcall push_xgset300C
	movw R10,R16
; }
; 
; /*
;  * Fires up the network interface. NIC interrupts
;  * should have been disabled when calling this
;  * function.
;  *
;  * \param mac Six byte unique MAC address.
;  */
; static int NicStart(CONST u_char * mac)
; {
;     u_char i;
; 
;     if (NicReset())
	xcall _NicReset
	cpi R16,0
	cpc R16,R17
	breq L64
X18:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L63
L64:
; 
;     /* Enable receiver. */
;     nic_bs(3);
	ldi R24,3
	sts 49166,R24
;     nic_outlb(NIC_ERCV, 7);
	ldi R24,7
	sts 49164,R24
;     nic_bs(0);
	clr R2
	sts 49166,R2
;     nic_outw(NIC_RCR, RCR_RXEN);
	ldi R24,256
	ldi R25,1
	sts 49156,R24
	ldi R24,1
	sts 49157,R24
; 
;     /* Enable transmitter and padding. */
;     nic_outw(NIC_TCR, TCR_PAD_EN | TCR_TXENA);
	ldi R24,129
	sts 49152,R24
	sts 49153,R2
; 
;     /* Configure the PHY. */
;     if (NicPhyConfig())
	xcall _NicPhyConfig
	cpi R16,0
	cpc R16,R17
	breq L66
X19:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L63
L66:
; 
;     /* Set MAC address. */
;     //printf("Set MAC %02X%02X%02X%02X%02X%02X", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
;     nic_bs(1);
	ldi R24,1
	sts 49166,R24
;     for (i = 0; i < 6; i++)
	clr R20
	xjmp L71
L68:
;         nic_outlb(NIC_IAR + i, mac[i]);
	mov R30,R20
	clr R31
	add R30,R10
	adc R31,R11
	ldd R2,z+0
	mov R30,R20
	clr R31
	subi R30,252  ; addi 49156
	sbci R31,63
	std z+0,R2
L69:
	inc R20
L71:
	cpi R20,6
	brlo L68
X20:
;     //printf("OK\n");
; 
;     /* Enable interrupts. */
;     nic_bs(2);
	ldi R24,2
	sts 49166,R24
;     nic_outlb(NIC_MSK, INT_ERCV | INT_RCV | INT_RX_OVRN);
	ldi R24,81
	sts 49165,R24
; 
;     return 0;
	clr R16
	clr R17
L63:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;             ni -> R22,R23
;            imr -> R20
;            isr -> R10
;            arg -> R20,R21
	.even
_NicInterrupt:
	xcall push_xgsetF00C
	movw R20,R16
; }
; 
; /*
;  * NIC interrupt entry.
;  */
; static void NicInterrupt(void *arg)
; {
;     u_char isr;
;     u_char imr;
;     NICINFO *ni = (NICINFO *) ((NUTDEVICE *) arg)->dev_dcb;
	movw R30,R20
	ldd R22,z+17
	ldd R23,z+18
; 
;     ni->ni_interrupts++;
	movw R24,R22
	adiw R24,14
	movw R6,R24
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R6
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
; 
;     /* Read the interrupt mask and disable all interrupts. */
;     nic_bs(2);
	ldi R24,2
	sts 49166,R24
;     imr = nic_inlb(NIC_MSK);
	lds R20,49165
;     nic_outlb(NIC_MSK, 0);
	clr R2
	sts 49165,R2
; 
;     /* Read the interrupt status and acknowledge all interrupts. */
;     isr = nic_inlb(NIC_IST);
	lds R10,49164
;     //printf("\n!%02X-%02X ", isr, imr);
;     isr &= imr;
	and R10,R20
; 
;     /*
;      * If this is a transmit interrupt, then a packet has been sent. 
;      * So we can clear the transmitter busy flag and wake up the 
;      * transmitter thread.
;      */
;     if (isr & INT_TX_EMPTY) {
	sbrs R10,2
	rjmp L73
X21:
;         nic_outlb(NIC_ACK, INT_TX_EMPTY);
	ldi R24,4
	sts 49164,R24
;         imr &= ~INT_TX_EMPTY;
	andi R20,251
;         NutEventPostFromIrq(&ni->ni_tx_rdy);
	movw R16,R22
	subi R16,254  ; offset = 2
	sbci R17,255
	xcall _NutEventPostFromIrq
;     }
	xjmp L74
L73:
;     /* Transmit error. */
;     else if (isr & INT_TX) {
	sbrs R10,1
	rjmp L75
X22:
;         /* re-enable transmit */
;         nic_bs(0);
	clr R2
	sts 49166,R2
;         nic_outw(NIC_TCR, nic_inlb(NIC_TCR) | TCR_TXENA);
	lds R24,49152
	clr R25
	ori R24,1
	sts 49152,R24
	lds R24,49152
	clr R25
	ori R24,1
	movw R2,R24
	mov R2,R3
	clr R3
	sbrc R2,7
	com R3
	sts 49153,R2
;         nic_bs(2);
	ldi R24,2
	sts 49166,R24
;         nic_outlb(NIC_ACK, INT_TX);
	sts 49164,R24
;         /* kill the packet */
;         nic_outlb(NIC_MMUCR, MMU_PKT);
	ldi R24,160
	sts 49152,R24
; 
;         NutEventPostFromIrq(&ni->ni_tx_rdy);
	movw R16,R22
	subi R16,254  ; offset = 2
	sbci R17,255
	xcall _NutEventPostFromIrq
;     }
L75:
L74:
; 
; 
;     /*
;      * If this is a receive interrupt, then wake up the receiver 
;      * thread.
;      */
;     if (isr & INT_RX_OVRN) {
	sbrs R10,4
	rjmp L77
X23:
;         nic_outlb(NIC_ACK, INT_RX_OVRN);
	ldi R24,16
	sts 49164,R24
;         //nic_outlb(NIC_MMUCR, MMU_TOP);
;     }
L77:
;     if (isr & INT_ERCV) {
	sbrs R10,6
	rjmp L79
X24:
;         nic_outlb(NIC_ACK, INT_ERCV);
	ldi R24,64
	sts 49164,R24
;         NutEventPostFromIrq(&ni->ni_rx_rdy);
	movw R16,R22
	xcall _NutEventPostFromIrq
;     }
L79:
;     if (isr & INT_RCV) {
	sbrs R10,0
	rjmp L81
X25:
;         nic_outlb(NIC_ACK, INT_RCV);
	ldi R24,1
	sts 49164,R24
;         imr &= ~INT_RCV;
	andi R20,254
;         NutEventPostFromIrq(&ni->ni_rx_rdy);
	movw R16,R22
	xcall _NutEventPostFromIrq
;     }
L81:
; 
;     if (isr & INT_ALLOC) {
	sbrs R10,3
	rjmp L83
X26:
;         imr &= ~INT_ALLOC;
	andi R20,247
;         NutEventPostFromIrq(&maq);
	ldi R16,<_maq
	ldi R17,>_maq
	xcall _NutEventPostFromIrq
;     }
L83:
;     //printf(" -%02X-%02X- ", nic_inlb(NIC_IST), inb(PINE) & 0x20);
;     nic_outlb(NIC_MSK, imr);
	sts 49165,R20
L72:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;             il -> R22
;             ih -> R20
;              l -> R22,R23
;            len -> R18,R19
;            buf -> R16,R17
	.even
_NicWrite:
	xcall push_xgsetF000
; }
; 
; /*
;  * Write data block to the NIC.
;  */
; static void NicWrite(u_char * buf, u_short len)
; {
;     register u_short l = len - 1;
	movw R22,R18
	subi R22,1
	sbci R23,0
;     register u_char ih = (u_short) l >> 8;
	movw R20,R22
	mov R20,R21
	clr R21
;     register u_char il = (u_char) l;
; 
;     if (!len)
	cpi R18,0
	cpc R18,R19
	brne L86
X27:
;         return;
	xjmp L85
L86:
L88:
; 
;     do {
L91:
;         do {
;             nic_outlb(NIC_DATA, *buf++);
	movw R30,R16
	ld R2,Z+
	movw R16,R30
	sts 49160,R2
;         } while (il-- != 0);
L92:
	mov R2,R22
	clr R3
	subi R22,1
	tst R2
	brne L91
X28:
;     } while (ih-- != 0);
L89:
	mov R2,R20
	clr R3
	subi R20,1
	tst R2
	brne L88
X29:
L85:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;             il -> R22
;             ih -> R20
;              l -> R22,R23
;            len -> R18,R19
;            buf -> R16,R17
	.even
_NicRead:
	xcall push_xgsetF000
; }
; 
; /*
;  * Read data block from the NIC.
;  */
; static void NicRead(u_char * buf, u_short len)
; {
;     register u_short l = len - 1;
	movw R22,R18
	subi R22,1
	sbci R23,0
;     register u_char ih = (u_short) l >> 8;
	movw R20,R22
	mov R20,R21
	clr R21
;     register u_char il = (u_char) l;
; 
;     if (!len)
	cpi R18,0
	cpc R18,R19
	brne L95
X30:
;         return;
	xjmp L94
L95:
L97:
; 
;     do {
L100:
;         do {
;             *buf++ = nic_inlb(NIC_DATA);
	lds R2,49160
	movw R30,R16
	st Z+,R2
	movw R16,R30
;         } while (il-- != 0);
L101:
	mov R2,R22
	clr R3
	subi R22,1
	tst R2
	brne L100
X31:
;     } while (ih-- != 0);
L98:
	mov R2,R20
	clr R3
	subi R20,1
	tst R2
	brne L97
X32:
L94:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;            fsw -> R10,R11
;            fbc -> R22,R23
;             nb -> R20,R21
	.even
_NicGetPacket:
	xcall push_xgsetF00C
	sbiw R28,2
; }
; 
; /*!
;  * \brief Fetch the next packet out of the receive ring buffer.
;  *
;  * Nic interrupts must be disabled when calling this funtion.
;  *
;  * \return Pointer to an allocated ::NETBUF. If there is no
;  *         no data available, then the function returns a
;  *         null pointer. If the NIC's buffer seems to be
;  *         corrupted, a pointer to 0xFFFF is returned.
;  */
; static NETBUF *NicGetPacket(void)
; {
;     NETBUF *nb = 0;
	clr R20
	clr R21
;     //u_char *buf;
;     u_short fsw;
;     u_short fbc;
; 
;     /* Check the fifo empty bit. If it is set, then there is 
;        nothing in the receiver fifo. */
;     nic_bs(2);
	ldi R24,2
	sts 49166,R24
;     if (nic_inw(NIC_FIFO) & 0x8000) {
	lds R2,49156
	lds R3,49156+1
	sbrs R3,7
	rjmp L104
X33:
;         return 0;
	clr R16
	clr R17
	xjmp L103
L104:
;     }
; 
;     /* Inialize pointer register. */
;     nic_outw(NIC_PTR, PTR_READ | PTR_RCV | PTR_AUTO_INCR);
	ldi R24,57344
	ldi R25,224
	sts 49158,R24
	ldi R24,224
	sts 49159,R24
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
; 
;     /* Read status word and byte count. */
;     fsw = nic_inw(NIC_DATA);
	lds R10,49160
	lds R11,49160+1
;     fbc = nic_inw(NIC_DATA);
	lds R22,49160
	lds R23,49160+1
;     //printf("[SW=%04X,BC=%04X]", fsw, fbc);
; 
;     /* Check for frame errors. */
;     if (fsw & 0xAC00) {
	movw R24,R10
	andi R24,0
	andi R25,172
	cpi R24,0
	cpc R24,R25
	breq L106
X34:
;         nb = (NETBUF *) 0xFFFF;
	ldi R20,65535
	ldi R21,255
;     }
	xjmp L107
L106:
;     /* Check the byte count. */
;     else if (fbc < 66 || fbc > 1524) {
	cpi R22,66
	ldi R30,0
	cpc R23,R30
	brlo L110
X35:
	ldi R24,1524
	ldi R25,5
	cp R24,R22
	cpc R25,R23
	brsh L108
X36:
L110:
;         nb = (NETBUF *) 0xFFFF;
	ldi R20,65535
	ldi R21,255
;     }
	xjmp L109
L108:
; 
;     else {
;         /* 
;          * Allocate a NETBUF. 
;          * Hack alert: Rev A chips never set the odd frame indicator.
;          */
;         fbc -= 3;
	subi R22,3
	sbci R23,0
;         nb = NutNetBufAlloc(0, NBAF_DATALINK, fbc);
	std y+1,R23
	std y+0,R22
	ldi R18,16
	clr R16
	clr R17
	xcall _NutNetBufAlloc
	movw R20,R16
; 
;         /* Perform the read. */
;         if (nb)
	cpi R16,0
	cpc R16,R17
	breq L111
X37:
;             NicRead(nb->nb_dl.vp, fbc);
	movw R18,R22
	movw R30,R20
	ldd R16,z+3
	ldd R17,z+4
	xcall _NicRead
L111:
;     }
L109:
L107:
; 
;     /* Release the packet. */
;     nic_outlb(NIC_MMUCR, MMU_TOP);
	ldi R24,128
	sts 49152,R24
; 
;     return nb;
	movw R16,R20
L103:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF00C
;            odd -> R10
;           imsk -> R20
;             sz -> R12,R13
;             nb -> R22,R23
	.even
_NicPutPacket:
	xcall push_xgsetF0FC
	movw R22,R16
	sbiw R28,2
; }
; 
; /*!
;  * \brief Load a packet into the nic's transmit ring buffer.
;  *
;  * Interupts must have been disabled when calling this function.
;  *
;  * \param nb Network buffer structure containing the packet to be sent.
;  *           The structure must have been allocated by a previous
;  *           call NutNetBufAlloc(). This routine will automatically
;  *           release the buffer in case of an error.
;  *
;  * \return 0 on success, -1 in case of any errors. Errors
;  *         will automatically release the network buffer 
;  *         structure.
;  */
; static int NicPutPacket(NETBUF * nb)
; {
;     u_short sz;
;     u_char odd = 0;
	clr R10
;     u_char imsk;
; 
;     //printf("[P]");
;     /*
;      * Calculate the number of bytes to be send. Do not send packets 
;      * larger than the Ethernet maximum transfer unit. The MTU
;      * consist of 1500 data bytes plus the 14 byte Ethernet header
;      * plus 4 bytes CRC. We check the data bytes only.
;      */
;     if ((sz = nb->nb_nw.sz + nb->nb_tp.sz + nb->nb_ap.sz) > ETHERMTU)
	movw R30,R22
	ldd R2,z+13
	ldd R3,z+14
	ldd R4,z+9
	ldd R5,z+10
	add R4,R2
	adc R5,R3
	ldd R2,z+17
	ldd R3,z+18
	add R4,R2
	adc R5,R3
	movw R12,R4
	ldi R24,1500
	ldi R25,5
	cp R24,R12
	cpc R25,R13
	brsh L114
X38:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L113
L114:
; 
;     /* Disable all interrupts. */
;     imsk = nic_inlb(NIC_MSK);
	lds R20,49165
;     nic_outlb(NIC_MSK, 0);
	clr R2
	sts 49165,R2
; 
;     /* Allocate packet buffer space. */
;     nic_bs(2);
	ldi R24,2
	sts 49166,R24
;     nic_outlb(NIC_MMUCR, MMU_ALO);
	ldi R24,32
	sts 49152,R24
;     if (NicMmuWait(100))
	ldi R16,100
	ldi R17,0
	xcall _NicMmuWait
	cpi R16,0
	cpc R16,R17
	breq L116
X39:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L113
L116:
; 
;     /* Enable interrupts including allocation success. */
;     nic_outlb(NIC_MSK, imsk | INT_ALLOC);
	mov R24,R20
	ori R24,8
	sts 49165,R24
; 
;     /* The MMU needs some time. Use it to calculate the byte count. */
;     sz += nb->nb_dl.sz;
	movw R30,R22
	ldd R2,z+5
	ldd R3,z+6
	add R12,R2
	adc R13,R3
;     sz += 6;
	movw R24,R12
	adiw R24,6
	movw R12,R24
;     if (sz & 1) {
	sbrs R24,0
	rjmp L121
X40:
;         sz++;
	adiw R24,1
	movw R12,R24
;         odd++;
	inc R10
;     }
	xjmp L121
L120:
; 
;     /* Wait for allocation success. */
;     while ((nic_inlb(NIC_IST) & INT_ALLOC) == 0) {
;         if (NutEventWait(&maq, 125)) {
	ldi R24,125
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	ldi R16,<_maq
	ldi R17,>_maq
	xcall _NutEventWait
	movw R14,R16
	cpi R16,0
	cpc R16,R17
	breq L123
X41:
;             nic_outlb(NIC_MMUCR, MMU_RST);
	ldi R24,64
	sts 49152,R24
;             NicMmuWait(1000);
	ldi R16,1000
	ldi R17,3
	xcall _NicMmuWait
;             nic_outlb(NIC_MMUCR, MMU_ALO);
	ldi R24,32
	sts 49152,R24
;             if (NicMmuWait(100) || (nic_inlb(NIC_IST) & INT_ALLOC) == 0) {
	ldi R16,100
	ldi R17,0
	xcall _NicMmuWait
	movw R14,R16
	cpi R16,0
	cpc R16,R17
	brne L127
X42:
	lds R2,49164
	sbrc R2,3
	rjmp L125
X43:
L127:
;                 if (NutEventWait(&maq, 125)) {
	ldi R24,125
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	ldi R16,<_maq
	ldi R17,>_maq
	xcall _NutEventWait
	movw R14,R16
	cpi R16,0
	cpc R16,R17
	breq L128
X44:
;                     return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L113
L128:
;                 }
;             }
L125:
;         }
L123:
;     }
L121:
	lds R2,49164
	sbrs R2,3
	rjmp L120
X45:
; 
;     /* Disable interrupts. */
;     imsk = nic_inlb(NIC_MSK);
	lds R20,49165
;     nic_outlb(NIC_MSK, 0);
	clr R2
	sts 49165,R2
; 
; 
;     nic_outlb(NIC_PNR, nic_inhb(NIC_PNR));
	lds R2,49155
	sts 49154,R2
; 
;     nic_outw(NIC_PTR, 0x4000);
	ldi R24,16384
	ldi R25,64
	sts 49158,R24
	ldi R24,64
	sts 49159,R24
; 
;     /* Transfer control word. */
;     nic_outlb(NIC_DATA, 0);
	clr R2
	sts 49160,R2
;     nic_outlb(NIC_DATA, 0);
	sts 49160,R2
; 
;     /* Transfer the byte count. */
;     nic_outw(NIC_DATA, sz);
	sts 49160,R12
	movw R2,R12
	mov R2,R3
	clr R3
	sts 49161,R2
; 
;     /* Transfer the Ethernet frame. */
;     NicWrite(nb->nb_dl.vp, nb->nb_dl.sz);
	movw R30,R22
	ldd R18,z+5
	ldd R19,z+6
	ldd R16,z+3
	ldd R17,z+4
	xcall _NicWrite
;     NicWrite(nb->nb_nw.vp, nb->nb_nw.sz);
	movw R30,R22
	ldd R18,z+9
	ldd R19,z+10
	ldd R16,z+7
	ldd R17,z+8
	xcall _NicWrite
;     NicWrite(nb->nb_tp.vp, nb->nb_tp.sz);
	movw R30,R22
	ldd R18,z+13
	ldd R19,z+14
	ldd R16,z+11
	ldd R17,z+12
	xcall _NicWrite
;     NicWrite(nb->nb_ap.vp, nb->nb_ap.sz);
	movw R30,R22
	ldd R18,z+17
	ldd R19,z+18
	ldd R16,z+15
	ldd R17,z+16
	xcall _NicWrite
; 
;     if (odd)
	tst R10
	breq L130
X46:
;         nic_outlb(NIC_DATA, 0);
	clr R2
	sts 49160,R2
L130:
; 
;     /* Transfer the control word. */
;     nic_outw(NIC_DATA, 0);
	clr R2
	sts 49160,R2
	sts 49161,R2
; 
;     /* Enqueue packet. */
;     if (NicMmuWait(100))
	ldi R16,100
	ldi R17,0
	xcall _NicMmuWait
	movw R10,R16
	cpi R16,0
	cpc R16,R17
	breq L132
X47:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L113
L132:
;     nic_outlb(NIC_MMUCR, MMU_ENQ);
	ldi R24,192
	sts 49152,R24
; 
;     /* Enable interrupts. */
;     imsk |= INT_TX | INT_TX_EMPTY;
	ori R20,6
;     nic_outlb(NIC_MSK, imsk);
	sts 49165,R20
; 
;     return 0;
	clr R16
	clr R17
L113:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF0FC
;           imsk -> R12
;            dev -> y+6
;             ni -> R10,R11
;            ifn -> R22,R23
;             nb -> R14,R15
;            arg -> R10,R11
	.even
_NicRxLanc::
	xcall push_xgsetF0FC
	movw R10,R16
	sbiw R28,8
; }
; 
; 
; /*! \fn NicRxLanc(void *arg)
;  * \brief NIC receiver thread.
;  *
;  */
; THREAD(NicRxLanc, arg)
; {
;     NUTDEVICE *dev;
;     IFNET *ifn;
;     NICINFO *ni;
;     NETBUF *nb;
;     u_char imsk;
; 
;     dev = arg;
	std y+7,R11
	std y+6,R10
;     ifn = (IFNET *) dev->dev_icb;
	movw R30,R10
	ldd R22,z+15
	ldd R23,z+16
;     ni = (NICINFO *) dev->dev_dcb;
	ldd R10,z+17
	ldd R11,z+18
; 
;     /*
;      * This is a temporary hack. Due to a change in initialization,
;      * we may not have got a MAC address yet. Wait until one has been
;      * set.
;      */
;     for (;;) {
L135:
;         if (*((u_long *) (ifn->if_mac)) && *((u_long *) (ifn->if_mac)) != 0xFFFFFFFFUL) {
	movw R30,R22
	ldd R2,z+1
	ldd R3,z+2
	ldd R4,z+3
	ldd R5,z+4
	std y+2,R2
	std y+3,R3
	std y+4,R4
	std y+5,R5
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	ldd R2,y+2
	ldd R3,y+3
	ldd R4,y+4
	ldd R5,y+5
	cp R2,R24
	cpc R3,R25
	cpc R4,R26
	cpc R5,R27
	breq L139
X48:
	ldi R24,255
	ldi R25,255
	ldi R26,255
	ldi R27,255
	cp R2,R24
	cpc R3,R25
	cpc R4,R26
	cpc R5,R27
	breq L139
X49:
;             break;
	xjmp L142
L139:
;         }
;         NutSleep(63);
	ldi R16,63
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;     }
	xjmp L135
L141:
; 
;     /*
;      * Do not continue unless we managed to start the NIC. We are
;      * trapped here if the Ethernet link cannot be established.
;      * This happens, for example, if no Ethernet cable is plugged
;      * in.
;      */
;     while(NicStart(ifn->if_mac)) {
;         NutSleep(1000);
	ldi R16,232
	ldi R17,3
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
;     }
L142:
	movw R16,R22
	subi R16,255  ; offset = 1
	sbci R17,255
	xcall _NicStart
	cpi R16,0
	cpc R16,R17
	brne L141
X50:
; 
;     LANC111_SIGNAL_MODE();
	in R24,0x3a
	ori R24,4
	out 0x3a,R24
	in R24,0x3a
	ori R24,8
	out 0x3a,R24
;     sbi(EIMSK, LANC111_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
; 
;     NutEventPost(&mutex);
	ldi R16,<_mutex
	ldi R17,>_mutex
	xcall _NutEventPost
; 
;     /* Run at high priority. */
;     NutThreadSetPriority(9);
	ldi R16,9
	xcall _NutThreadSetPriority
; 
;     for (;;) {
L144:
; 
;         /*
;          * Wait for the arrival of new packets or
;          * check the receiver every two second.
;          */
;         NutEventWait(&ni->ni_rx_rdy, 2000);
	ldi R24,208
	ldi R25,7
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	movw R16,R10
	xcall _NutEventWait
; 
;         /*
;          * Fetch all packets from the NIC's internal
;          * buffer and pass them to the registered handler.
;          */
;         imsk = nic_inlb(NIC_MSK);
	lds R12,49165
;         nic_outlb(NIC_MSK, 0);
	clr R2
	sts 49165,R2
	xjmp L149
L148:
;         while ((nb = NicGetPacket()) != 0) {
;             if (nb != (NETBUF *) 0xFFFF) {
	movw R24,R14
	cpi R24,255
	ldi R30,255
	cpc R25,R30
	breq L151
X51:
;                 ni->ni_rx_packets++;
	movw R24,R10
	adiw R24,6
	movw R6,R24
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R6
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;                 (*ifn->if_recv) (dev, nb);
	movw R18,R14
	ldd R16,y+6
	ldd R17,y+7
	movw R30,R22
	ldd R26,z+25
	ldd R27,z+26
	movw R30,R26
	xcall  xicall
;             }
L151:
;         }
L149:
	xcall _NicGetPacket
	movw R20,R16
	movw R14,R20
	cpi R16,0
	cpc R16,R17
	breq X54
	xjmp L148
X54:
X52:
;         nic_outlb(NIC_MSK, imsk | INT_RCV | INT_ERCV);
	mov R24,R12
	ori R24,65
	sts 49165,R24
;     }
	xjmp L144
X53:
L134:
	.dbline 0 ; func end
	adiw R28,8
	xjmp pop_xgsetF0FC
	.area data(ram, con, rel)
L154:
	.blkb 4
	.area idata
	.word 5000,0
	.area data(ram, con, rel)
	.area text(rom, con, rel)
;             ni -> R22,R23
;             rc -> R20,R21
;             nb -> R10,R11
;            dev -> R22,R23
	.even
_LancOutput::
	xcall push_xgsetF00C
	movw R10,R18
	movw R22,R16
	sbiw R28,2
; }
; 
; /*!
;  * \brief Send Ethernet packet.
;  *
;  * \param dev   Identifies the device to use.
;  * \param nb    Network buffer structure containing the packet to be sent.
;  *              The structure must have been allocated by a previous
;  *              call NutNetBufAlloc().
;  *
;  * \return 0 on success, -1 in case of any errors.
;  */
; int LancOutput(NUTDEVICE * dev, NETBUF * nb)
; {
;     static u_long mx_wait = 5000;
;     int rc = -1;
	ldi R20,-1
	ldi R21,-1
;     NICINFO *ni;
; 
;     /*
;      * After initialization we are waiting for a long time to give
;      * the PHY a chance to establish an Ethernet link.
;      */
;     if (NutEventWait(&mutex, mx_wait) == 0) {
	lds R4,L154+2
	lds R5,L154+2+1
	lds R2,L154
	lds R3,L154+1
	std y+0,R4
	std y+1,R5
	movw R18,R2
	ldi R16,<_mutex
	ldi R17,>_mutex
	xcall _NutEventWait
	cpi R16,0
	cpc R16,R17
	breq X57
	xjmp L155
X57:
X55:
;         ni = (NICINFO *) dev->dev_dcb;
	movw R30,R22
	ldd R22,z+17
	ldd R23,z+18
; 
;         if (NicPutPacket(nb) == 0) {
	movw R16,R10
	xcall _NicPutPacket
	cpi R16,0
	cpc R16,R17
	brne L157
X56:
;             ni->ni_tx_packets++;
	subi R22,246  ; offset = 10
	sbci R23,255
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R22
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;             rc = 0;
	clr R20
	clr R21
;             /* Ethernet works. Set a long waiting time in case we
;                temporarly lose the link next time. */
;             mx_wait = 5000;
	ldi R24,136
	ldi R25,19
	ldi R26,0
	ldi R27,0
	sts L154+1,R25
	sts L154,R24
	sts L154+2+1,R27
	sts L154+2,R26
;         }
L157:
;         NutEventPost(&mutex);
	ldi R16,<_mutex
	ldi R17,>_mutex
	xcall _NutEventPost
;     }
	xjmp L156
L155:
;     /*
;      * Probably no Ethernet link. Significantly reduce the waiting
;      * time, so following transmission will soon return an error.
;      */
;     else {
;         mx_wait = 500;
	ldi R24,244
	ldi R25,1
	ldi R26,0
	ldi R27,0
	sts L154+1,R25
	sts L154,R24
	sts L154+2+1,R27
	sts L154+2,R26
;     }
L156:
;     return rc;
	movw R16,R20
L153:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF00C
;            dev -> R20,R21
	.even
_LancInit::
	st -y,R20
	st -y,R21
	movw R20,R16
	sbiw R28,4
; }
; 
; /*!
;  * \brief Initialize Ethernet hardware.
;  *
;  * Resets the LAN91C111 Ethernet controller, initializes all required 
;  * hardware registers and starts a background thread for incoming 
;  * Ethernet traffic.
;  *
;  * Applications should do not directly call this function. It is 
;  * automatically executed during during device registration by 
;  * NutRegisterDevice().
;  *
;  * If the network configuration hasn't been set by the application
;  * before registering the specified device, this function will
;  * call NutNetLoadConfig() to get the MAC address.
;  *
;  * \param dev Identifies the device to initialize.
;  */
; int LancInit(NUTDEVICE * dev)
; {
;     /* Disable NIC interrupt and clear NICINFO structure. */
;     cbi(EIMSK, LANC111_SIGNAL_IRQ);
	in R24,0x39
	andi R24,223
	out 0x39,R24
;     memset(dev->dev_dcb, 0, sizeof(NICINFO));
	ldi R24,34
	ldi R25,0
	std y+1,R25
	std y+0,R24
	clr R18
	clr R19
	movw R30,R20
	ldd R16,z+17
	ldd R17,z+18
	xcall _memset
; 
;     /* Register interrupt handler and enable interrupts. */
;     if (NutRegisterIrqHandler(&LANC111_SIGNAL, NicInterrupt, dev))
	std y+1,R21
	std y+0,R20
	ldi R18,<PL_NicInterrupt
	ldi R19,>PL_NicInterrupt
	ldi R16,<_sig_INTERRUPT5
	ldi R17,>_sig_INTERRUPT5
	xcall _NutRegisterIrqHandler
	cpi R16,0
	cpc R16,R17
	breq L160
X58:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L159
L160:
; 
;     /*
;      * Start the receiver thread.
;      */
;     NutThreadCreate("rxi5", NicRxLanc, dev, 640);
	ldi R24,640
	ldi R25,2
	std y+3,R25
	std y+2,R24
	std y+1,R21
	std y+0,R20
	ldi R18,<PL_NicRxLanc
	ldi R19,>PL_NicRxLanc
	ldi R16,<L162
	ldi R17,>L162
	xcall _NutThreadCreate
; 
;     //NutSleep(500);
; 
;     return 0;
	clr R16
	clr R17
L159:
	.dbline 0 ; func end
	adiw R28,4
	ld R21,y+
	ld R20,y+
	ret
	.area data(ram, con, rel)
_ifn_eth0:
	.blkb 1
	.area idata
	.byte 6
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 4
	.area idata
	.word 0,0
	.area data(ram, con, rel)
	.blkb 4
	.area idata
	.word 0,0
	.area data(ram, con, rel)
	.blkb 4
	.area idata
	.word 0,0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 1500
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_NutEtherInput
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_LancOutput
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_NutEtherOutput
	.area data(ram, con, rel)
_devSmsc111::
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 'e,'t
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 'h,48
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 3
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word _ifn_eth0
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word _dcb_eth0
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_LancInit
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.area bss(ram, con, rel)
_dcb_eth0:
	.blkb 34
_maq:
	.blkb 2
_mutex:
	.blkb 2
	.area data(ram, con, rel)
L162:
	.blkb 5
	.area idata
	.byte 'r,'x,'i,53,0
	.area data(ram, con, rel)
; }
; 
; static NICINFO dcb_eth0;
; 
; /*!
;  * \brief Network interface information structure.
;  *
;  * Used to call.
;  */
; static IFNET ifn_eth0 = {
;     IFT_ETHER,                  /*!< \brief Interface type. */
;     {0, 0, 0, 0, 0, 0},         /*!< \brief Hardware net address. */
;     0,                          /*!< \brief IP address. */
;     0,                          /*!< \brief Remote IP address for point to point. */
;     0,                          /*!< \brief IP network mask. */
;     ETHERMTU,                   /*!< \brief Maximum size of a transmission unit. */
;     0,                          /*!< \brief Packet identifier. */
;     0,                          /*!< \brief Linked list of arp entries. */
;     NutEtherInput,              /*!< \brief Routine to pass received data to, if_recv(). */
;     LancOutput,                 /*!< \brief Driver output routine, if_send(). */
;     NutEtherOutput              /*!< \brief Media output routine, if_output(). */
; };
; 
; /*!
;  * \brief Device information structure.
;  *
;  * A pointer to this structure must be passed to NutRegisterDevice() 
;  * to bind this Ethernet device driver to the Nut/OS kernel.
;  * An application may then call NutNetIfConfig() with the name \em eth0 
;  * of this driver to initialize the network interface.
;  * 
;  */
; NUTDEVICE devSmsc111 = {
;     0,                          /* Pointer to next device. */
;     {'e', 't', 'h', '0', 0, 0, 0, 0, 0},        /* Unique device name. */
;     IFTYP_NET,                  /* Type of device. */
;     0,                          /* Base address. */
;     0,                          /* First interrupt number. */
;     &ifn_eth0,                  /* Interface control block. */
;     &dcb_eth0,                  /* Driver control block. */
;     LancInit,                   /* Driver initialization routine. */
;     0,                          /* Driver specific control function. */
;     0,                          /* Read from device. */
;     0,                          /* Write to device. */
;     0,                          /* Write from program space data to device. */
;     0,                          /* Open a device or file. */
;     0,                          /* Close a device or file. */
;     0                           /* Request file size. */
; };
; 
; /*@}*/
	.area func_lit
PL_LancInit:	.word `_LancInit
PL_NutEtherOutput:	.word `_NutEtherOutput
PL_LancOutput:	.word `_LancOutput
PL_NutEtherInput:	.word `_NutEtherInput
PL_NicRxLanc:	.word `_NicRxLanc
PL_NicInterrupt:	.word `_NicInterrupt
