	.module nicrtl.c
	.area text(rom, con, rel)
	.even
_Delay16Cycles:
; /*
;  * Copyright (C) 2001-2005 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  */
; 
; /*
;  * $Log: nicrtl.c,v $
;  * Revision 1.1  2005/07/26 18:02:40  haraldkipp
;  * Moved from dev.
;  *
;  * Revision 1.15  2005/04/30 16:42:41  chaac
;  * Fixed bug in handling of NUTDEBUG. Added include for cfg/os.h. If NUTDEBUG
;  * is defined in NutConf, it will make effect where it is used.
;  *
;  * Revision 1.14  2005/02/10 07:06:18  hwmaier
;  * Changes to incorporate support for AT90CAN128 CPU
;  *
;  * Revision 1.13  2005/02/05 20:42:38  haraldkipp
;  * Force compiler error for leftover debug prints.
;  *
;  * Revision 1.12  2005/01/24 21:11:50  freckle
;  * renamed NutEventPostFromIRQ into NutEventPostFromIrq
;  *
;  * Revision 1.11  2005/01/22 19:24:46  haraldkipp
;  * Changed AVR port configuration names from PORTx to AVRPORTx.
;  *
;  * Revision 1.10  2005/01/21 16:49:46  freckle
;  * Seperated calls to NutEventPostAsync between Threads and IRQs
;  *
;  * Revision 1.9  2004/12/17 15:31:28  haraldkipp
;  * Support of rising edge interrupts for hardware w/o inverter gate.
;  * Fixed compilation issue for hardware with RTL reset port.
;  * Thanks to FOCUS Software Engineering Pty Ltd.
;  *
;  * Revision 1.8  2004/09/10 10:36:01  haraldkipp
;  * ICCAVR compile problems fixed
;  *
;  * Revision 1.7  2004/08/25 10:41:00  haraldkipp
;  * Hardware dependent definitions are configurable. For performance reasons the
;  * base address is not kept in a variable any longer. It is now a preprocessor
;  * macro and the parameters during device registration are ignored. The speed
;  * improvements provided by Kolja Waschk for the LAN91C111 had been implemented
;  * here too. The driver will not touch a port anymore unless a reset port bit
;  * had been configured. For Ethernut 1.1 bit 4 PORTE must be specified in the
;  * configuration. Finally, an EEPROM emulation had been added, which can use
;  * address bus bits instead of wasting additional port pins. The required
;  * hardware has been implemented on Rev.-G Ethernut 1.3 boards. This fixes the
;  * Realtek full duplex problem.
;  *
;  * Revision 1.6  2004/05/17 19:14:53  haraldkipp
;  * Added Bengt Florin's RTL8019 driver mods
;  *
;  * Revision 1.5  2004/03/16 16:48:27  haraldkipp
;  * Added Jan Dubiec's H8/300 port.
;  *
;  * Revision 1.4  2003/08/05 20:05:59  haraldkipp
;  * Bugfix. Empty MAC address is FF, not 00
;  *
;  * Revision 1.3  2003/07/17 09:39:56  haraldkipp
;  * Optimized controller feeding.
;  * Ignore LSB of packet status.
;  *
;  * Revision 1.2  2003/05/15 14:09:16  haraldkipp
;  * Much better performance under heavy traffic.
;  *
;  * Revision 1.1.1.1  2003/05/09 14:40:48  haraldkipp
;  * Initial using 3.2.1
;  *
;  */
; 
; #include <cfg/os.h>
; #include <cfg/arch/avr.h>
; 
; #include <string.h>
; 
; #include <sys/atom.h>
; #include <sys/heap.h>
; #include <sys/thread.h>
; #include <sys/event.h>
; #include <sys/timer.h>
; #include <sys/confnet.h>
; 
; #include <net/ether.h>
; #include <net/if_var.h>
; 
; #include <dev/irqreg.h>
; #include <dev/nicrtl.h>
; #include "rtlregs.h"
; 
; #ifdef NUTDEBUG
; #include <sys/osdebug.h>
; #include <net/netdebug.h>
; #endif
; 
; /*
;  * Determine ports, which had not been explicitely configured.
;  */
; #if (RTL_EESK_AVRPORT == AVRPORTB)
; #define RTL_EESK_PIN    PINB
; #define RTL_EESK_DDR    DDRB
; 
; #elif (RTL_EESK_AVRPORT == AVRPORTC)
; #define RTL_EE_MEMBUS
; #define RTL_EESK_PIN    PINC
; #define RTL_EESK_DDR    DDRC
; 
; #elif (RTL_EESK_AVRPORT == AVRPORTD)
; #define RTL_EESK_PIN    PIND
; #define RTL_EESK_DDR    DDRD
; 
; #elif (RTL_EESK_AVRPORT == AVRPORTE)
; #define RTL_EESK_PIN    PINE
; #define RTL_EESK_DDR    DDRE
; 
; #elif (RTL_EESK_AVRPORT == AVRPORTF)
; #define RTL_EESK_PIN    PINF
; #define RTL_EESK_DDR    DDRF
; 
; #endif /* RTL_EESK_AVRPORT */
; 
; #if (RTL_EEDO_AVRPORT == AVRPORTB)
; #define RTL_EEDO_PORT   PORTB
; #define RTL_EEDO_DDR    DDRB
; 
; #elif (RTL_EEDO_AVRPORT == AVRPORTC)
; #define RTL_EE_MEMBUS
; #define RTL_EEDO_PORT   PORTC
; #define RTL_EEDO_DDR    DDRC
; 
; #elif (RTL_EEDO_AVRPORT == AVRPORTD)
; #define RTL_EEDO_PORT   PORTD
; #define RTL_EEDO_DDR    DDRD
; 
; #elif (RTL_EEDO_AVRPORT == AVRPORTE)
; #define RTL_EEDO_PORT   PORTE
; #define RTL_EEDO_DDR    DDRE
; 
; #elif (RTL_EEDO_AVRPORT == AVRPORTF)
; #define RTL_EEDO_PORT   PORTF
; #define RTL_EEDO_DDR    DDRF
; 
; #endif /* RTL_EEDO_AVRPORT */
; 
; #if (RTL_EEMU_AVRPORT == AVRPORTB)
; #define RTL_EEMU_PORT   PORTB
; #define RTL_EEMU_DDR    DDRB
; 
; #elif (RTL_EEMU_AVRPORT == AVRPORTC)
; #define RTL_EE_MEMBUS
; #define RTL_EEMU_PORT   PORTC
; #define RTL_EEMU_DDR    DDRC
; 
; #elif (RTL_EEMU_AVRPORT == AVRPORTD)
; #define RTL_EEMU_PORT   PORTD
; #define RTL_EEMU_DDR    DDRD
; 
; #elif (RTL_EEMU_AVRPORT == AVRPORTE)
; #define RTL_EEMU_PORT   PORTE
; #define RTL_EEMU_DDR    DDRE
; 
; #elif (RTL_EEMU_AVRPORT == AVRPORTF)
; #define RTL_EEMU_PORT   PORTF
; #define RTL_EEMU_DDR    DDRF
; 
; #endif /* RTL_EEMU_AVRPORT */
; 
; #if (RTL_RESET_AVRPORT == AVRPORTB)
; #define RTL_RESET_PORT   PORTB
; #define RTL_RESET_DDR    DDRB
; 
; #elif (RTL_RESET_AVRPORT == AVRPORTD)
; #define RTL_RESET_PORT   PORTD
; #define RTL_RESET_DDR    DDRD
; 
; #elif (RTL_RESET_AVRPORT == AVRPORTE)
; #define RTL_RESET_PORT   PORTE
; #define RTL_RESET_DDR    DDRE
; 
; #elif (RTL_RESET_AVRPORT == AVRPORTF)
; #define RTL_RESET_PORT   PORTF
; #define RTL_RESET_DDR    DDRF
; 
; #endif /* RTL_RESET_AVRPORT */
; 
; 
; /*!
;  * \brief Interrupt used.
;  */
; #if (RTL_SIGNAL_IRQ == INT0)
; #define RTL_SIGNAL sig_INTERRUPT0
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRA, ISC00); sbi(EICRA, ISC01)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT1)
; #define RTL_SIGNAL sig_INTERRUPT1
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRA, ISC10); sbi(EICRA, ISC11)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT2)
; #define RTL_SIGNAL sig_INTERRUPT2
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRA, ISC20); sbi(EICRA, ISC21)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT3)
; #define RTL_SIGNAL sig_INTERRUPT3
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRA, ISC30); sbi(EICRA, ISC31)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT4)
; #define RTL_SIGNAL sig_INTERRUPT4
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRB, ISC40); sbi(EICRB, ISC41)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT6)
; #define RTL_SIGNAL sig_INTERRUPT6
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRB, ISC60); sbi(EICRB, ISC61)
; #endif
; 
; #elif (RTL_SIGNAL_IRQ == INT7)
; #define RTL_SIGNAL sig_INTERRUPT7
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRB, ISC70); sbi(EICRB, ISC71)
; #endif
; 
; #else
; #define RTL_SIGNAL sig_INTERRUPT5
; #ifdef __AVR_ENHANCED__
; #define RTL_RISING_EDGE_MODE()   sbi(EICRB, ISC50); sbi(EICRB, ISC51)
; #endif
; 
; #endif /* RTL_SIGNAL_IRQ */
; 
; 
; /*!
;  * \brief Size of a single ring buffer page.
;  */
; #define NIC_PAGE_SIZE   0x100
; 
; /*!
;  * \brief First ring buffer page address.
;  */
; #define NIC_START_PAGE  0x40
; 
; /*!
;  * \brief Last ring buffer page address plus 1.
;  */
; #define NIC_STOP_PAGE   0x60
; 
; /*!
;  * \brief Number of pages in a single transmit buffer.
;  *
;  * This should be at least the MTU size.
;  */
; #define NIC_TX_PAGES    6
; 
; /*!
;  * \brief Number of transmit buffers.
;  *
;  * The initial value had been 2. The idea was to use two alternating
;  * buffers. However, this had never been implemented and we took over
;  * Bengt Florin's change, defining 1 transmit buffer only and give
;  * more buffer space to the receiver.
;  *
;  * The controller memory layout is now
;  *
;  * - 0x4000..0x45ff  1.5K bytes transmit buffer
;  * - 0x4600..0x5fff  6.5K bytes receive buffer
;  *
;  * The routines are still not using the buffers in an optimal way,
;  * as transmission is limited to 1 packet at a time. In fact several
;  * smaller packets would fit into the 1.5 kByte buffer. On the other
;  * hand, filling the buffer with more than one transmission packet
;  * may result in other bad effects, like pulling the IORDY line more
;  * often.
;  */
; #define NIC_TX_BUFFERS      1
; 
; #define NIC_FIRST_TX_PAGE   NIC_START_PAGE
; #define NIC_FIRST_RX_PAGE   (NIC_FIRST_TX_PAGE + NIC_TX_PAGES * NIC_TX_BUFFERS)
; 
; #define NIC_CR_PAGE0 (0)
; #define NIC_CR_PAGE1 (NIC_CR_PS0)
; #define NIC_CR_PAGE2 (NIC_CR_PS1)
; #define NIC_CR_PAGE3 (NIC_CR_PS1 | NIC_CR_PS0)
; 
; /*
;  * This delay has been added by Bengt Florin and is used to minimize
;  * the effect of the IORDY line during reads. Bengt contributed a
;  * more versatile loop, which unfortunately wasn't portable to the
;  * ImageCraft compiler.
;  *
;  * Both versions depend on the CPU clock and had been tested with
;  * 14.7456 MHz.
;  */
; static INLINE void Delay16Cycles(void)
; {
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
L2:
	.dbline 0 ; func end
	ret
;              j -> R20
;              i -> R22
	.even
_NicReset:
	xcall push_xgsetF000
; }
; 
; /*!
;  * \addtogroup xgNicRtl
;  */
; /*@{*/
; 
; /*!
;  * Realtek packet header.
;  */
; struct nic_pkt_header {
;     u_char ph_status;           /*!< \brief Status, contents of RSR register */
;     u_char ph_nextpg;           /*!< \brief Page for next packet */
;     u_short ph_size;            /*!< \brief Size of header and packet in octets */
; };
; 
; #define NICINB(reg)         (*((volatile u_char *)RTL_BASE_ADDR + reg))
; #define NICOUTB(reg, val)   (*((volatile u_char *)RTL_BASE_ADDR + reg) = val)
; 
; /*!
;  * \brief Reset the Ethernet controller.
;  *
;  */
; static int NicReset(void)
; {
;     u_char i;
;     u_char j;
; 
; /*
;  * Toggle the hardware reset line. Since Ethernut version 1.3 the
;  * hardware reset pin of the nic is no longer connected to bit 4
;  * on port E, but wired to the board reset line.
;  */
; #ifdef RTL_RESET_BIT
;     sbi(RTL_RESET_DDR, RTL_RESET_BIT);
;     sbi(RTL_RESET_PORT, RTL_RESET_BIT);
;     NutDelay(WAIT100);
;     cbi(RTL_RESET_PORT, RTL_RESET_BIT);
;     NutDelay(WAIT250);
;     NutDelay(WAIT250);
; #endif
; 
;     /*
;      * Do the software reset by reading from the reset register followed
;      * by writing to the reset register. Wait until the controller enters
;      * the reset state.
;      */
;     for (j = 0; j < 20; j++) {
	clr R20
	xjmp L7
L4:
;         i = NICINB(NIC_RESET);
	lds R22,33567
;         NutDelay(WAIT5);
	ldi R16,5
	xcall _NutDelay
;         NICOUTB(NIC_RESET, i);
	sts 33567,R22
;         for (i = 0; i < 20; i++) {
	clr R22
	xjmp L11
L8:
;             NutDelay(WAIT50);
	ldi R16,50
	xcall _NutDelay
;             /*
;              * We got the reset bit. However, Ethernut 1.1 may
;              * still fail because the NIC hasn't got it's hardware
;              * reset and the data lines remain in tristate. So we
;              * read noise instead of the register. To solve this
;              * problem, we will verify the NIC's id.
;              */
;             if ((NICINB(NIC_PG0_ISR) & NIC_ISR_RST) != 0 &&     /* */
	lds R2,33543
	sbrs R2,7
	rjmp L12
X0:
	lds R24,33546
	cpi R24,80
	brne L12
X1:
	lds R24,33547
	cpi R24,112
	brne L12
X2:
;                 NICINB(NIC_PG0_RBCR0) == 0x50 &&        /* */
;                 NICINB(NIC_PG0_RBCR1) == 0x70)
;                 return 0;
	clr R16
	clr R17
	xjmp L3
L12:
;         }
L9:
	inc R22
L11:
	cpi R22,20
	brlo L8
X3:
;     }
L5:
	inc R20
L7:
	cpi R20,20
	brlo L4
X4:
;     return -1;
	ldi R16,-1
	ldi R17,-1
L3:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;            cnt -> R20,R21
	.even
_DetectNicEeprom:
	st -y,R20
	st -y,R21
; }
; 
; static int DetectNicEeprom(void)
; {
; #ifdef RTL_EESK_BIT
;     register u_int cnt = 0;
	clr R20
	clr R21
; 
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
; 
;     /*
;      * Prepare the EEPROM emulation port bits. Configure the EEDO
;      * and the EEMU lines as outputs and set both lines to high.
;      */
;     sbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	sbi 0x15,6
;     sbi(RTL_EEDO_DDR, RTL_EEDO_BIT);
	sbi 0x14,6
; #ifdef RTL_EEMU_BIT
;     sbi(RTL_EEMU_PORT, RTL_EEMU_BIT);
	sbi 0x15,7
;     sbi(RTL_EEMU_DDR, RTL_EEMU_BIT);
	sbi 0x14,7
; #endif
;     NutDelay(20);
	ldi R16,20
	xcall _NutDelay
; 
;     /*
;      * Force the chip to re-read the EEPROM contents.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2 | NIC_CR_PS0 | NIC_CR_PS1);
	ldi R24,225
	sts 33536,R24
;     NICOUTB(NIC_PG3_EECR, NIC_EECR_EEM0);
	ldi R24,64
	sts 33537,R24
; 
;     /*
;      * No external memory access beyond this point.
;      */
; #ifdef RTL_EE_MEMBUS
;     /*
;      * No external memory access beyond this point.
;      */
; #ifdef __AVR_ENHANCED__
;     /* On the ATmega 128 we release bits 5-7 as normal port pins. */
;     outb(XMCRB, inb(XMCRB) | _BV(XMM0) | _BV(XMM1));
	lds R24,108
	ori R24,3
	sts 108,R24
; #else
;     /* On the ATmega 103 we have to disable the external memory interface. */
;     cbi(MCUCR, SRE);
; #endif
; #endif
; 
;     /*
;      * Check, if the chip toggles our EESK input. If not, we do not
;      * have EEPROM emulation hardware.
;      */
;     if (bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT)) {
	sbis 0x13,5
	rjmp L22
X5:
L17:
;         while (++cnt && bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT));
L18:
	movw R24,R20
	adiw R24,1
	movw R20,R24
	cpi R24,0
	cpc R24,R25
	breq L20
X6:
	sbic 0x13,5
	rjmp L17
X7:
L20:
;     } else {
	xjmp L16
L21:
;         while (++cnt && bit_is_clear(RTL_EESK_PIN, RTL_EESK_BIT));
L22:
	movw R24,R20
	adiw R24,1
	movw R20,R24
	cpi R24,0
	cpc R24,R25
	breq L24
X8:
	sbis 0x13,5
	rjmp L21
X9:
L24:
;     }
L16:
; 
; #ifdef RTL_EE_MEMBUS
;     /*
;      * Enable memory interface.
;      */
; #ifdef __AVR_ENHANCED__
;     /* On the ATmega 128 we release bits 5-7 as normal port pins. */
;     outb(XMCRB, inb(XMCRB) & ~(_BV(XMM0) | _BV(XMM1)));
	lds R24,108
	andi R24,252
	sts 108,R24
; #else
;     /* On the ATmega 103 we have to disable the external memory interface. */
;     sbi(MCUCR, SRE);
; #endif
; #endif
; 
;     /* Reset port outputs to default. */
;     cbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	cbi 0x15,6
;     cbi(RTL_EEDO_DDR, RTL_EEDO_BIT);
	cbi 0x14,6
; #ifdef RTL_EEMU_BIT
;     cbi(RTL_EEMU_PORT, RTL_EEMU_BIT);
	cbi 0x15,7
;     cbi(RTL_EEMU_DDR, RTL_EEMU_BIT);
	cbi 0x14,7
; #endif
; 
;     /* Restore previous interrupt enable state. */
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
L25:
; 
;     /* Wait until controller ready. */
;     while (NICINB(NIC_CR) != (NIC_CR_STP | NIC_CR_RD2));
L26:
	lds R24,33536
	cpi R24,33
	brne L25
X10:
; 
;     return cnt ? 0 : -1;
	cpi R20,0
	cpc R20,R21
	breq L28
X11:
	clr R20
	clr R21
	xjmp L29
L28:
	ldi R20,-1
	ldi R21,-1
L29:
	movw R16,R20
L14:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
	.area lit(rom, con, rel)
_nic_eeprom:
	.byte 255,255
	.byte 255,48
	.byte 255,255
	.byte 255,255
	.byte 255,255
	.byte 255,255
	.byte 255,255
	.byte 255,255
	.byte 255,255
	.area text(rom, con, rel)
;            val -> R10
;            cnt -> R20
;            clk -> R22
	.even
_EmulateNicEeprom:
	xcall push_xgsetF00C
; #else
;     return -1;
; #endif
; }
; 
; #ifdef RTL_EESK_BIT
; /*
;  * Emulated EEPROM contents.
;  *
;  * In jumper mode our influence is quite limited, only CONFIG3 and CONFIG4
;  * can be modified.
;  */
; static prog_char nic_eeprom[18] = {
;     0xFF,                       /* CONFIG2: jPL1 jPL0   0      jBS4   jBS3   jBS2  jBS1  jBS0  */
;     0xFF,                       /* CONFIG1: 1    jIRQS2 jIRQS1 jIRQS0 jIOS3  jIOS2 jIOS1 jIOS0 */
; 
;     0xFF,                       /* CONFIG4: -    -      -      -      -      -     -     IOMS  */
;     0x30,                       /* CONFIG3  PNP  FUDUP  LEDS1  LEDS0  -      0     PWRDN ACTB  */
; 
;     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, /* MAC */
;     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF      /* ID */
; };
; #endif
; 
; /*!
;  * \brief EEPROM emulator.
;  *
;  * Forces the chip to re-read the EEPROM contents and emulates a serial
;  * EEPROM.
;  *
;  * If the hardware does not support this feature, then this call will
;  * never return. Thus, make sure to have the driver properly configured.
;  */
; static void EmulateNicEeprom(void)
; {
; #ifdef RTL_EESK_BIT
;     register u_char clk;
;     register u_char cnt;
;     register u_char val;
; 
;     /*
;      * Disable all interrupts. This routine requires critical timing
;      * and optionally may disable the memory interface.
;      */
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
; 
;     /*
;      * Prepare the EEPROM emulation port bits. Configure the EEDO and
;      * the EEMU lines as outputs and set EEDO to low and EEMU to high.
;      */
;     cbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	cbi 0x15,6
;     sbi(RTL_EEDO_DDR, RTL_EEDO_BIT);
	sbi 0x14,6
; #ifdef RTL_EEMU_BIT
;     sbi(RTL_EEMU_PORT, RTL_EEMU_BIT);
	sbi 0x15,7
;     sbi(RTL_EEMU_DDR, RTL_EEMU_BIT);
	sbi 0x14,7
; #endif
;     NutDelay(20);
	ldi R16,20
	xcall _NutDelay
; 
;     /*
;      * Start EEPROM configuration. Stop/abort any activity and select
;      * configuration page 3. Setting bit EEM0 will force the controller
;      * to read the EEPROM contents.
;      */
; 
;     /* Select page 3, stop and abort/complete. */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2 | NIC_CR_PS0 | NIC_CR_PS1);
	ldi R24,225
	sts 33536,R24
;     NICOUTB(NIC_PG3_EECR, NIC_EECR_EEM0);
	ldi R24,64
	sts 33537,R24
; 
;     /*
;      * We can avoid wasting port pins for EEPROM emulation by using the
;      * upper bits of the address bus.
;      */
; #ifdef RTL_EE_MEMBUS
;     /*
;      * No external memory access beyond this point.
;      */
; #ifdef __AVR_ENHANCED__
;     /* On the ATmega 128 we release bits 5-7 as normal port pins. */
;     outb(XMCRB, inb(XMCRB) | _BV(XMM0) | _BV(XMM1));
	lds R24,108
	ori R24,3
	sts 108,R24
; #else
;     /* On the ATmega 103 we have to disable the external memory interface. */
;     cbi(MCUCR, SRE);
; #endif
; #endif
; 
;     /*
;      * Loop for all EEPROM words.
;      */
;     for (cnt = 0; cnt < sizeof(nic_eeprom);) {
	clr R20
	xjmp L34
L31:
; 
;         /*
;          *
;          * 1 start bit, always high
;          * 2 op-code bits
;          * 7 address bits
;          * 1 dir change bit, always low
;          */
;         for (clk = 0; clk < 11; clk++) {
	clr R22
	xjmp L38
L39:
;             while (bit_is_clear(RTL_EESK_PIN, RTL_EESK_BIT));
L40:
	sbis 0x13,5
	rjmp L39
X12:
L42:
;             while (bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT));
L43:
	sbic 0x13,5
	rjmp L42
X13:
;         }
L36:
	inc R22
L38:
	cpi R22,11
	brlo L40
X14:
; 
;         /*
;          * Shift out the high byte, MSB first. Our data changes at the EESK
;          * rising edge. Data is sampled by the Realtek at the falling edge.
;          */
;         val = PRG_RDB(nic_eeprom + cnt);
	ldi R24,<_nic_eeprom
	ldi R25,>_nic_eeprom
	mov R30,R20
	clr R31
	add R30,R24
	adc R31,R25
	lpm R10,Z
;         cnt++;
	inc R20
;         for (clk = 0x80; clk; clk >>= 1) {
	ldi R22,128
	xjmp L48
L49:
;             while (bit_is_clear(RTL_EESK_PIN, RTL_EESK_BIT));
L50:
	sbis 0x13,5
	rjmp L49
X15:
;             if (val & clk)
	mov R2,R10
	and R2,R22
	breq L55
X16:
;                 sbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	sbi 0x15,6
L54:
;             while (bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT));
L55:
	sbic 0x13,5
	rjmp L54
X17:
;             cbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	cbi 0x15,6
;         }
L46:
	lsr R22
L48:
	tst R22
	brne L50
X18:
; 
;         /*
;          * Shift out the low byte.
;          */
;         val = PRG_RDB(nic_eeprom + cnt);
	ldi R24,<_nic_eeprom
	ldi R25,>_nic_eeprom
	mov R30,R20
	clr R31
	add R30,R24
	adc R31,R25
	lpm R10,Z
;         cnt++;
	inc R20
;         for (clk = 0x80; clk; clk >>= 1) {
	ldi R22,128
	xjmp L60
L61:
;             while (bit_is_clear(RTL_EESK_PIN, RTL_EESK_BIT));
L62:
	sbis 0x13,5
	rjmp L61
X19:
;             if (val & clk)
	mov R2,R10
	and R2,R22
	breq L67
X20:
;                 sbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	sbi 0x15,6
L66:
;             while (bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT));
L67:
	sbic 0x13,5
	rjmp L66
X21:
;             cbi(RTL_EEDO_PORT, RTL_EEDO_BIT);
	cbi 0x15,6
;         }
L58:
	lsr R22
L60:
	tst R22
	brne L62
X22:
; 
; 
;         /* 5 remaining clock cycles. */
;         for (clk = 0; clk < 5; clk++) {
	clr R22
	xjmp L72
L73:
;             while (bit_is_clear(RTL_EESK_PIN, RTL_EESK_BIT));
L74:
	sbis 0x13,5
	rjmp L73
X23:
L76:
;             while (bit_is_set(RTL_EESK_PIN, RTL_EESK_BIT));
L77:
	sbic 0x13,5
	rjmp L76
X24:
;         }
L70:
	inc R22
L72:
	cpi R22,5
	brlo L74
X25:
;     }
L32:
L34:
	cpi R20,18
	brsh X27
	xjmp L31
X27:
X26:
; 
; #ifdef RTL_EE_MEMBUS
;     /*
;      * Enable memory interface.
;      */
; #ifdef __AVR_ENHANCED__
;     /* On the ATmega 128 we release bits 5-7 as normal port pins. */
;     outb(XMCRB, inb(XMCRB) & ~(_BV(XMM0) | _BV(XMM1)));
	lds R24,108
	andi R24,252
	sts 108,R24
; #else
;     /* On the ATmega 103 we have to disable the external memory interface. */
;     sbi(MCUCR, SRE);
; #endif
; #endif
; 
;     /* Reset port outputs to default. */
;     cbi(RTL_EEDO_DDR, RTL_EEDO_BIT);
	cbi 0x14,6
; #ifdef RTL_EEMU_BIT
;     cbi(RTL_EEMU_PORT, RTL_EEMU_BIT);
	cbi 0x15,7
;     cbi(RTL_EEMU_DDR, RTL_EEMU_BIT);
	cbi 0x14,7
; #endif
; 
;     /* Restore previous interrupt enable state. */
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
L30:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;              i -> R20
;            mac -> R10,R11
	.even
_NicStart:
	xcall push_xgset300C
	movw R10,R16
; #endif
; }
; 
; /*
;  * Fires up the network interface. NIC interrupts
;  * should have been disabled when calling this
;  * function.
;  */
; static int NicStart(CONST u_char * mac)
; {
;     u_char i;
; 
;     if (NicReset()) {
	xcall _NicReset
	cpi R16,0
	cpc R16,R17
	breq L80
X28:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L79
L80:
;     }
;     if (DetectNicEeprom() == 0) {
	xcall _DetectNicEeprom
	cpi R16,0
	cpc R16,R17
	brne L82
X29:
;         EmulateNicEeprom();
	xcall _EmulateNicEeprom
;     }
L82:
; 
;     /*
;      * Mask all interrupts and clear any interrupt status flag to set the
;      * INT pin back to low.
;      */
;     NICOUTB(NIC_PG0_IMR, 0);
	clr R2
	sts 33551,R2
;     NICOUTB(NIC_PG0_ISR, 0xff);
	ldi R24,255
	sts 33543,R24
; 
;     /*
;      * During reset the nic loaded its initial configuration from an
;      * external eeprom. On the ethernut board we do not have any
;      * configuration eeprom, but simply tied the eeprom data line to
;      * high level. So we have to clear some bits in the configuration
;      * register. Switch to register page 3.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2 | NIC_CR_PS0 | NIC_CR_PS1);
	ldi R24,225
	sts 33536,R24
; 
;     /*
;      * The nic configuration registers are write protected unless both
;      * EEM bits are set to 1.
;      */
;     NICOUTB(NIC_PG3_EECR, NIC_EECR_EEM0 | NIC_EECR_EEM1);
	ldi R24,192
	sts 33537,R24
; 
;     /*
;      * Network media had been set to 10Base2 by the virtual EEPROM and
;      * will be set now to auto detect. This will initiate a link test.
;      * We don't force 10BaseT, because this would disable the link test.
;      */
;     NICOUTB(NIC_PG3_CONFIG2, NIC_CONFIG2_BSELB);
	ldi R24,32
	sts 33541,R24
; 
;     /*
;      * Disable sleep and power down.
;      *
;      * The virtual EEPROM (resistor tight to VCC) will set all bits of
;      * CONFIG3 to 1. Unfortunately we are not able to modify the full
;      * duplex bit. The only solution is to use a real EEPROM or emulate
;      * one.
;      */
;     NICOUTB(NIC_PG3_CONFIG3, NIC_CONFIG3_LEDS1 | NIC_CONFIG3_LEDS1);
	sts 33542,R24
; 
;     /*
;      * Reenable write protection of the nic configuration registers
;      * and wait for link test to complete.
;      */
;     NICOUTB(NIC_PG3_EECR, 0);
	sts 33537,R2
;     NutDelay(255);
	ldi R16,255
	xcall _NutDelay
; 
;     /*
;      * Switch to register page 0 and set data configuration register
;      * to byte-wide DMA transfers, normal operation (no loopback),
;      * send command not executed and 8 byte fifo threshold.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2);
	ldi R24,33
	sts 33536,R24
;     NICOUTB(NIC_PG0_DCR, NIC_DCR_LS | NIC_DCR_FT1);
	ldi R24,72
	sts 33550,R24
; 
;     /*
;      * Clear remote dma byte count register.
;      */
;     NICOUTB(NIC_PG0_RBCR0, 0);
	clr R2
	sts 33546,R2
;     NICOUTB(NIC_PG0_RBCR1, 0);
	sts 33547,R2
; 
;     /*
;      * Temporarily set receiver to monitor mode and transmitter to
;      * internal loopback mode. Incoming packets will not be stored
;      * in the nic ring buffer and no data will be send to the network.
;      */
;     NICOUTB(NIC_PG0_RCR, NIC_RCR_MON);
	ldi R24,32
	sts 33548,R24
;     NICOUTB(NIC_PG0_TCR, NIC_TCR_LB0);
	ldi R24,2
	sts 33549,R24
; 
;     /*
;      * Configure the nic's ring buffer page layout.
;      * NIC_PG0_BNRY: Last page read.
;      * NIC_PG0_PSTART: First page of receiver buffer.
;      * NIC_PG0_PSTOP: Last page of receiver buffer.
;      */
;     NICOUTB(NIC_PG0_TPSR, NIC_FIRST_TX_PAGE);
	ldi R24,64
	sts 33540,R24
;     NICOUTB(NIC_PG0_BNRY, NIC_STOP_PAGE - 1);
	ldi R24,95
	sts 33539,R24
;     NICOUTB(NIC_PG0_PSTART, NIC_FIRST_RX_PAGE);
	ldi R24,70
	sts 33537,R24
;     NICOUTB(NIC_PG0_PSTOP, NIC_STOP_PAGE);
	ldi R24,96
	sts 33538,R24
; 
;     /*
;      * Once again clear interrupt status register.
;      */
;     NICOUTB(NIC_PG0_ISR, 0xff);
	ldi R24,255
	sts 33543,R24
; 
;     /*
;      * Switch to register page 1 and copy our MAC address into the nic.
;      * We are still in stop mode.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2 | NIC_CR_PS0);
	ldi R24,97
	sts 33536,R24
;     for (i = 0; i < 6; i++)
	clr R20
	xjmp L87
L84:
;         NICOUTB(NIC_PG1_PAR0 + i, mac[i]);
	mov R30,R20
	clr R31
	add R30,R10
	adc R31,R11
	ldd R2,z+0
	mov R30,R20
	clr R31
	subi R30,255  ; addi 33537
	sbci R31,124
	std z+0,R2
L85:
	inc R20
L87:
	cpi R20,6
	brlo L84
X30:
; 
;     /*
;      * Clear multicast filter bits to disable all packets.
;      */
;     for (i = 0; i < 8; i++)
	clr R20
	xjmp L91
L88:
;         NICOUTB(NIC_PG1_MAR0 + i, 0);
	clr R2
	mov R30,R20
	clr R31
	subi R30,248  ; addi 33544
	sbci R31,124
	std z+0,R2
L89:
	inc R20
L91:
	cpi R20,8
	brlo L88
X31:
; 
;     /*
;      * Set current page pointer to one page after the boundary pointer.
;      */
;     NICOUTB(NIC_PG1_CURR, NIC_FIRST_RX_PAGE);
	ldi R24,70
	sts 33543,R24
; 
;     /*
;      * Switch back to register page 0, remaining in stop mode.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2);
	ldi R24,33
	sts 33536,R24
; 
;     /*
;      * Take receiver out of monitor mode and enable it for accepting
;      * broadcasts.
;      */
;     NICOUTB(NIC_PG0_RCR, NIC_RCR_AB);
	ldi R24,4
	sts 33548,R24
; 
;     /*
;      * Clear all interrupt status flags and enable interrupts.
;      */
;     NICOUTB(NIC_PG0_ISR, 0xff);
	ldi R24,255
	sts 33543,R24
;     /* Note: transmitter if polled, thus no NIC_IMR_PTXE */
;     NICOUTB(NIC_PG0_IMR, NIC_IMR_PRXE | NIC_IMR_RXEE | NIC_IMR_TXEE | NIC_IMR_OVWE);
	ldi R24,29
	sts 33551,R24
; 
;     /*
;      * Fire up the nic by clearing the stop bit and setting the start bit.
;      * To activate the local receive dma we must also take the nic out of
;      * the local loopback mode.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD2);
	ldi R24,34
	sts 33536,R24
;     NICOUTB(NIC_PG0_TCR, 0);
	clr R2
	sts 33549,R2
; 
;     NutDelay(255);
	ldi R16,255
	xcall _NutDelay
; 
;     return 0;
	clr R16
	clr R17
L79:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;              i -> R16
	.even
_NicCompleteDma:
; }
; 
; /*!
;  * Complete remote DMA.
;  */
; static void NicCompleteDma(void)
; {
;     u_char i;
; 
;     /*
;      * Complete remote dma.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD2);
	ldi R24,34
	sts 33536,R24
; 
;     /*
;      * Check that we have a DMA complete flag.
;      */
;     for (i = 0; i <= 20; i++)
	clr R16
	xjmp L96
L93:
;         if (NICINB(NIC_PG0_ISR) & NIC_ISR_RDC)
	lds R2,33543
	sbrs R2,6
	rjmp L97
X32:
;             break;
	xjmp L95
L97:
L94:
	inc R16
L96:
	ldi R24,20
	cp R24,R16
	brsh L93
X33:
L95:
; 
;     /*
;      * Reset remote dma complete flag.
;      */
;     NICOUTB(NIC_PG0_ISR, NIC_ISR_RDC);
	ldi R24,64
	sts 33543,R24
L92:
	.dbline 0 ; func end
	ret
;             il -> R22
;             ih -> R20
;              l -> R22,R23
;            len -> R18,R19
;            buf -> R16,R17
	.even
_NicWrite:
	xcall push_xgsetF000
; }
; 
; /*
;  * Write data block to the NIC.
;  */
; static void NicWrite(u_char * buf, u_short len)
; {
;     register u_short l = len - 1;
	movw R22,R18
	subi R22,1
	sbci R23,0
;     register u_char ih = (u_short) l >> 8;
	movw R20,R22
	mov R20,R21
	clr R21
;     register u_char il = (u_char) l;
; 
;     if (!len)
	cpi R18,0
	cpc R18,R19
	brne L100
X34:
;         return;
	xjmp L99
L100:
L102:
; 
;     do {
L105:
;         do {
;             NICOUTB(NIC_IOPORT, *buf++);
	movw R30,R16
	ld R2,Z+
	movw R16,R30
	sts 33552,R2
;         } while (il-- != 0);
L106:
	mov R2,R22
	clr R3
	subi R22,1
	tst R2
	brne L105
X35:
;     } while (ih-- != 0);
L103:
	mov R2,R20
	clr R3
	subi R20,1
	tst R2
	brne L102
X36:
L99:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;             il -> R22
;             ih -> R20
;              l -> R22,R23
;            len -> R18,R19
;            buf -> R16,R17
	.even
_NicRead:
	xcall push_xgsetF000
; }
; 
; /*
;  * Read data block from the NIC.
;  */
; static void NicRead(u_char * buf, u_short len)
; {
;     register u_short l = len - 1;
	movw R22,R18
	subi R22,1
	sbci R23,0
;     register u_char ih = (u_short) l >> 8;
	movw R20,R22
	mov R20,R21
	clr R21
;     register u_char il = (u_char) l;
; 
;     if (!len)
	cpi R18,0
	cpc R18,R19
	brne L109
X37:
;         return;
	xjmp L108
L109:
L111:
; 
;     do {
L114:
;         do {
;             *buf++ = NICINB(NIC_IOPORT);
	lds R2,33552
	movw R30,R16
	st Z+,R2
	movw R16,R30
;         } while (il-- != 0);
L115:
	mov R2,R22
	clr R3
	subi R22,1
	tst R2
	brne L114
X38:
;     } while (ih-- != 0);
L112:
	mov R2,R20
	clr R3
	subi R20,1
	tst R2
	brne L111
X39:
L108:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
;             sz -> R20,R21
;        padding -> R10
;              i -> R22,R23
;             nb -> R22,R23
	.even
_NicPutPacket:
	xcall push_xgsetF00C
	movw R22,R16
; }
; 
; 
; /*!
;  * \brief Load a packet into the nic's transmit ring buffer.
;  *
;  *
;  * \param base NIC hardware base address.
;  * \param nb Network buffer structure containing the packet to be sent.
;  *           The structure must have been allocated by a previous
;  *           call NutNetBufAlloc().
;  *
;  * \return 0 on success, -1 in case of any errors. Errors
;  *         will automatically release the network buffer
;  *         structure.
;  */
; static int NicPutPacket(NETBUF * nb)
; {
;     u_short sz;
;     u_short i;
;     u_char padding = 0;
	clr R10
; 
;     /*
;      * Calculate the number of bytes to be send. Do not
;      * send packets larger than 1514 bytes.
;      *
;      * The previous version was wrong by specifying a maximum
;      * of 1518, because it didn't take the CRC into account,
;      * which is generated by the hardware and automatically
;      * appended. Thanks to Bengt Florin, who discovered this.
;      */
;     sz = nb->nb_dl.sz + nb->nb_nw.sz + nb->nb_tp.sz + nb->nb_ap.sz;
	movw R30,R22
	ldd R2,z+9
	ldd R3,z+10
	ldd R20,z+5
	ldd R21,z+6
	add R20,R2
	adc R21,R3
	ldd R2,z+13
	ldd R3,z+14
	add R20,R2
	adc R21,R3
	ldd R2,z+17
	ldd R3,z+18
	add R20,R2
	adc R21,R3
;     if (sz > 1514)
	ldi R24,1514
	ldi R25,5
	cp R24,R20
	cpc R25,R21
	brsh L118
X40:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L117
L118:
; 
;     /*
;      * The controller will not append pad bytes,
;      * so we have to do this.
;      */
;     if (sz < 60) {
	cpi R20,60
	ldi R30,0
	cpc R21,R30
	brsh L123
X41:
;         padding = (u_char) (60 - sz);
	ldi R24,60
	ldi R25,0
	mov R10,R24
	sub R10,R20
	sbc R11,R21
;         sz = 60;
	ldi R20,60
	ldi R21,0
;     }
	xjmp L123
L122:
; 
;     /*
;      * Bengt Florin introduces polling mode for the transmitter. Be
;      * aware, that this may introduce other problems. If a high
;      * priority thread is waiting for the transmitter, it may hold
;      * the CPU for more than 1.2 milliseconds in worst cases.
;      */
;     while (NICINB(NIC_CR) & NIC_CR_TXP)
;         NutThreadYield();
	xcall _NutThreadYield
L123:
	lds R2,33536
	sbrc R2,2
	rjmp L122
X42:
; 
;     /* we don't want to be interrupted by NIC owerflow */
;     cbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	andi R24,223
	out 0x39,R24
; 
;     /*
;      * Set remote dma byte count
;      * and start address.
;      */
;     NICOUTB(NIC_PG0_RBCR0, sz);
	sts 33546,R20
;     NICOUTB(NIC_PG0_RBCR1, sz >> 8);
	movw R2,R20
	mov R2,R3
	clr R3
	sts 33547,R2
;     NICOUTB(NIC_PG0_RSAR0, 0);
	clr R2
	sts 33544,R2
;     NICOUTB(NIC_PG0_RSAR1, NIC_FIRST_TX_PAGE);
	ldi R24,64
	sts 33545,R24
; 
;     /*
;      * Peform the write.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD1);
	ldi R24,18
	sts 33536,R24
; 
;     /*
;      * Transfer the Ethernet frame.
;      */
;     NicWrite(nb->nb_dl.vp, nb->nb_dl.sz);
	movw R30,R22
	ldd R18,z+5
	ldd R19,z+6
	ldd R16,z+3
	ldd R17,z+4
	xcall _NicWrite
;     NicWrite(nb->nb_nw.vp, nb->nb_nw.sz);
	movw R30,R22
	ldd R18,z+9
	ldd R19,z+10
	ldd R16,z+7
	ldd R17,z+8
	xcall _NicWrite
;     NicWrite(nb->nb_tp.vp, nb->nb_tp.sz);
	movw R30,R22
	ldd R18,z+13
	ldd R19,z+14
	ldd R16,z+11
	ldd R17,z+12
	xcall _NicWrite
;     NicWrite(nb->nb_ap.vp, nb->nb_ap.sz);
	movw R30,R22
	ldd R18,z+17
	ldd R19,z+18
	ldd R16,z+15
	ldd R17,z+16
	xcall _NicWrite
; 
;     /*
;      * Add pad bytes.
;      */
;     for (i = 0; i < padding; i++)
	clr R22
	clr R23
	xjmp L128
L125:
;         NICOUTB(NIC_IOPORT, 0);
	clr R2
	sts 33552,R2
L126:
	subi R22,255  ; offset = 1
	sbci R23,255
L128:
	mov R2,R10
	clr R3
	cp R22,R2
	cpc R23,R3
	brlo L125
X43:
; 
;     /*
;      * Complete remote dma.
;      */
;     NicCompleteDma();
	xcall _NicCompleteDma
; 
;     /*
;      * Number of bytes to be transmitted.
;      */
;     NICOUTB(NIC_PG0_TBCR0, (sz & 0xff));
	movw R24,R20
	andi R25,0
	sts 33541,R24
;     NICOUTB(NIC_PG0_TBCR1, ((sz >> 8) & 0xff));
	movw R24,R20
	mov R24,R25
	clr R25
	andi R25,0
	sts 33542,R24
; 
;     /*
;      * First page of packet to be transmitted.
;      */
;     NICOUTB(NIC_PG0_TPSR, NIC_FIRST_TX_PAGE);
	ldi R24,64
	sts 33540,R24
; 
;     /*
;      * Start transmission.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_TXP | NIC_CR_RD2);
	ldi R24,38
	sts 33536,R24
; 
;     sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
; 
;     return 0;
	clr R16
	clr R17
L117:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;        nextpg1 -> R14
;          count -> R10,R11
;           curr -> R14
;           drop -> R10
;             nb -> R20,R21
;           bnry -> R12
;         nextpg -> R22
;            hdr -> y+2
;            buf -> R14,R15
;              i -> R22,R23
	.even
_NicGetPacket:
	xcall push_xgsetF0FC
	sbiw R28,6
; }
; 
; /*!
;  * \brief Fetch the next packet out of the receive ring buffer.
;  *
;  * Nic interrupts must be disabled when calling this funtion.
;  *
;  * \return Pointer to an allocated ::NETBUF. If there is no
;  *         no data available, then the function returns a
;  *         null pointer. If the NIC's buffer seems to be
;  *         corrupted, a pointer to 0xFFFF is returned.
;  */
; static NETBUF *NicGetPacket(void)
; {
;     NETBUF *nb = 0;
	clr R20
	clr R21
;     struct nic_pkt_header hdr;
;     u_short count;
;     u_char *buf;
;     u_char nextpg;
;     u_char bnry;
;     u_char curr;
;     u_short i;
;     u_char drop = 0;
	clr R10
; 
;     /* we don't want to be interrupted by NIC owerflow */
;     cbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	andi R24,223
	out 0x39,R24
;     /*
;      * Get the current page pointer. It points to the page where the NIC
;      * will start saving the next incoming packet.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD2 | NIC_CR_PS0);
	ldi R24,98
	sts 33536,R24
;     Delay16Cycles();
	xcall _Delay16Cycles
;     curr = NICINB(NIC_PG1_CURR);
	lds R14,33543
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD2);
	ldi R24,34
	sts 33536,R24
; 
;     /*
;      * Get the pointer to the last page we read from. The following page
;      * is the one where we start reading. If it's equal to the current
;      * page pointer, then there's nothing to read. In this case we return
;      * a null pointer.
;      */
;     if ((bnry = NICINB(NIC_PG0_BNRY) + 1) >= NIC_STOP_PAGE)
	lds R24,33539
	subi R24,255    ; addi 1
	mov R12,R24
	cpi R24,96
	brlo L130
X44:
;         bnry = NIC_FIRST_RX_PAGE;
	ldi R24,70
	mov R12,R24
L130:
; 
;     if (bnry == curr) {
	cp R12,R14
	brne L132
X45:
;         sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
;         return 0;
	clr R16
	clr R17
	xjmp L129
L132:
;     }
; 
;     /*
;      * Read the NIC specific packet header.
;      */
;     NICOUTB(NIC_PG0_RBCR0, sizeof(struct nic_pkt_header));
	ldi R24,4
	sts 33546,R24
;     NICOUTB(NIC_PG0_RBCR1, 0);
	clr R2
	sts 33547,R2
;     NICOUTB(NIC_PG0_RSAR0, 0);
	sts 33544,R2
;     NICOUTB(NIC_PG0_RSAR1, bnry);
	sts 33545,R12
;     buf = (u_char *) & hdr;
	movw R24,R28
	adiw R24,2
	movw R14,R24
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD0);
	ldi R24,10
	sts 33536,R24
;     Delay16Cycles();
	xcall _Delay16Cycles
;     for (i = 0; i < sizeof(struct nic_pkt_header); i++)
	clr R22
	clr R23
	xjmp L137
L134:
;         *buf++ = NICINB(NIC_IOPORT);
	lds R2,33552
	movw R30,R14
	st Z+,R2
	movw R14,R30
L135:
	subi R22,255  ; offset = 1
	sbci R23,255
L137:
	cpi R22,4
	ldi R30,0
	cpc R23,R30
	brlo L134
X46:
;     NicCompleteDma();
	xcall _NicCompleteDma
; 
;     /*
;      *  Check packet length. Silently discard packets of illegal size.
;      */
;     if (hdr.ph_size < 60 + sizeof(struct nic_pkt_header) ||     /* */
	ldd R24,y+4
	ldd R25,y+5
	cpi R24,64
	ldi R30,0
	cpc R25,R30
	brlo L142
X47:
	ldi R24,1518
	ldi R25,5
	ldd R2,y+4
	ldd R3,y+5
	cp R24,R2
	cpc R25,R3
	brsh L138
X48:
L142:
;         hdr.ph_size > 1514 + sizeof(struct nic_pkt_header)) {
;         drop = 1;
	clr R10
	inc R10
;     }
L138:
; 
;     /*
;      * Calculate the page of the next packet. If it differs from the
;      * pointer in the packet header, we return with errorcode.
;      */
;     nextpg = bnry + (hdr.ph_size >> 8) + ((hdr.ph_size & 0xFF) != 0);
	ldd R24,y+4
	ldd R25,y+5
	andi R25,0
	cpi R24,0
	cpc R24,R25
	breq L145
X49:
	ldi R22,1
	ldi R23,0
	xjmp L146
L145:
	clr R22
	clr R23
L146:
	ldd R2,y+4
	ldd R3,y+5
	mov R2,R3
	clr R3
	mov R4,R12
	clr R5
	add R4,R2
	adc R5,R3
	add R4,R22
	adc R5,R23
	mov R22,R4
;     if (nextpg >= NIC_STOP_PAGE) {
	cpi R22,96
	brlo L147
X50:
;         nextpg -= NIC_STOP_PAGE;
	subi R22,96
;         nextpg += NIC_FIRST_RX_PAGE;
	subi R22,186    ; addi 70
;     }
L147:
;     if (nextpg != hdr.ph_nextpg) {
	ldd R2,y+3
	cp R22,R2
	breq L149
X51:
;         u_char nextpg1 = nextpg + 1;
	mov R24,R22
	subi R24,255    ; addi 1
	mov R14,R24
;         if (nextpg1 >= NIC_STOP_PAGE) {
	cpi R24,96
	brlo L152
X52:
;             nextpg1 -= NIC_STOP_PAGE;
	subi R24,96
;             nextpg1 += NIC_FIRST_RX_PAGE;
	subi R24,186    ; addi 70
	mov R14,R24
;         }
L152:
;         if (nextpg1 != hdr.ph_nextpg) {
	ldd R2,y+3
	cp R14,R2
	breq L154
X53:
;             sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
;             return (NETBUF *) 0xFFFF;
	ldi R16,65535
	ldi R17,255
	xjmp L129
L154:
;         }
;         nextpg = nextpg1;
	mov R22,R14
;     }
L149:
; 
;     /*
;      * Check packet status. It should have set bit 0, but
;      * even without this bit packets seem to be OK.
;      */
;     if (!drop && ((hdr.ph_status & 0x0E) == 0)) {
	tst R10
	brne L157
X54:
	ldd R24,y+2
	andi R24,14
	brne L157
X55:
;         /*
;          * Allocate a NETBUF.
;          * Omit the fcs.
;          */
;         count = hdr.ph_size - 4;
	ldd R24,y+4
	ldd R25,y+5
	sbiw R24,4
	movw R10,R24
;         if ((nb = NutNetBufAlloc(0, NBAF_DATALINK, count))) {
	std y+1,R11
	std y+0,R10
	ldi R18,16
	clr R16
	clr R17
	xcall _NutNetBufAlloc
	movw R20,R16
	cpi R20,0
	cpc R20,R21
	breq L160
X56:
;             /*
;              * Set remote dma byte count and
;              * start address. Don't read the
;              * header again.
;              */
;             NICOUTB(NIC_PG0_RBCR0, count);
	sts 33546,R10
;             NICOUTB(NIC_PG0_RBCR1, count >> 8);
	movw R2,R10
	mov R2,R3
	clr R3
	sts 33547,R2
;             NICOUTB(NIC_PG0_RSAR0, sizeof(struct nic_pkt_header));
	ldi R24,4
	sts 33544,R24
;             NICOUTB(NIC_PG0_RSAR1, bnry);
	sts 33545,R12
; 
;             /*
;              * Perform the read.
;              */
;             NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD0);
	ldi R24,10
	sts 33536,R24
;             Delay16Cycles();
	xcall _Delay16Cycles
;             NicRead(nb->nb_dl.vp, count);
	movw R18,R10
	movw R30,R20
	ldd R16,z+3
	ldd R17,z+4
	xcall _NicRead
;             NicCompleteDma();
	xcall _NicCompleteDma
;         }
L160:
;     }
L157:
; 
;     /*
;      * Set boundary register to the last page we read.
;      * This also drops packets with errors
;      */
;     if (--nextpg < NIC_FIRST_RX_PAGE)
	mov R24,R22
	subi R24,1
	mov R10,R24
	mov R22,R24
	cpi R22,70
	brsh L162
X57:
;         nextpg = NIC_STOP_PAGE - 1;
	ldi R22,95
L162:
;     NICOUTB(NIC_PG0_BNRY, nextpg);
	sts 33539,R22
; 
;     sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
;     return nb;
	movw R16,R20
L129:
	.dbline 0 ; func end
	adiw R28,6
	xjmp pop_xgsetF0FC
;             cr -> R10
;           curr -> R18
;         resend -> R16
	.even
_NicOverflow:
	st -y,R10
	st -y,R11
; }
; 
; /*
;  * \brief Handle NIC overflows.
;  *
;  * When a receiver buffer overflow occurs, the NIC will defer any subsequent
;  * action until properly restarted.
;  *
;  * This routine is called within interrupt context, which introduces a big
;  * problem. It waits for the last transmission to finish, which may take
;  * several milliseconds. Since Nut/OS 3.5, we do not support nested interrupts
;  * on AVR systems anymore. So this routine may now increase interrupt
;  * latency in an unacceptable way. The solution might be to handle overflows
;  * in the receiver thread.
;  *
;  * In any case, this routines needs a major redesign. But it has been
;  * tested in its current form to gracefully withstand ping floods. Thanks
;  * to Bengt Florin for contributing his code, which provides much more
;  * stability than its predecessor.
;  */
; static u_char NicOverflow(void)
; {
;     u_char cr;
;     u_char resend = 0;
	clr R16
L165:
;     u_char curr;
; 
;     /*
;      * Wait for any transmission in progress. Save the command register,
;      * so we can later determine, if NIC transmitter has been interrupted.
;      * or reception in progress.
;      */
;     while (NICINB(NIC_CR) & NIC_CR_TXP);
L166:
	lds R2,33536
	sbrc R2,2
	rjmp L165
X58:
;     cr = NICINB(NIC_CR);
	lds R10,33536
; 
;     /*
;      * Get the current page pointer. It points to the page where the NIC
;      * will start saving the next incoming packet.
;      */
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2 | NIC_CR_PS0);
	ldi R24,97
	sts 33536,R24
;     curr = NICINB(NIC_PG1_CURR);
	lds R18,33543
;     NICOUTB(NIC_CR, NIC_CR_STP | NIC_CR_RD2);
	ldi R24,33
	sts 33536,R24
; 
;     /* Clear remote byte count register. */
;     NICOUTB(NIC_PG0_RBCR0, 0);
	clr R2
	sts 33546,R2
;     NICOUTB(NIC_PG0_RBCR1, 0);
	sts 33547,R2
; 
;     /* Check for any incomplete transmission. */
;     if ((cr & NIC_CR_TXP) && ((NICINB(NIC_PG0_ISR) & (NIC_ISR_PTX | NIC_ISR_TXE)) == 0)) {
	sbrs R10,2
	rjmp L168
X59:
	lds R24,33543
	andi R24,10
	brne L168
X60:
;         resend = 1;
	ldi R16,1
;     }
L168:
; 
;     /* Enter loopback mode and restart the NIC. */
;     NICOUTB(NIC_PG0_TCR, NIC_TCR_LB0);
	ldi R24,2
	sts 33549,R24
;     NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_RD2);
	ldi R24,34
	sts 33536,R24
; 
;     /*
;      * Discard all packets from the receiver buffer. Set boundary
;      * register to the last page we read.
;      */
;     if (--curr < NIC_FIRST_RX_PAGE) {
	mov R24,R18
	subi R24,1
	mov R18,R24
	cpi R18,70
	brsh L170
X61:
;         curr = NIC_STOP_PAGE - 1;
	ldi R18,95
;     }
L170:
;     NICOUTB(NIC_PG0_BNRY, curr);
	sts 33539,R18
; 
;     /* Switch from loopback to normal mode mode. */
;     NICOUTB(NIC_PG0_TCR, 0);
	clr R2
	sts 33549,R2
; 
;     /* Re-invoke any interrupted transmission. */
;     if (resend) {
	tst R16
	breq L172
X62:
;         NICOUTB(NIC_CR, NIC_CR_STA | NIC_CR_TXP | NIC_CR_RD2);
	ldi R24,38
	sts 33536,R24
;     }
L172:
; 
;     /* Finally clear the overflow flag */
;     NICOUTB(NIC_PG0_ISR, NIC_ISR_OVW);
	ldi R24,16
	sts 33543,R24
;     return resend;
L164:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	ret
;             ni -> R10,R11
;            isr -> R12
;            arg -> R10,R11
	.even
_NicInterrupt:
	xcall push_xgsetF0FC
	movw R10,R16
; }
; 
; 
; /*
;  * \brief NIC interrupt entry.
;  */
; static void NicInterrupt(void *arg)
; {
;     u_char isr;
;     NICINFO *ni = (NICINFO *) ((NUTDEVICE *) arg)->dev_dcb;
	movw R30,R10
	ldd R10,z+17
	ldd R11,z+18
; 
;     ni->ni_interrupts++;
	movw R24,R10
	adiw R24,10
	ldi R20,1
	ldi R21,0
	ldi R22,0
	ldi R23,0
	movw R30,R24
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
; 
;     isr = NICINB(NIC_PG0_ISR);
	lds R12,33543
;     NICOUTB(NIC_PG0_ISR, isr);
	sts 33543,R12
; 
;     /*
;      * Recover from receive buffer overflow. This may take some
;      * time, so we enable global interrupts but keep NIC
;      * interrupts disabled.
;      */
;     if (isr & NIC_ISR_OVW) {
	sbrs R12,4
	rjmp L175
X63:
;         /* The AVR platform uses a dedicated interrupt stack, which
;          * forbids interrupt nesting. */
; #if !defined(__AVR__)
;         cbi(EIMSK, RTL_SIGNAL_IRQ);
;         sei();
; #endif
;         NicOverflow();
	xcall _NicOverflow
; #if !defined(__AVR__)
;         cli();
;         sbi(EIMSK, RTL_SIGNAL_IRQ);
; #endif
;         ni->ni_rx_overruns++;
	movw R24,R10
	adiw R24,14
	ldi R20,1
	ldi R21,0
	ldi R22,0
	ldi R23,0
	movw R30,R24
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;     } else {
	xjmp L176
L175:
;         /*
;          * If this is a transmit interrupt, then a packet has been sent.
;          * So we can clear the transmitter busy flag and wake up the
;          * transmitter thread.
;          */
;         if (isr & NIC_ISR_TXE)
	sbrs R12,3
	rjmp L177
X64:
;             ni->ni_tx_errors++;
	movw R24,R10
	adiw R24,34
	movw R14,R24
	ldi R20,1
	ldi R21,0
	ldi R22,0
	ldi R23,0
	movw R30,R24
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
L177:
; 
;         /*
;          * If this is a receive interrupt, then wake up the receiver
;          * thread.
;          */
;         if (isr & NIC_ISR_PRX)
	sbrs R12,0
	rjmp L179
X65:
;             NutEventPostFromIrq(&ni->ni_rx_rdy);
	movw R16,R10
	xcall _NutEventPostFromIrq
L179:
; 
;         if (isr & NIC_ISR_RXE) {
	sbrs R12,2
	rjmp L181
X66:
;             ni->ni_rx_frame_errors += NICINB(NIC_PG0_CNTR0);
	movw R24,R10
	adiw R24,18
	lds R2,33549
	clr R3
	clr R4
	clr R5
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;             ni->ni_rx_crc_errors += NICINB(NIC_PG0_CNTR1);
	movw R24,R10
	adiw R24,22
	movw R12,R24
	lds R2,33550
	clr R3
	clr R4
	clr R5
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;             ni->ni_rx_missed_errors += NICINB(NIC_PG0_CNTR2);
	movw R24,R10
	adiw R24,26
	movw R10,R24
	lds R2,33551
	clr R3
	clr R4
	clr R5
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;         }
L181:
;     }
L176:
L174:
	.dbline 0 ; func end
	xjmp pop_xgsetF0FC
;            dev -> R22,R23
;             ni -> R10,R11
;            ifn -> R20,R21
;             nb -> R12,R13
;            arg -> R10,R11
	.even
_NicRx::
	xcall push_xgsetF0FC
	movw R10,R16
	sbiw R28,2
; }
; 
; /*! \fn NicRx(void *arg)
;  * \brief NIC receiver thread.
;  *
;  *
;  * It runs with high priority.
;  */
; THREAD(NicRx, arg)
; {
;     NUTDEVICE *dev;
;     IFNET *ifn;
;     NICINFO *ni;
;     NETBUF *nb;
; 
;     dev = arg;
	movw R22,R10
;     ifn = (IFNET *) dev->dev_icb;
	movw R30,R22
	ldd R20,z+15
	ldd R21,z+16
;     ni = (NICINFO *) dev->dev_dcb;
	ldd R10,z+17
	ldd R11,z+18
; 
;     NutThreadSetPriority(9);
	ldi R16,9
	xcall _NutThreadSetPriority
;     /*
;      * This is a temporary hack. Due to a change in initialization,
;      * we may not have got a MAC address yet. Wait until one has been
;      * set.
;      */
;     if ((ifn->if_mac[0] & ifn->if_mac[1] & ifn->if_mac[2]) == 0xFF) {
	movw R30,R20
	ldd R2,z+2
	ldd R24,z+1
	and R24,R2
	ldd R2,z+3
	and R24,R2
	cpi R24,255
	breq X73
	xjmp L190
X73:
X67:
	xjmp L187
L186:
;         while ((ifn->if_mac[0] & ifn->if_mac[1] & ifn->if_mac[2]) == 0xFF)
;             NutSleep(125);
	ldi R16,125
	ldi R17,0
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
L187:
	movw R30,R20
	ldd R2,z+2
	ldd R24,z+1
	and R24,R2
	ldd R2,z+3
	and R24,R2
	cpi R24,255
	breq L186
X68:
;         cbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	andi R24,223
	out 0x39,R24
;         NicStart(ifn->if_mac);
	movw R16,R30
	subi R16,255  ; offset = 1
	sbci R17,255
	xcall _NicStart
;         sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
;     }
	xjmp L190
L189:
; 
;     while (1) {
;         NutEventWait(&ni->ni_rx_rdy, 0);
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	movw R16,R10
	xcall _NutEventWait
L192:
;         /*
;          * Fetch all packets from the NIC's internal
;          * buffer and pass them to the registered handler.
;          */
;         do {
;             nb = NicGetPacket();
	xcall _NicGetPacket
	movw R12,R16
; 
;             /* The sanity check may fail because the controller is too busy.
;                restart the NIC. */
;             if ((u_short) nb == 0xFFFF) {
	cpi R16,255
	ldi R30,255
	cpc R17,R30
	brne L195
X69:
;                 NicStart(ifn->if_mac);
	movw R16,R20
	subi R16,255  ; offset = 1
	sbci R17,255
	xcall _NicStart
;                 ni->ni_rx_size_errors++;
	movw R24,R10
	adiw R24,30
	movw R6,R24
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R6
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;             } else if (nb) {
	xjmp L196
L195:
	tst R12
	brne X70
	tst R13
	breq L197
X70:
;                 ni->ni_rx_packets++;
	movw R24,R10
	adiw R24,2
	movw R14,R24
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R14
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;                 (*ifn->if_recv) (dev, nb);
	movw R18,R12
	movw R16,R22
	movw R30,R20
	ldd R26,z+25
	ldd R27,z+26
	movw R30,R26
	xcall  xicall
;             }
L197:
L196:
;         } while (nb);
L193:
	tst R12
	breq X74
	xjmp L192
X74:
	tst R13
	breq X75
	xjmp L192
X75:
X71:
;     }
L190:
	xjmp L189
X72:
L183:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF0FC
;             ni -> R22,R23
;             rc -> R20,R21
;             nb -> R10,R11
;            dev -> R22,R23
	.even
_NicOutput::
	xcall push_xgsetF00C
	movw R10,R18
	movw R22,R16
; }
; 
; /*!
;  * \brief Send Ethernet packet.
;  *
;  * \param dev   Identifies the device to use.
;  * \param nb    Network buffer structure containing the packet to be sent.
;  *              The structure must have been allocated by a previous
;  *              call NutNetBufAlloc().
;  *
;  * \return 0 on success, -1 in case of any errors.
;  */
; int NicOutput(NUTDEVICE * dev, NETBUF * nb)
; {
;     int rc = -1;
	ldi R20,-1
	ldi R21,-1
;     NICINFO *ni = (NICINFO *) dev->dev_dcb;
	movw R30,R22
	ldd R22,z+17
	ldd R23,z+18
; 
;     if (NicPutPacket(nb) == 0) {
	movw R16,R10
	xcall _NicPutPacket
	cpi R16,0
	cpc R16,R17
	brne L200
X76:
;         ni->ni_tx_packets++;
	movw R24,R22
	adiw R24,6
	movw R6,R24
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R6
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;         rc = 0;
	clr R20
	clr R21
;     }
L200:
;     return rc;
	movw R16,R20
L199:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;             ni -> R10,R11
;            ifn -> R20,R21
;            dev -> R22,R23
	.even
_NicInit::
	xcall push_xgsetF00C
	movw R22,R16
	sbiw R28,4
; }
; 
; /*!
;  * \brief Initialize Ethernet hardware.
;  *
;  * Resets RTL8019AS Ethernet controller, initializes all required
;  * hardware registers and starts a background thread for incoming
;  * Ethernet traffic.
;  *
;  * Applications should do not directly call this function. It is
;  * automatically executed during during device registration by
;  * NutRegisterDevice(). Note, that base address and interrupt number
;  * passed to NutRegisterDevice() are silently ignored by this driver
;  * for performance reasons. These values can be changed only by
;  * using the Nut/OS Configurator to rebuild the system.
;  *
;  * If the network configuration hasn't been set by the application
;  * before registering the specified device, this function will
;  * call NutNetLoadConfig() to get the MAC address.
;  *
;  * \param dev Identifies the device to initialize.
;  */
; int NicInit(NUTDEVICE * dev)
; {
;     IFNET *ifn;
;     NICINFO *ni;
; 
;     /*
;      * We need to know our MAC address. If no configuration is
;      * available, load it now.
;      */
;     if (confnet.cd_size == 0)
	lds R2,_confnet
	tst R2
	brne L203
X77:
;         NutNetLoadConfig(dev->dev_name);
	movw R16,R22
	subi R16,254  ; offset = 2
	sbci R17,255
	xcall _NutNetLoadConfig
L203:
; 
;     ifn = dev->dev_icb;
	movw R30,R22
	ldd R20,z+15
	ldd R21,z+16
;     memcpy(ifn->if_mac, confnet.cdn_mac, 6);
	ldi R24,6
	ldi R25,0
	std y+1,R25
	std y+0,R24
	ldi R18,<_confnet+10
	ldi R19,>_confnet+10
	movw R16,R20
	subi R16,255  ; offset = 1
	sbci R17,255
	xcall _memcpy
;     ni = (NICINFO *) dev->dev_dcb;
	movw R30,R22
	ldd R10,z+17
	ldd R11,z+18
;     memset(ni, 0, sizeof(NICINFO));
	ldi R24,38
	ldi R25,0
	std y+1,R25
	std y+0,R24
	clr R18
	clr R19
	movw R16,R10
	xcall _memset
; 
;     /*
;      * Start the receiver thread.
;      */
;     NutThreadCreate("rxi5", NicRx, dev, 640);
	ldi R24,640
	ldi R25,2
	std y+3,R25
	std y+2,R24
	std y+1,R23
	std y+0,R22
	ldi R18,<PL_NicRx
	ldi R19,>PL_NicRx
	ldi R16,<L206
	ldi R17,>L206
	xcall _NutThreadCreate
;     NutSleep(WAIT500);
	ldi R16,244
	ldi R17,1
	ldi R18,0
	ldi R19,0
	xcall _NutSleep
; 
;     /*
;      * Register interrupt handler and enable interrupts.
;      */
;     if (NutRegisterIrqHandler(&RTL_SIGNAL, NicInterrupt, dev))
	std y+1,R23
	std y+0,R22
	ldi R18,<PL_NicInterrupt
	ldi R19,>PL_NicInterrupt
	ldi R16,<_sig_INTERRUPT5
	ldi R17,>_sig_INTERRUPT5
	xcall _NutRegisterIrqHandler
	cpi R16,0
	cpc R16,R17
	breq L207
X78:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L202
L207:
; 
;     cbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	andi R24,223
	out 0x39,R24
; #ifdef RTL_IRQ_RISING_EDGE
;     /* Support of rising edge interrupts for HW w/o inverter gate */
;     RTL_RISING_EDGE_MODE();
; #endif
; 
;     if (ifn->if_mac[0] | ifn->if_mac[1] | ifn->if_mac[2])
	movw R30,R20
	ldd R2,z+2
	ldd R3,z+1
	or R3,R2
	ldd R2,z+3
	or R3,R2
	breq L209
X79:
;         if (NicStart(ifn->if_mac))
	movw R16,R30
	subi R16,255  ; offset = 1
	sbci R17,255
	xcall _NicStart
	cpi R16,0
	cpc R16,R17
	breq L211
X80:
;             return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L202
L211:
L209:
; 
;     sbi(EIMSK, RTL_SIGNAL_IRQ);
	in R24,0x39
	ori R24,32
	out 0x39,R24
; 
;     return 0;
	clr R16
	clr R17
L202:
	.dbline 0 ; func end
	adiw R28,4
	xjmp pop_xgsetF00C
	.area data(ram, con, rel)
L206:
	.blkb 5
	.area idata
	.byte 'r,'x,'i,53,0
	.area data(ram, con, rel)
; }
; 
; 
; /*@}*/
	.area func_lit
PL_NicInterrupt:	.word `_NicInterrupt
PL_NicRx:	.word `_NicRx
