	.module vs1001k.c
	.area data(ram, con, rel)
_vs_status:
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.area text(rom, con, rel)
;              b -> R16
	.even
_VsSdiPutByte:
; /*
;  * Copyright (C) 2003 by Pavel Chromy. All rights reserved.
;  * Copyright (C) 2001-2003 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  * -
;  *
;  * This software has been inspired by all the valuable work done by
;  * Jesper Hansen <jesperh@telia.com>. Many thanks for all his help.
;  */
; 
; /*
;  * $Log: vs1001k.c,v $
;  * Revision 1.1  2005/07/26 18:02:40  haraldkipp
;  * Moved from dev.
;  *
;  * Revision 1.3  2004/03/16 16:48:27  haraldkipp
;  * Added Jan Dubiec's H8/300 port.
;  *
;  * Revision 1.2  2003/07/21 18:06:34  haraldkipp
;  * Buffer function removed. The driver is now using the banked memory routines.
;  * New functions allows the application to enable/disable decoder interrupts.
;  *
;  * Revision 1.1.1.1  2003/05/09 14:40:58  haraldkipp
;  * Initial using 3.2.1
;  *
;  * Revision 1.12  2003/05/06 18:35:21  harald
;  * ICCAVR port
;  *
;  * Revision 1.11  2003/04/21 16:43:54  harald
;  * Added more comments.
;  * Avoid initializing static globals to zero.
;  * New function VsSdiWrite/_P checks DREQ
;  * Removed decoder interrupt en/disable from low level routines.
;  * Keep decoder in reset state until ports have been initialized.
;  * Do not send initial zero bytes as the datasheet recommends.
;  * A single nop is sufficient delay during reset active.
;  * Clear interrupt flag after reset to avoid useless interrupt.
;  * Available buffer size corrected.
;  * New function to read header information.
;  * New function invokes decoder memory test.
;  * Beep makes use of VsSdiWrite.
;  *
;  * Revision 1.10  2003/04/18 14:46:08  harald
;  * Copyright update by the maintainer, after none of the original code had
;  * been left. We have a clean BSD licence now.
;  * This release had been prepared by Pavel Chromy.
;  * BSYNC vs. transfer in progress issue in VsSdiPutByte().
;  * Fixed possible transfer in progress issue in VsPlayerFeed().
;  * HW reset may be forced by VS_SM_RESET mode bit
;  *
;  * Revision 1.9  2003/04/07 20:29:20  harald
;  * Redesigned by Pavel Chromy
;  *
;  * Revision 1.9  2003/04/04 15:01:00  mac
;  * VS_STATUS_EMTY is reported correctly.
;  *
;  * Revision 1.9  2003/02/14 13:39:00  mac
;  * Several serious bugs fixed,
;  * interrupt routine completely remade.
;  * Unreliable spurious interrupts detection removed.
;  * Mpeg frame detection removed.
;  * Watermark check removed (this was rather limiting)
;  * Can be optionaly compiled not to use SPI
;  *
;  * Revision 1.8  2003/02/04 17:50:55  harald
;  * Version 3 released
;  *
;  * Revision 1.7  2003/01/14 16:15:19  harald
;  * Sending twice the number of zeros to end MP3 stream.
;  * Check for spurious interrupts to detect hanging chip.
;  * Simpler portable inline assembler for short delays.
;  *
;  * Revision 1.6  2002/11/02 15:15:13  harald
;  * Library dependencies removed
;  *
;  * Revision 1.5  2002/09/15 16:44:14  harald
;  * *** empty log message ***
;  *
;  * Revision 1.4  2002/08/16 17:49:02  harald
;  * First public release
;  *
;  * Revision 1.3  2002/06/26 17:29:08  harald
;  * First pre-release with 2.4 stack
;  *
;  */
; 
; #include <sys/atom.h>
; #include <sys/event.h>
; #include <sys/timer.h>
; #include <sys/heap.h>
; 
; #include <dev/irqreg.h>
; #include <dev/vs1001k.h>
; 
; #include <sys/bankmem.h>
; 
; #include <stddef.h> /* NULL definition */
; 
; /*!
;  * \addtogroup xgVs1001
;  */
; /*@{*/
; 
; 
; static volatile u_char vs_status = VS_STATUS_STOPPED;
; static volatile u_short vs_flush;
; 
; 
; 
; /*
;  * \brief Write a byte to the VS1001 data interface.
;  *
;  * The caller is responsible for checking the DREQ line. Also make sure,
;  * that decoder interrupts are disabled.
;  *
;  * \param b Byte to be shifted to the decoder's data interface.
;  */
; static INLINE void VsSdiPutByte(u_char b)
; {
L3:
; #ifdef VS_NOSPI
;     u_char mask = 0x80;
;     sbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
;     while (mask) {
;         if (b & mask)
;             sbi(VS_SI_PORT, VS_SI_BIT);
;         else
;             cbi(VS_SI_PORT, VS_SI_BIT);
; 
;         sbi(VS_SS_PORT, VS_SS_BIT);
;         mask >>= 1;
;         cbi(VS_SS_PORT, VS_SS_BIT);
;         cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
;     }
; #else
;     /* Wait for previous SPI transfer to finish. */
;     loop_until_bit_is_set(SPSR, SPIF);
L4:
	sbis 0xe,7
	rjmp L3
X0:
;     sbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
	sbi 0x18,5
;     outp(b, SPDR);
	out 0xf,R16
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     _NOP();
	nop
;     cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
	cbi 0x18,5
L2:
	.dbline 0 ; func end
	ret
;            try -> R20,R21
;            len -> R10,R11
;           data -> R22,R23
	.even
_VsSdiWrite:
	xcall push_xgsetF00C
	movw R10,R18
	movw R22,R16
; #endif
; }
; 
; /*!
;  * \brief Write a specified number of bytes to the VS1001 data interface.
;  *
;  * This function will check the DREQ line. Decoder interrupts must have 
;  * been disabled before calling this function.
;  */
; static int VsSdiWrite(CONST u_char * data, u_short len)
; {
;     u_short try = 5000;
	ldi R20,5000
	ldi R21,19
	xjmp L8
L10:
; 
;     while (len--) {
;         while (try-- && bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT));
L11:
	movw R2,R20
	subi R20,1
	sbci R21,0
	tst R2
	brne X1
	tst R3
	breq L13
X1:
	sbis 0x1,6
	rjmp L10
X2:
L13:
;         VsSdiPutByte(*data);
	movw R30,R22
	ldd R16,z+0
	xcall _VsSdiPutByte
;         data++;
	subi R22,255  ; offset = 1
	sbci R23,255
;     }
L8:
	movw R2,R10
	movw R24,R2
	sbiw R24,1
	movw R10,R24
	tst R2
	brne L11
	tst R3
	brne L11
X3:
;     return try ? 0 : -1;
	cpi R20,0
	cpc R20,R21
	breq L14
X4:
	clr R20
	clr R21
	xjmp L15
L14:
	ldi R20,-1
	ldi R21,-1
L15:
	movw R16,R20
L6:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;            try -> R20,R21
;            len -> R10,R11
;           data -> R22,R23
	.even
_VsSdiWrite_P:
	xcall push_xgsetF00C
	movw R10,R18
	movw R22,R16
; }
; 
; /*!
;  * \brief Write a specified number of bytes from program space to the 
;  *        VS1001 data interface.
;  *
;  * This function is similar to VsSdiWrite() except that the data is 
;  * located in program space.
;  */
; static int VsSdiWrite_P(PGM_P data, u_short len)
; {
;     u_short try = 5000;
	ldi R20,5000
	ldi R21,19
	xjmp L18
L20:
; 
;     while (len--) {
;         while (try-- && bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT));
L21:
	movw R2,R20
	subi R20,1
	sbci R21,0
	tst R2
	brne X5
	tst R3
	breq L23
X5:
	sbis 0x1,6
	rjmp L20
X6:
L23:
;         VsSdiPutByte(PRG_RDB(data));
	movw R30,R22
	lpm R16,Z
	xcall _VsSdiPutByte
;         data++;
	subi R22,255  ; offset = 1
	sbci R23,255
;     }
L18:
	movw R2,R10
	movw R24,R2
	sbiw R24,1
	movw R10,R24
	tst R2
	brne L21
	tst R3
	brne L21
X7:
;     return try ? 0 : -1;
	cpi R20,0
	cpc R20,R21
	breq L24
X8:
	clr R20
	clr R21
	xjmp L25
L24:
	ldi R20,-1
	ldi R21,-1
L25:
	movw R16,R20
L16:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;           mask -> R20
;           data -> R16
	.even
_VsSciPutByte:
	st -y,R20
	st -y,R21
; }
; 
; /*!
;  * \brief Write a byte to the serial control interface.
;  *
;  * Decoder interrupts must have been disabled and the decoder chip
;  * select must have been enabled before calling this function.
;  */
; static INLINE void VsSciPutByte(u_char data)
; {
;     u_char mask = 0x80;
	ldi R20,128
	xjmp L28
L27:
; 
;     /*
;      * Loop until all 8 bits are processed.
;      */
;     while (mask) {
; 
;         /* Set data line. */
;         if (data & mask)
	mov R2,R16
	and R2,R20
	breq L30
X9:
;             sbi(VS_SI_PORT, VS_SI_BIT);
	sbi 0x18,2
	xjmp L31
L30:
;         else
;             cbi(VS_SI_PORT, VS_SI_BIT);
	cbi 0x18,2
L31:
; 
;         /* Toggle clock and shift mask. */
;         sbi(VS_SCK_PORT, VS_SCK_BIT);
	sbi 0x18,0
;         mask >>= 1;
	lsr R20
;         cbi(VS_SCK_PORT, VS_SCK_BIT);
	cbi 0x18,0
;     }
L28:
	tst R20
	brne L27
X10:
L26:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;           data -> R16
;           mask -> R18
	.even
_VsSciGetByte:
; }
; 
; /*!
;  * \brief Read a byte from the serial control interface.
;  *
;  * Decoder interrupts must have been disabled and the decoder chip
;  * select must have been enabled before calling this function.
;  */
; static INLINE u_char VsSciGetByte(void)
; {
;     u_char mask = 0x80;
	ldi R18,128
;     u_char data = 0;
	clr R16
	xjmp L34
L33:
; 
;     /*
;      * Loop until all 8 bits are processed.
;      */
;     while (mask) {
;         /* Toggle clock and get the data. */
;         sbi(VS_SCK_PORT, VS_SCK_BIT);
	sbi 0x18,0
;         if (bit_is_set(VS_SO_PIN, VS_SO_BIT))
	sbis 0x16,3
	rjmp L36
X11:
;             data |= mask;
	or R16,R18
L36:
;         mask >>= 1;
	lsr R18
;         cbi(VS_SCK_PORT, VS_SCK_BIT);
	cbi 0x18,0
;     }
L34:
	tst R18
	brne L33
X12:
;     return data;
L32:
	.dbline 0 ; func end
	ret
;           data -> R10,R11
;            reg -> R12
	.even
_VsRegWrite:
	xcall push_xgset003C
	movw R10,R18
	mov R12,R16
; }
; 
; /*!
;  * \brief Write to a decoder register.
;  *
;  * Decoder interrupts must have been disabled before calling this function.
;  */
; static void VsRegWrite(u_char reg, u_short data)
; {
;     /* Select chip. */
;     cbi(VS_XCS_PORT, VS_XCS_BIT);
	cbi 0x18,4
; 
; #ifndef VS_NOSPI
;     /* Disable SPI */
;     cbi(SPCR, SPE);
	cbi 0xd,6
; #endif
; 
;     VsSciPutByte(VS_OPCODE_WRITE);
	ldi R16,2
	xcall _VsSciPutByte
;     VsSciPutByte(reg);
	mov R16,R12
	xcall _VsSciPutByte
;     VsSciPutByte((u_char) (data >> 8));
	movw R16,R10
	mov R16,R17
	clr R17
	xcall _VsSciPutByte
;     VsSciPutByte((u_char) data);
	mov R16,R10
	xcall _VsSciPutByte
; 
; #ifndef VS_NOSPI
;     /* Re-enable SPI. Hint given by Jesper Hansen. */
;     outp(BV(MSTR) | BV(SPE), SPCR);
	ldi R24,80
	out 0xd,R24
;     outp(inp(SPSR), SPSR);
	in R2,0xe
	out 0xe,R2
; #endif
; 
;     /* Deselect chip. */
;     sbi(VS_XCS_PORT, VS_XCS_BIT);
	sbi 0x18,4
L38:
	.dbline 0 ; func end
	xjmp pop_xgset003C
;           data -> R10,R11
;            reg -> R10
	.even
_VsRegRead:
	st -y,R10
	st -y,R11
	mov R10,R16
; }
; 
; /*
;  * \brief Read from a register.
;  *
;  * Decoder interrupts must have been disabled before calling this function.
;  * 
;  * \return Register contents.
;  */
; static u_short VsRegRead(u_char reg)
; {
;     u_short data;
; 
;     /* Disable interrupts and select chip. */
;     cbi(VS_XCS_PORT, VS_XCS_BIT);
	cbi 0x18,4
; 
; #ifndef VS_NOSPI
;     /* Disable SPI. */
;     cbi(SPCR, SPE);
	cbi 0xd,6
; #endif
; 
;     VsSciPutByte(VS_OPCODE_READ);
	ldi R16,3
	xcall _VsSciPutByte
;     VsSciPutByte(reg);
	mov R16,R10
	xcall _VsSciPutByte
;     data = VsSciGetByte() << 8;
	xcall _VsSciGetByte
	mov R10,R16
	clr R11
	mov R11,R10
	clr R10
;     data |= VsSciGetByte();
	xcall _VsSciGetByte
	mov R2,R16
	clr R3
	or R10,R2
	or R11,R3
; 
; #ifndef VS_NOSPI
;     /* Re-enable SPI. Changed due to a hint by Jesper. */
;     outp(BV(MSTR) | BV(SPE), SPCR);
	ldi R24,80
	out 0xd,R24
;     outp(inp(SPSR), SPSR);
	in R2,0xe
	out 0xe,R2
; #endif
; 
;     /* Deselect chip and enable interrupts. */
;     sbi(VS_XCS_PORT, VS_XCS_BIT);
	sbi 0x18,4
; 
;     return data;
	movw R16,R10
L39:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	ret
;             rc -> R10
;         enable -> R16
	.even
_VsPlayerInterrupts::
	xcall push_xgset300C
; }
; 
; /*!
;  * \brief Enable or disable player interrupts.
;  *
;  * This routine is typically used by applications when dealing with 
;  * unprotected buffers.
;  *
;  * \param enable Disables interrupts when zero. Otherwise interrupts
;  *               are enabled.
;  *
;  * \return Zero if interrupts were disabled before this call.
;  */
; u_char VsPlayerInterrupts(u_char enable)
; {
;     u_char rc;
; 
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;     rc = (inb(EIMSK) & _BV(VS_DREQ_BIT)) != 0;
	in R2,0x39
	sbrs R2,6
	rjmp L41
X13:
	ldi R20,1
	ldi R21,0
	xjmp L42
L41:
	clr R20
	clr R21
L42:
	mov R10,R20
;     if(enable)
	tst R16
	breq L43
X14:
;         sbi(EIMSK, VS_DREQ_BIT);
	in R24,0x39
	ori R24,64
	out 0x39,R24
	xjmp L44
L43:
;     else
;         cbi(EIMSK, VS_DREQ_BIT);
	in R24,0x39
	andi R24,191
	out 0x39,R24
L44:
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
; 
;     return rc;
	mov R16,R10
L40:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;      available -> y+0
;             bp -> R12,R13
;       consumed -> R20,R21
;            ief -> R10
;              j -> R22
;            arg -> R10,R11
	.even
_VsPlayerFeed:
	xcall push_xgsetF0FC
	movw R10,R16
	sbiw R28,2
; }
; 
; /*
;  * \brief Feed the decoder with data.
;  *
;  * This function serves two purposes: 
;  * - It is called by VsPlayerKick() to initially fill the decoder buffer.
;  * - It is used as an interrupt handler for the decoder.
;  */
; static void VsPlayerFeed(void *arg)
; {
;     u_char ief;
;     u_char j = 32;
	ldi R22,32
; 
;     /* Cancel interrupt if not running. */
;     if (vs_status != VS_STATUS_RUNNING || bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT))
	lds R24,_vs_status
	cpi R24,1
	brne L48
X15:
	sbic 0x1,6
	rjmp L46
X16:
L48:
;         return;
	xjmp L45
L46:
; 
;     /*
;      * We are hanging around here some time and may block other important
;      * interrupts. Disable decoder interrupts and enable global interrupts.
;      */
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R10,R16
;     sei();
	sei
; 
;     /* 
;      * Flush the internal VS buffer. 
;      */
;     if(vs_flush) {
	lds R2,_vs_flush
	lds R3,_vs_flush+1
	tst R2
	brne X17
	tst R3
	breq L49
X17:
L51:
;         do {
;             VsSdiPutByte(0);
	clr R16
	xcall _VsSdiPutByte
;             if (--vs_flush == 0) {
	lds R24,_vs_flush
	lds R25,_vs_flush+1
	sbiw R24,1
	sts _vs_flush+1,R25
	sts _vs_flush,R24
	cpi R24,0
	cpc R24,R25
	brne L54
X18:
;                 /* Decoder internal buffer is flushed. */
;                 vs_status = VS_STATUS_EMPTY;
	ldi R24,4
	sts _vs_status,R24
;                 break;
	xjmp L50
L54:
;             }
;             /* Allow 32 bytes to be sent as long as DREQ is set, This includes
;                the one in progress. */
;             if (bit_is_set(VS_DREQ_PIN, VS_DREQ_BIT))
	sbis 0x1,6
	rjmp L56
X19:
;                 j = 32;
	ldi R22,32
L56:
;         } while(j--);
L52:
	mov R20,R22
	clr R21
	subi R22,1
	tst R20
	brne L51
X20:
;     }
	xjmp L50
L49:
; 
;     /* 
;      * Feed the decoder until its buffer is full or we ran out of data.
;      */
;     else {
;         char *bp = 0;
	clr R12
	clr R13
;         size_t consumed = 0;
	clr R20
	clr R21
;         size_t available = 0;
	clr R2
	clr R3
	std y+1,R3
	std y+0,R2
L58:
;         do {
;             if(consumed >= available) {
	ldd R2,y+0
	ldd R3,y+1
	cp R20,R2
	cpc R21,R3
	brlo L61
X21:
;                 /* Commit previously consumed bytes. */
;                 if(consumed) {
	cpi R20,0
	cpc R20,R21
	breq L63
X22:
;                     NutSegBufReadCommit(consumed);
	movw R16,R20
	xcall _NutSegBufReadCommit
;                     consumed = 0;
	clr R20
	clr R21
;                 }
L63:
;                 /* All bytes consumed, request new. */
;                 bp = NutSegBufReadRequest(&available);
	movw R16,R28
	xcall _NutSegBufReadRequest
	movw R14,R16
	movw R12,R14
;                 if(available == 0) {
	ldd R2,y+0
	ldd R3,y+1
	tst R2
	brne L65
	tst R3
	brne L65
X23:
;                     /* End of stream. */
;                     vs_status = VS_STATUS_EOF;
	ldi R24,2
	sts _vs_status,R24
;                     break;
	xjmp L60
L65:
;                 }
;             }
L61:
;             /* We have some data in the buffer, feed it. */
;             VsSdiPutByte(*bp);
	movw R30,R12
	ldd R16,z+0
	xcall _VsSdiPutByte
;             bp++;
	movw R24,R12
	adiw R24,1
	movw R12,R24
;             consumed++;
	subi R20,255  ; offset = 1
	sbci R21,255
; 
;             /* Allow 32 bytes to be sent as long as DREQ is set, This includes
;                the one in progress. */
;             if (bit_is_set(VS_DREQ_PIN, VS_DREQ_BIT))
	sbis 0x1,6
	rjmp L67
X24:
;                 j = 32;
	ldi R22,32
L67:
;         } while(j--);
L59:
	mov R14,R22
	clr R15
	subi R22,1
	tst R14
	brne L58
X25:
L60:
; 
;         /* Finally re-enable the producer buffer. */
;         NutSegBufReadLast(consumed);
	movw R16,R20
	xcall _NutSegBufReadLast
;     }
L50:
;     VsPlayerInterrupts(ief);
	mov R16,R10
	xcall _VsPlayerInterrupts
L45:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF0FC
	.even
_VsPlayerKick::
; }
; 
; /*!
;  * \brief Start playback.
;  *
;  * This routine will send the first MP3 data bytes to the
;  * decoder, until it is completely filled. The data buffer
;  * should have been filled before calling this routine.
;  *
;  * Decoder interrupts will be enabled.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerKick(void)
; {
;     /*
;      * Start feeding the decoder with data.
;      */
;     VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
;     vs_status = VS_STATUS_RUNNING;
	ldi R24,1
	sts _vs_status,R24
;     VsPlayerFeed(NULL);
	clr R16
	clr R17
	xcall _VsPlayerFeed
;     VsPlayerInterrupts(1);
	ldi R16,1
	xcall _VsPlayerInterrupts
; 
;     return 0;
	clr R16
	clr R17
L69:
	.dbline 0 ; func end
	ret
;            ief -> R10
	.even
_VsPlayerStop::
	st -y,R10
	st -y,R11
; }
; 
; /*!
;  * \brief Stops the playback.
;  *
;  * This routine will stops the MP3 playback, VsPlayerKick() may be used 
;  * to resume the playback.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerStop(void)
; {
;     u_char ief;
; 
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R10,R16
;     /* Check whether we need to stop at all to not overwrite other than running status */
;     if (vs_status == VS_STATUS_RUNNING)
	lds R24,_vs_status
	cpi R24,1
	brne L71
X26:
;         vs_status = VS_STATUS_STOPPED;
	clr R2
	sts _vs_status,R2
L71:
;     VsPlayerInterrupts(ief);
	mov R16,R10
	xcall _VsPlayerInterrupts
; 
;     return 0;
	clr R16
	clr R17
L70:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	ret
	.even
_VsPlayerFlush::
; }
; 
; 
; /*!
;  * \brief Sets up decoder internal buffer flushing.
;  *
;  * This routine will set up internal VS buffer flushing,
;  * unless the buffer is already empty and starts the playback
;  * if necessary. The internal VS buffer is flushed in VsPlayerFeed()
;  * at the end of the stream.
;  *
;  * Decoder interrupts will be enabled.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerFlush(void)
; {
;     VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
;     /* Set up fluhing unless both buffers are empty. */
;     if (vs_status != VS_STATUS_EMPTY || NutSegBufUsed()) {
	lds R24,_vs_status
	cpi R24,4
	brne L76
X27:
	xcall _NutSegBufUsed
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	cp R16,R24
	cpc R17,R25
	cpc R18,R26
	cpc R19,R27
	breq L74
X28:
L76:
;         if (vs_flush == 0)
	lds R2,_vs_flush
	lds R3,_vs_flush+1
	tst R2
	brne L77
	tst R3
	brne L77
X29:
;             vs_flush = VS_FLUSH_BYTES;
	ldi R24,2048
	ldi R25,8
	sts _vs_flush+1,R25
	sts _vs_flush,R24
L77:
;         /* start the playback if necessary */
;         if (vs_status != VS_STATUS_RUNNING)
	lds R24,_vs_status
	cpi R24,1
	breq L79
X30:
;             VsPlayerKick();
	xcall _VsPlayerKick
L79:
;     }
L74:
;     VsPlayerInterrupts(1);
	ldi R16,1
	xcall _VsPlayerInterrupts
;     return 0;
	clr R16
	clr R17
L73:
	.dbline 0 ; func end
	ret
;          dummy -> R10
	.even
_VsPlayerInit::
	st -y,R10
	st -y,R11
	sbiw R28,2
; }
; 
; 
; /*!
;  * \brief Initialize the VS1001 hardware interface.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerInit(void)
; {
;     /* Disable decoder interrupts. */
;     VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
; 
;     /* Keep decoder in reset state. */
;     cbi(VS_RESET_PORT, VS_RESET_BIT);
	cbi 0x18,7
;     sbi(VS_RESET_DDR, VS_RESET_BIT);
	sbi 0x17,7
; 
;     /* Set BSYNC output low. */
;     cbi(VS_BSYNC_PORT, VS_BSYNC_BIT);
	cbi 0x18,5
;     sbi(VS_BSYNC_DDR, VS_BSYNC_BIT);
	sbi 0x17,5
; 
;     /* Set MP3 chip select output low. */
;     sbi(VS_XCS_PORT, VS_XCS_BIT);
	sbi 0x18,4
;     sbi(VS_XCS_DDR, VS_XCS_BIT);
	sbi 0x17,4
; 
;     /* Set DREQ input with pullup. */
;     sbi(VS_DREQ_PORT, VS_DREQ_BIT);
	sbi 0x3,6
;     cbi(VS_DREQ_DDR, VS_DREQ_BIT);
	cbi 0x2,6
; 
;     /* Init SPI Port. */
;     sbi(VS_SI_DDR, VS_SI_BIT);
	sbi 0x17,2
;     sbi(VS_SS_DDR, VS_SS_BIT);
	sbi 0x17,1
;     cbi(VS_SO_DDR, VS_SO_BIT);
	cbi 0x17,3
; 
;     /* Set SCK output low. */
;     cbi(VS_SCK_PORT, VS_SCK_BIT);
	cbi 0x18,0
;     sbi(VS_SCK_DDR, VS_SCK_BIT);
	sbi 0x17,0
; 
; #ifndef VS_NOSPI
;     {
;         u_char dummy;           /* Required by some compilers. */
; 
;         /* 
;          * Init SPI mode to no interrupts, enabled, MSB first, master mode, 
;          * rising clock and fosc/4 clock speed. Send an initial zero byte to 
;          * make sure SPIF is set. Note, that the decoder reset line is still
;          * active.
;          */
;         outp(BV(MSTR) | BV(SPE), SPCR);
	ldi R24,80
	out 0xd,R24
;         dummy = inp(SPSR);
	in R10,0xe
;         outp(0, SPDR);
	clr R2
	out 0xf,R2
;     }
; #endif
; 
;     /* Rising edge will generate interrupts. */
;     sbi(EICR, 5);
	in R24,0x3a
	ori R24,32
	out 0x3a,R24
;     sbi(EICR, 4);
	in R24,0x3a
	ori R24,16
	out 0x3a,R24
; 
;     /* Register the interrupt routine */
;     NutRegisterIrqHandler(&sig_INTERRUPT6, VsPlayerFeed, NULL);
	clr R3
	std y+1,R3
	std y+0,R2
	ldi R18,<PL_VsPlayerFeed
	ldi R19,>PL_VsPlayerFeed
	ldi R16,<_sig_INTERRUPT6
	ldi R17,>_sig_INTERRUPT6
	xcall _NutRegisterIrqHandler
; 
;     /* Release decoder reset line. */
;     sbi(VS_RESET_PORT, VS_RESET_BIT);
	sbi 0x18,7
;     NutDelay(4);
	ldi R16,4
	xcall _NutDelay
; 
;     /* Force frequency change (see datasheet). */
;     VsRegWrite(VS_CLOCKF_REG, 0x9800);
	ldi R18,38912
	ldi R19,152
	ldi R16,3
	xcall _VsRegWrite
;     VsRegWrite(VS_INT_FCTLH_REG, 0x8008);
	ldi R18,32776
	ldi R19,128
	ldi R16,2
	xcall _VsRegWrite
;     NutDelay(200);
	ldi R16,200
	xcall _NutDelay
; 
;     /* Clear any spurious interrupt. */
;     outp(BV(VS_DREQ_BIT), EIFR);
	ldi R24,64
	out 0x38,R24
; 
;     return 0;
	clr R16
	clr R17
L81:
	.dbline 0 ; func end
	adiw R28,2
	ld R11,y+
	ld R10,y+
	ret
;           mode -> R20,R21
	.even
_VsPlayerReset::
	st -y,R20
	st -y,R21
	movw R20,R16
; }
; 
; /*!
;  * \brief Software reset the decoder.
;  *
;  * This function is typically called after VsPlayerInit() and at the end
;  * of each track.
;  *
;  * \param mode Any of the following flags may be or'ed
;  * - VS_SM_DIFF Left channel inverted.
;  * - VS_SM_FFWD Fast forward.
;  * - VS_SM_RESET Force hardware reset.
;  * - VS_SM_PDOWN Switch to power down mode.
;  * - VS_SM_BASS Bass/treble enhancer.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerReset(u_short mode)
; {
;     /* Disable decoder interrupts and feeding. */
;     VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
;     vs_status = VS_STATUS_STOPPED;
	clr R2
	sts _vs_status,R2
; 
;     /* Software reset, set modes of decoder. */
;     VsRegWrite(VS_MODE_REG, VS_SM_RESET | mode);
	movw R18,R20
	ori R18,4
	clr R16
	xcall _VsRegWrite
;     NutDelay(2);
	ldi R16,2
	xcall _NutDelay
; 
;     /*
;      * Check for correct reset.
;      */
;     if ((mode & VS_SM_RESET) != 0 || bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT)) {
	sbrc R20,2
	rjmp L85
X31:
	sbic 0x1,6
	rjmp L83
X32:
L85:
;         /* If not succeeded, give it one more chance and try hw reset,
;            HW reset may also be forced by VS_SM_RESET mode bit. */
;         cbi(VS_RESET_PORT, VS_RESET_BIT);
	cbi 0x18,7
;         _NOP();
	nop
;         sbi(VS_RESET_PORT, VS_RESET_BIT);
	sbi 0x18,7
;         NutDelay(4);
	ldi R16,4
	xcall _NutDelay
; 
;         /* Set the requested modes. */
;         VsRegWrite(VS_MODE_REG, VS_SM_RESET | mode);
	movw R18,R20
	ori R18,4
	clr R16
	xcall _VsRegWrite
;         NutDelay(2);
	ldi R16,2
	xcall _NutDelay
;         if (bit_is_clear(VS_DREQ_PIN, VS_DREQ_BIT))
	sbic 0x1,6
	rjmp L86
X33:
;             return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L82
L86:
;     }
L83:
; 
;     /* Force frequency change (see datasheet). */
;     VsRegWrite(VS_CLOCKF_REG, 0x9800);
	ldi R18,38912
	ldi R19,152
	ldi R16,3
	xcall _VsRegWrite
;     VsRegWrite(VS_INT_FCTLH_REG, 0x8008);
	ldi R18,32776
	ldi R19,128
	ldi R16,2
	xcall _VsRegWrite
;     NutDelay(2);
	ldi R16,2
	xcall _NutDelay
; 
;     /* Clear any spurious interrupts. */
;     outp(BV(VS_DREQ_BIT), EIFR);
	ldi R24,64
	out 0x38,R24
; 
;     return 0;
	clr R16
	clr R17
L82:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;            ief -> R10
;           mode -> R12,R13
	.even
_VsPlayerSetMode::
	xcall push_xgset003C
	movw R12,R16
; }
; 
; /*!
;  * \brief Set mode register of the decoder.
;  *
;  * \param mode Any of the following flags may be or'ed
;  * - VS_SM_DIFF Left channel inverted.
;  * - VS_SM_FFWD Fast forward.
;  * - VS_SM_RESET Software reset.
;  * - VS_SM_PDOWN Switch to power down mode.
;  * - VS_SM_BASS Bass/treble enhancer.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsPlayerSetMode(u_short mode)
; {
;     u_char ief;
;     
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R10,R16
;     VsRegWrite(VS_MODE_REG, mode);
	movw R18,R12
	clr R16
	xcall _VsRegWrite
;     VsPlayerInterrupts(ief);
	mov R16,R10
	xcall _VsPlayerInterrupts
; 
;     return 0;
	clr R16
	clr R17
L88:
	.dbline 0 ; func end
	xjmp pop_xgset003C
;            ief -> R12
;             rc -> R10,R11
	.even
_VsPlayTime::
	xcall push_xgset003C
; }
; 
; /*!
;  * \brief Returns play time since last reset.
;  *
;  * \return Play time since reset in seconds
;  */
; u_short VsPlayTime(void)
; {
;     u_short rc;
;     u_char ief;
; 
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R12,R16
;     rc = VsRegRead(VS_DECODE_TIME_REG);
	ldi R16,4
	xcall _VsRegRead
	movw R10,R16
;     VsPlayerInterrupts(ief);
	mov R16,R12
	xcall _VsPlayerInterrupts
; 
;     return rc;
	movw R16,R10
L89:
	.dbline 0 ; func end
	xjmp pop_xgset003C
	.even
_VsGetStatus::
; }
; 
; /*!
;  * \brief Returns status of the player.
;  *
;  * \return Any of the following value:
;  * - VS_STATUS_STOPPED Player is ready to be started by VsPlayerKick().
;  * - VS_STATUS_RUNNING Player is running.
;  * - VS_STATUS_EOF Player has reached the end of a stream after VsPlayerFlush() has been called.
;  * - VS_STATUS_EMPTY Player runs out of data. VsPlayerKick() will restart it.
;  */
; u_char VsGetStatus(void)
; {
;     return vs_status;
	lds R16,_vs_status
L90:
	.dbline 0 ; func end
	ret
	.area lit(rom, con, rel)
L92:
	.byte 'M,234
	.byte 'm,'T
	.byte 0,0
	.byte 0,0
	.area text(rom, con, rel)
;            ief -> R12
;             rc -> R10,R11
	.even
_VsMemoryTest::
	xcall push_xgset003C
; }
; 
; #ifdef __GNUC__
; 
; /*!
;  * \brief Query MP3 stream header information.
;  *
;  * \param vshi Pointer to VS_HEADERINFO structure.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsGetHeaderInfo(VS_HEADERINFO * vshi)
; {
;     u_short *usp = (u_short *) vshi;
;     u_char ief;
; 
;     ief = VsPlayerInterrupts(0);
;     *usp = VsRegRead(VS_HDAT1_REG);
;     *++usp = VsRegRead(VS_HDAT0_REG);
;     VsPlayerInterrupts(ief);
; 
;     return 0;
; }
; #endif
; 
; /*!
;  * \brief Initialize decoder memory test and return result.
;  *
;  * \return Memory test result.
;  * - Bit 0: Good X ROM
;  * - Bit 1: Good Y ROM (high)
;  * - Bit 2: Good Y ROM (low)
;  * - Bit 3: Good Y RAM
;  * - Bit 4: Good X RAM
;  * - Bit 5: Good Instruction RAM (high)
;  * - Bit 6: Good Instruction RAM (low)
;  */
; u_short VsMemoryTest(void)
; {
;     u_short rc;
;     u_char ief;
;     static prog_char mtcmd[] = { 0x4D, 0xEA, 0x6D, 0x54, 0x00, 0x00, 0x00, 0x00 };
; 
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R12,R16
;     VsSdiWrite_P(mtcmd, sizeof(mtcmd));
	ldi R18,8
	ldi R19,0
	ldi R16,<L92
	ldi R17,>L92
	xcall _VsSdiWrite_P
;     NutDelay(40);
	ldi R16,40
	xcall _NutDelay
;     rc = VsRegRead(VS_HDAT0_REG);
	ldi R16,8
	xcall _VsRegRead
	movw R10,R16
;     VsPlayerInterrupts(ief);
	mov R16,R12
	xcall _VsPlayerInterrupts
; 
;     return rc;
	movw R16,R10
L91:
	.dbline 0 ; func end
	xjmp pop_xgset003C
;            ief -> R10
;          right -> R14
;           left -> R12
	.even
_VsSetVolume::
	xcall push_xgset00FC
	mov R14,R18
	mov R12,R16
; }
; 
; /*!
;  * \brief Set volume.
;  *
;  * \param left  Left channel volume.
;  * \param right Right channel volume.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsSetVolume(u_char left, u_char right)
; {
;     u_char ief;
; 
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R10,R16
;     VsRegWrite(VS_VOL_REG, (((u_short) left) << 8) | (u_short) right);
	mov R2,R14
	clr R3
	mov R18,R12
	clr R19
	mov R19,R18
	clr R18
	or R18,R2
	or R19,R3
	ldi R16,11
	xcall _VsRegWrite
;     VsPlayerInterrupts(ief);
	mov R16,R10
	xcall _VsPlayerInterrupts
; 
;     return 0;
	clr R16
	clr R17
L93:
	.dbline 0 ; func end
	xjmp pop_xgset00FC
	.area lit(rom, con, rel)
L95:
	.byte 'S,239
	.byte 'n
L96:
	.byte 'E,'x
	.byte 'i,'t
L97:
	.byte 0,0
	.byte 0,0
	.area text(rom, con, rel)
;            ief -> R10
;             ms -> y+4
;           fsin -> y+2
	.even
_VsBeep::
	xcall push_arg4
	st -y,R10
	st -y,R11
; }
; 
; /*!
;  * \brief Sine wave beep.
;  *
;  * \param fsin Frequency.
;  * \param ms   Duration.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; int VsBeep(u_char fsin, u_char ms)
; {
;     u_char ief;
;     static prog_char on[] = { 0x53, 0xEF, 0x6E };
;     static prog_char off[] = { 0x45, 0x78, 0x69, 0x74 };
;     static prog_char end[] = { 0x00, 0x00, 0x00, 0x00 };
; 
;     /* Disable decoder interrupts. */
;     ief = VsPlayerInterrupts(0);
	clr R16
	xcall _VsPlayerInterrupts
	mov R10,R16
; 
;     fsin = 56 + (fsin & 7) * 9;
	ldd R24,y+2
	andi R24,7
	ldi R25,9
	mul R25,R24
	mov R24,R0
	subi R24,200    ; addi 56
	std y+2,R24
;     VsSdiWrite_P(on, sizeof(on));
	ldi R18,3
	ldi R19,0
	ldi R16,<L95
	ldi R17,>L95
	xcall _VsSdiWrite_P
;     VsSdiWrite(&fsin, 1);
	ldi R18,1
	ldi R19,0
	movw R16,R28
	subi R16,254  ; offset = 2
	sbci R17,255
	xcall _VsSdiWrite
;     VsSdiWrite_P(end, sizeof(end));
	ldi R18,4
	ldi R19,0
	ldi R16,<L97
	ldi R17,>L97
	xcall _VsSdiWrite_P
;     NutDelay(ms);
	ldd R16,y+4
	xcall _NutDelay
;     VsSdiWrite_P(off, sizeof(off));
	ldi R18,4
	ldi R19,0
	ldi R16,<L96
	ldi R17,>L96
	xcall _VsSdiWrite_P
;     VsSdiWrite_P(end, sizeof(end));
	ldi R18,4
	ldi R19,0
	ldi R16,<L97
	ldi R17,>L97
	xcall _VsSdiWrite_P
; 
;     /* Enable decoder interrupts. */
;     VsPlayerInterrupts(ief);
	mov R16,R10
	xcall _VsPlayerInterrupts
; 
;     return 0;
	clr R16
	clr R17
L94:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	adiw R28,4
	ret
	.area bss(ram, con, rel)
_vs_flush:
	.blkb 2
; }
; 
; 
; /*@}*/
	.area func_lit
PL_VsPlayerFeed:	.word `_VsPlayerFeed
