	.module pnutfs.c
	.area text(rom, con, rel)
;            blk -> R0,R1
	.even
_BankSelect::
; /*
;  * Copyright (C) 2004-2005 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  */
; 
; /*!
;  * \file fs/pnutfs.c
;  * \brief Peanut File System.
;  *
;  * \verbatim
;  *
;  * $Log: pnutfs.c,v $
;  * Revision 1.7  2005/09/08 10:12:44  olereinhardt
;  * Added #ifdef statement in NutSegBufEnable to avoid compiler warning
;  * if no banked mem is used.
;  *
;  * Revision 1.6  2005/09/07 16:23:41  christianwelzel
;  * Added support for MMnet02. Bankswitching is now handled in bankmem.h
;  *
;  * Revision 1.5  2005/08/02 17:46:47  haraldkipp
;  * Major API documentation update.
;  *
;  * Revision 1.4  2005/05/16 08:33:59  haraldkipp
;  * Added banking support for Arthernet.
;  *
;  * Revision 1.3  2005/02/21 11:10:21  olereinhardt
;  * Changed deckaration of the "root" variable to compile with unix emulation
;  *
;  * Revision 1.2  2005/02/07 18:57:47  haraldkipp
;  * ICCAVR compile errors fixed
;  *
;  * Revision 1.1  2005/02/05 20:35:21  haraldkipp
;  * Peanut added
;  *
;  * \endverbatim
;  */
; 
; #include <compiler.h>
; 
; #include <stdlib.h>
; #include <string.h>
; #include <time.h>
; #include <fcntl.h>
; #include <dirent.h>
; #include <errno.h>
; 
; #include <sys/stat.h>
; #include <sys/file.h>
; #include <sys/bankmem.h>
; 
; #include <fs/fs.h>
; #include <dev/pnut.h>
; 
; /*!
;  * \addtogroup xgPNut
;  */
; /*@{*/
; 
; /*!
;  * \name Peanut File System Configuration
;  *
;  * The Nut/OS Configurator may be used to override the default values.
;  */
; /*@{*/
; 
; /*! \brief Size of a filesystem block. 
;  *
;  * \showinitializer
;  */
; #ifndef PNUT_BLOCK_SIZE
; #define PNUT_BLOCK_SIZE     512
; #endif
; 
; /*! 
;  * \brief Size of a directory entry.
;  *
;  * 
;  * \showinitializer
;  */
; #ifndef PNUT_DIRENT_SIZE
; #define PNUT_DIRENT_SIZE  32
; #endif
; 
; /*! 
;  * \brief Maximum number of blocks per node.
;  *
;  * Peanut supports only one node per file. Thus, this number
;  * multiplied by PNUT_BLOCK_SIZE specifies the maximum file
;  * size.
;  *
;  * Changings this number will change the size of the node
;  * structure, which must fit into a single filesystem block.
;  * 
;  * \showinitializer
;  */
; #ifndef PNUT_BLOCKS_PER_NODE
; #define PNUT_BLOCKS_PER_NODE      250
; #endif
; 
; #ifndef PNUTBANK_COUNT
; #ifdef ARTHERNET1
; /* Default for Arthernet 1. */
; #define PNUTBANK_COUNT 15
; #elif MMNET02
; /* Default for MMnte02. */
; #define PNUTBANK_COUNT 6
; #else
; /* Default for Ethernet 2. */
; #define PNUTBANK_COUNT 30
; #endif
; #endif
; 
; /*@}*/
; 
; 
; #ifndef SEEK_SET
; #  define SEEK_SET        0     /* Seek from beginning of file.  */
; #  define SEEK_CUR        1     /* Seek from current position.  */
; #  define SEEK_END        2     /* Set file pointer to EOF plus "offset" */
; #endif
; 
; #define NODETYPE_REG        0
; #define NODETYPE_DIR        1
; 
; 
; typedef short PNUT_BLKNUM;
; 
; /*!
;  * \brief Node structure.
;  *
;  * The Peanut filesystem divides the total space into blocks. It
;  * distinguishes between unallocated blocks, data blocks and blocks
;  * containing node informations. Each allocated data block is 
;  * associated to a node block. Unallocated blocks are collected
;  * in a linked list.
;  *
;  * Peanut knows two types of nodes, file nodes and directory nodes.
;  * Each file node represents a file entry and the associated data 
;  * blocks contain the contents of this file. Each directory node
;  * represents a directory and the associated data blocks contain
;  * the entries in this directory. Each directory entry in use
;  * contains a link to to a node block.
;  *
;  * \note The node structure must fit into a block.
;  */
; typedef struct {
;     u_char node_type;           /*!< Type of this node, 0=file, 1=dir */
;     u_char node_refs;           /*!< Reference counter: Number of active open calls. */
;     /*!
;      * \brief Number of links.
;      *
;      * For file nodes this number keeps track of the number of directory 
;      * entries pointing to it. PNUT currently doesn't support link type
;      * entries, thus in file entries this is always 1.
;      *
;      * For directory nodes the number represents the number
;      * of subdirectories plus 2.
;      */
;     u_short node_links;
;     u_long node_size;           /*!< Total size of the data area. */
;     time_t node_mtime;          /*!< Time of last modification. */
;     PNUT_BLKNUM node_blocks[PNUT_BLOCKS_PER_NODE];      /*!< Allocated data blocks. */
; } PNUT_NODE;
; 
; /*!
;  * \brief Maximum length of a base name.
;  *
;  * This is a calculated value and depends on the definition of the
;  * block size and the size of a directory entry.
;  */
; #define PNUT_MAX_NAMELEN    (PNUT_DIRENT_SIZE - sizeof(PNUT_BLKNUM) - sizeof(u_char) - 1)
; 
; /*!
;  * \brief Directory entry structure.
;  *
;  * Each directory node contains a list of data blocks. Each of these
;  * data blocks contains a number of directory entries. Each directory
;  * entry points to a node block, which may be a file node or a
;  * directory node.
;  */
; typedef struct {
;     /*!
;      * \brief Node block index of this entry.
;      *
;      * This may point to a file or directory node.
;      */
;     PNUT_BLKNUM dir_node;
;     /*!
;      * \brief Allocation flag.
;      *
;      * Set to one if this entry is in use.
;      */
;     u_char dir_inuse;
;     /*!
;      * \brief Name of this entry.
;      *
;      * Contains the name ot the file or directory.
;      */
;     char dir_name[PNUT_MAX_NAMELEN + 1];
; } PNUT_DIRENTRY;
; 
; /*! 
;  * \brief Maximum size  of a file or directory.
;  *
;  * This is a calculated value and depends on the definition of the
;  * block size and the number of blocks per directory entry.
;  */
; #define PNUT_MAX_FILESIZE    (PNUT_BLOCKS_PER_NODE * PNUT_BLOCK_SIZE)
; 
; /*!
;  * \brief Directory search result information structure.
;  */
; typedef struct {
;     /*!
;      * \brief Node of the entry found. 
;      */
;     PNUT_BLKNUM fr_node;
; 
;     /*!
;      * \brief Parent node of the entry found.
;      */
;     PNUT_BLKNUM fr_pnode;
; 
;     /*! 
;      * \brief Last path component.
;      *
;      * Set if all parents found. 
;      */
;     CONST char *fr_name;
; } PNUT_FINDRESULT;
; 
; /*!
;  * \brief PNUT file descriptor type.
;  */
; typedef struct _PNUTFILE PNUTFILE;
; 
; /*!
;  * \brief PNUT file descriptor structure.
;  */
; struct _PNUTFILE {
;     PNUT_BLKNUM f_node;         /* Node of the file or directory. */
;     u_long f_pos;               /* Current file pointer position. */
;     u_int f_flag;               /* File mode. */
; };
; 
; /*! \brief Root node of the filesystem. */
; static PNUT_BLKNUM root;
; 
; /* -------------------- Banking hardware routines ------------------- */
; 
; /*! \brief Size of each memory bank. */
; #ifndef NUTBANK_SIZE
; #define NUTBANK_SIZE 16384
; #endif
; 
; /*! 
;  * \brief Total number of blocks on this device.
;  *
;  * This value is calulated by multiplying the number of memory banks 
;  * reserved for the file system by the number of blocks per bank.
;  *
;  * For example, if all 30 banks, which are available on Ethernut 2, are 
;  * reserved for PNUT and if the size of a block is 512 bytes, then 
;  * 960 blocks are available.
;  */
; #ifndef PNUT_TOTAL_BLOCKS
; #define PNUT_TOTAL_BLOCKS   (PNUTBANK_COUNT * (NUTBANK_SIZE / PNUT_BLOCK_SIZE))
; #endif
; 
; #define BLOCKS_PER_BANK  (NUTBANK_SIZE / PNUT_BLOCK_SIZE)
; 
; #ifndef NUTBANK_SR
; #define NUTBANK_SR 0xFF00
; #endif
; 
; #ifndef NUTBANK_START
; #define NUTBANK_START 0x8000
; #endif
; 
; #define NUTBANK_PTR ((char *)NUTBANK_START)
; 
; /*!
;  * \brief Make the bank visible, which contains the specified block.
;  *
;  * On Ethernut 2 we select a bank by writing the bank number to
;  * the bank register base address plus the bank number.
;  *
;  * For example, writing 0x12 to 0xFF12 selects bank 18 (0x12),
;  * assuming that 0xFF00 is the bank register base address.
;  *
;  * \param blk Block number to access.
;  */
; void BankSelect(PNUT_BLKNUM blk)
; {
; 
; // This is a hack to avoid compiler warning if no banking is enabled... 
; // But I don't like moving code to header files at all.. (Ole Reinhardt)
; #if NUTBANK_COUNT 
;     int bank = blk / BLOCKS_PER_BANK;
; #endif
;     // Bankswitching is now handled in bankmem.h
;     NutSegBufEnable(bank);
L6:
	.dbline 0 ; func end
	ret
;            blk -> R20,R21
	.even
_BankNodePointer::
	st -y,R20
	st -y,R21
	movw R20,R16
; }
; 
; /*!
;  * \brief Select specified bank and return pointer to block.
;  */
; PNUT_NODE *BankNodePointer(PNUT_BLKNUM blk)
; {
;     if (blk < 0) {
	cpi R20,0
	ldi R30,0
	cpc R21,R30
	brge L8
X0:
;         return NULL;
	clr R16
	clr R17
	xjmp L7
L8:
;     }
;     BankSelect(blk);
	movw R16,R20
	xcall _BankSelect
;     return (PNUT_NODE *) & NUTBANK_PTR[(blk % BLOCKS_PER_BANK) * PNUT_BLOCK_SIZE];
	ldi R18,32
	ldi R19,0
	movw R16,R20
	xcall mod16s
	movw R18,R16
	ldi R16,512
	ldi R17,2
	xcall empy16s
	ldi R24,32768
	ldi R25,128
	add R16,R24
	adc R17,R25
L7:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
	.area data(ram, con, rel)
_blockFreeList:
	.blkb 2
	.area idata
	.word -1
	.area data(ram, con, rel)
	.area text(rom, con, rel)
;        bankptr -> R10,R11
;             rc -> R20,R21
	.even
_PnutBlockAlloc:
	xcall push_xgset300C
	sbiw R28,2
; }
; 
; 
; /* ------------------------ Block routines ------------------------ */
; 
; /*
;  * \brief Index of the first free block.
;  *
;  * All free blocks are collected in a linked list. This global variable 
;  * contains the index of the first free block. Each free block contains 
;  * nothing but the index of the next free block, which is -1 in the last 
;  * member of this list.
;  *
;  * If this variable is set to -1, then there are no more free blocks 
;  * available. During file system initialization all available blocks are 
;  * added to this list.
;  */
; static PNUT_BLKNUM blockFreeList = -1;
; 
; /*!
;  * \brief Allocate a block.
;  *
;  * Removes the block in front of the list of free blocks.
;  *
;  * \return Index of the free block or -1 if none is available.
;  */
; static PNUT_BLKNUM PnutBlockAlloc(void)
; {
;     PNUT_BLKNUM rc = blockFreeList;
	lds R20,_blockFreeList
	lds R21,_blockFreeList+1
; 
;     if (rc >= 0) {
	cpi R20,0
	ldi R30,0
	cpc R21,R30
	brlt L11
X1:
;         PNUT_BLKNUM *bankptr = (PNUT_BLKNUM *) BankNodePointer(blockFreeList);
	movw R16,R20
	xcall _BankNodePointer
	movw R10,R16
;         blockFreeList = *bankptr;
	movw R30,R10
	ldd R2,z+0
	ldd R3,z+1
	sts _blockFreeList+1,R3
	sts _blockFreeList,R2
;         /* Clear the block contents. */
;         memset(bankptr, 0, PNUT_BLOCK_SIZE);
	ldi R24,512
	ldi R25,2
	std y+1,R25
	std y+0,R24
	clr R18
	clr R19
	xcall _memset
;     }
L11:
;     return rc;
	movw R16,R20
L10:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgset300C
;        bankptr -> R12,R13
;            blk -> R10,R11
	.even
_PnutBlockRelease:
	xcall push_xgset003C
	movw R10,R16
; }
; 
; /*!
;  * \brief Release a block.
;  *
;  * Inserts a given block in front of the list of free blocks.
;  *
;  * \param blk Index of the block to be released.
;  */
; static void PnutBlockRelease(PNUT_BLKNUM blk)
; {
;     PNUT_BLKNUM *bankptr = (PNUT_BLKNUM *) BankNodePointer(blk);
	movw R16,R10
	xcall _BankNodePointer
	movw R12,R16
; 
;     *bankptr = blockFreeList;
	lds R2,_blockFreeList
	lds R3,_blockFreeList+1
	movw R30,R12
	std z+1,R3
	std z+0,R2
;     blockFreeList = blk;
	sts _blockFreeList+1,R11
	sts _blockFreeList,R10
L13:
	.dbline 0 ; func end
	xjmp pop_xgset003C
;             np -> R22,R23
;              i -> R12,R13
;           node -> R20,R21
;           type -> R10
	.even
_PnutNodeAlloc:
	xcall push_xgsetF03C
	mov R10,R16
; }
; 
; /* ------------------------ Node routines ------------------------ */
; 
; /*!
;  * \brief Allocate a node.
;  *
;  * Creates a new node of a specified type.
;  *
;  * \param type Type of this node, either NODETYPE_REG or NODETYPE_DIR.
;  *
;  * \return Block index of this node or -1 if there are no more blocks.
;  */
; static PNUT_BLKNUM PnutNodeAlloc(u_char type)
; {
;     PNUT_BLKNUM node = PnutBlockAlloc();
	xcall _PnutBlockAlloc
	movw R20,R16
; 
;     if (node >= 0) {
	cpi R16,0
	ldi R30,0
	cpc R17,R30
	brlt L15
X2:
;         int i;
;         PNUT_NODE *np = BankNodePointer(node);
	xcall _BankNodePointer
	movw R22,R16
; 
;         /* Clear the data block list of this node. */
;         for (i = 0; i < PNUT_BLOCKS_PER_NODE; i++) {
	clr R12
	clr R13
L17:
;             np->node_blocks[i] = -1;
	ldi R16,2
	ldi R17,0
	movw R18,R12
	xcall empy16s
	movw R30,R16
	movw R24,R22
	adiw R24,12
	add R30,R24
	adc R31,R25
	ldi R24,-1
	ldi R25,-1
	std z+1,R25
	std z+0,R24
;         }
L18:
	movw R24,R12
	adiw R24,1
	movw R12,R24
	cpi R24,250
	ldi R30,0
	cpc R25,R30
	brlt L17
X3:
; 
;         /* Set the type and the last modification date. */
;         np->node_type = type;
	movw R30,R22
	std z+0,R10
;         np->node_mtime = time(0);
	clr R16
	clr R17
	xcall _time
	movw R30,R22
	std z+8,R16
	std z+9,R17
	std z+10,R18
	std z+11,R19
;     }
L15:
;     return node;
	movw R16,R20
L14:
	.dbline 0 ; func end
	xjmp pop_xgsetF03C
;              i -> R20,R21
;           node -> R10,R11
	.even
_PnutNodeDataRelease:
	xcall push_xgsetF00C
	movw R10,R16
; }
; 
; /*
;  * \brief Free all blocks of a specified node.
;  */
; static void PnutNodeDataRelease(PNUT_BLKNUM node)
; {
;     int i;
; 
;     for (i = 0; i < PNUT_BLOCKS_PER_NODE; i++) {
	clr R20
	clr R21
L22:
;         if (BankNodePointer(node)->node_blocks[i] >= 0) {
	movw R16,R10
	xcall _BankNodePointer
	movw R2,R16
	ldi R16,2
	ldi R17,0
	movw R18,R20
	xcall empy16s
	movw R30,R16
	movw R24,R2
	adiw R24,12
	add R30,R24
	adc R31,R25
	ldd R24,z+0
	ldd R25,z+1
	cpi R24,0
	ldi R30,0
	cpc R25,R30
	brlt L26
X4:
;             PnutBlockRelease(BankNodePointer(node)->node_blocks[i]);
	movw R16,R10
	xcall _BankNodePointer
	movw R2,R16
	ldi R16,2
	ldi R17,0
	movw R18,R20
	xcall empy16s
	movw R30,R16
	movw R24,R2
	adiw R24,12
	add R30,R24
	adc R31,R25
	ldd R16,z+0
	ldd R17,z+1
	xcall _PnutBlockRelease
;             BankNodePointer(node)->node_blocks[i] = -1;
	movw R16,R10
	xcall _BankNodePointer
	movw R22,R16
	ldi R16,2
	ldi R17,0
	movw R18,R20
	xcall empy16s
	movw R30,R16
	movw R24,R22
	adiw R24,12
	add R30,R24
	adc R31,R25
	ldi R24,-1
	ldi R25,-1
	std z+1,R25
	std z+0,R24
;         }
L26:
;     }
L23:
	subi R20,255  ; offset = 1
	sbci R21,255
	cpi R20,250
	ldi R30,0
	cpc R21,R30
	brge X6
	xjmp L22
X6:
X5:
;     BankNodePointer(node)->node_size = 0;
	movw R16,R10
	xcall _BankNodePointer
	movw R20,R16
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R20
	std z+4,R24
	std z+5,R25
	std z+6,R26
	std z+7,R27
;     BankNodePointer(node)->node_mtime = time(0);
	movw R16,R10
	xcall _BankNodePointer
	movw R20,R16
	clr R16
	clr R17
	xcall _time
	movw R30,R20
	std z+8,R16
	std z+9,R17
	std z+10,R18
	std z+11,R19
L21:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;           node -> R10,R11
	.even
_PnutNodeRelease:
	st -y,R10
	st -y,R11
	movw R10,R16
; }
; 
; /*!
;  * \brief Release a node and all its data blocks.
;  */
; static void PnutNodeRelease(PNUT_BLKNUM node)
; {
;     PnutNodeDataRelease(node);
	movw R16,R10
	xcall _PnutNodeDataRelease
;     PnutBlockRelease(node);
	movw R16,R10
	xcall _PnutBlockRelease
L28:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	ret
;         blkptr -> R22,R23
;         blkpos -> R12,R13
;            blk -> R22,R23
;         blkidx -> R14,R15
;             rc -> R20,R21
;         create -> y+20
;           size -> R10,R11
;         buffer -> y+16
;            pos -> y+12
;           node -> y+10
	.even
_PnutNodeGetDataPtr:
	xcall push_arg4
	xcall push_xgsetF0FC
	ldd R10,y+18
	ldd R11,y+19
; }
; 
; /*!
;  * \brief Retrieve a pointer into a node's data block.
;  *
;  * This function also returns the number of bytes available within
;  * the data block at the retrieved pointer position.
;  *
;  * \param node   Node number.
;  * \param pos    Data position.
;  * \param buffer Receives the pointer to the specified position.
;  * \param size   Receives the remaining number of bytes within the block.
;  * \param create If the position is beyond the current data size and if
;  *               this parameter is not equal zero, then a new data block 
;  *               will be allocated.
;  *
;  * \return 0 on success, otherwise returns an error code.
;  */
; static int PnutNodeGetDataPtr(PNUT_BLKNUM node, u_long pos, void **buffer, size_t * size, int create)
; {
;     int blkidx;                 /* Number of full blocks */
;     int rc = EINVAL;
	ldi R20,22
	ldi R21,0
; 
;     *buffer = NULL;
	clr R2
	clr R3
	ldd R30,y+16
	ldd R31,y+17
	std z+1,R3
	std z+0,R2
;     *size = 0;
	movw R30,R10
	std z+1,R3
	std z+0,R2
; 
;     /* Calculate the block index. Make sure it fits in our maximum file size. */
;     if ((blkidx = pos / PNUT_BLOCK_SIZE) < PNUT_BLOCKS_PER_NODE) {
	ldi R24,9
	ldi R25,0
	ldd R16,y+12
	ldd R17,y+13
	ldd R18,y+14
	ldd R19,y+15
	st -y,R24
	xcall lsr32
	movw R14,R16
	cpi R16,250
	ldi R30,0
	cpc R17,R30
	brlt X12
	xjmp L30
X12:
X7:
;         PNUT_BLKNUM blk;
; 
;         /* Get the data block number. Optionally allocate a new block. */
;         if ((blk = BankNodePointer(node)->node_blocks[blkidx]) < 0 && create) {
	ldd R16,y+10
	ldd R17,y+11
	xcall _BankNodePointer
	movw R2,R16
	ldi R16,2
	ldi R17,0
	movw R18,R14
	xcall empy16s
	movw R30,R16
	movw R24,R2
	adiw R24,12
	add R30,R24
	adc R31,R25
	ldd R12,z+0
	ldd R13,z+1
	movw R22,R12
	cpi R22,0
	ldi R30,0
	cpc R23,R30
	brge L32
X8:
	ldd R0,y+20
	ldd R1,y+21
	tst R0
	brne X9
	tst R1
	breq L32
X9:
;             if ((blk = PnutBlockAlloc()) < 0) {
	xcall _PnutBlockAlloc
	movw R12,R16
	movw R22,R12
	cpi R16,0
	ldi R30,0
	cpc R17,R30
	brge L34
X10:
;                 rc = ENOSPC;
	ldi R20,28
;             } else {
	xjmp L35
L34:
;                 BankNodePointer(node)->node_blocks[blkidx] = blk;
	ldd R16,y+10
	ldd R17,y+11
	xcall _BankNodePointer
	movw R12,R16
	ldi R16,2
	ldi R17,0
	movw R18,R14
	xcall empy16s
	movw R30,R16
	movw R24,R12
	adiw R24,12
	add R30,R24
	adc R31,R25
	std z+1,R23
	std z+0,R22
;             }
L35:
;         }
L32:
; 
;         /* 
;          * If we got a valid block, then set the pointer at the specified
;          * position and the remaining bytes within this block.
;          */
;         if (blk >= 0) {
	cpi R22,0
	ldi R30,0
	cpc R23,R30
	brlt L36
X11:
;             char *blkptr = (char *) BankNodePointer(blk);
	movw R16,R22
	xcall _BankNodePointer
	movw R12,R16
	movw R22,R12
;             int blkpos = pos % PNUT_BLOCK_SIZE; /* Position within block */
	ldi R24,255
	ldi R25,1
	ldi R26,0
	ldi R27,0
	ldd R2,y+12
	ldd R3,y+13
	ldd R4,y+14
	ldd R5,y+15
	and R2,R24
	and R3,R25
	and R4,R26
	and R5,R27
	movw R12,R2
; 
;             *buffer = blkptr + blkpos;
	add R2,R16
	adc R3,R17
	ldd R30,y+16
	ldd R31,y+17
	std z+1,R3
	std z+0,R2
;             *size = PNUT_BLOCK_SIZE - blkpos;
	ldi R24,512
	ldi R25,2
	sub R24,R12
	sbc R25,R13
	movw R30,R10
	std z+1,R25
	std z+0,R24
;             rc = 0;
	clr R20
	clr R21
;         }
L36:
;     }
L30:
;     return rc;
	movw R16,R20
L29:
	.dbline 0 ; func end
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             rc -> R10,R11
;          entry -> y+14
;           size -> y+12
;            pos -> y+8
;           node -> R12,R13
	.even
_PnutDirIsEmpty:
	xcall push_xgsetF0FC
	movw R12,R16
	sbiw R28,16
; }
; 
; 
; /* ------------------------ Directory routines ------------------------ */
; 
; /*!
;  * \brief Check if a directory is empty.
;  */
; static int PnutDirIsEmpty(PNUT_BLKNUM node)
; {
;     int rc = 1;
	ldi R24,1
	ldi R25,0
	movw R10,R24
;     u_long pos;
;     size_t size;
;     PNUT_DIRENTRY *entry;
; 
;     /* Loop through the data blocks of this directory node. */
;     for (pos = 0; pos < BankNodePointer(node)->node_size; pos += PNUT_DIRENT_SIZE) {
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	std y+8,R20
	std y+9,R21
	std y+10,R22
	std y+11,R23
	xjmp L42
L39:
; 
;         /* Get the pointer to the next directory entry. */
;         if (PnutNodeGetDataPtr(node, pos, (void *) &entry, &size, 0) || size == 0) {
	clr R2
	clr R3
	std y+7,R3
	std y+6,R2
	movw R24,R28
	adiw R24,12
	std y+5,R25
	std y+4,R24
	movw R24,R28
	adiw R24,14
	std y+3,R25
	std y+2,R24
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	std y+0,R4
	std y+1,R5
	movw R18,R2
	movw R16,R12
	xcall _PnutNodeGetDataPtr
	cpi R16,0
	cpc R16,R17
	brne L45
X13:
	ldd R2,y+12
	ldd R3,y+13
	tst R2
	brne L43
	tst R3
	brne L43
X14:
L45:
;             /* No more entries. */
;             break;
	xjmp L41
L43:
;         }
; 
;         if (size >= PNUT_DIRENT_SIZE) {
	ldd R24,y+12
	ldd R25,y+13
	cpi R24,32
	ldi R30,0
	cpc R25,R30
	brlo L46
X15:
;             if (entry->dir_inuse && strcmp(entry->dir_name, ".") && strcmp(entry->dir_name, "..")) {
	ldd R30,y+14
	ldd R31,y+15
	ldd R2,z+2
	tst R2
	breq L48
X16:
	ldi R18,<L50
	ldi R19,>L50
	movw R16,R30
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strcmp
	cpi R16,0
	cpc R16,R17
	breq L48
X17:
	ldi R18,<L51
	ldi R19,>L51
	ldd R16,y+14
	ldd R17,y+15
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strcmp
	movw R14,R16
	cpi R16,0
	cpc R16,R17
	breq L48
X18:
;                 /* Entry found */
;                 rc = 0;
	clr R10
	clr R11
;                 break;
	xjmp L41
L48:
;             }
;         }
L46:
;     }
L40:
	ldi R20,32
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
L42:
	movw R16,R12
	xcall _BankNodePointer
	movw R14,R16
	movw R30,R14
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+8
	ldd R7,y+9
	ldd R8,y+10
	ldd R9,y+11
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh X20
	xjmp L39
X20:
X19:
L41:
;     return rc;
	movw R16,R10
L38:
	.dbline 0 ; func end
	adiw R28,16
	xjmp pop_xgsetF0FC
;             rc -> R10,R11
;           size -> y+12
;            pos -> y+8
;          entry -> R12,R13
;            len -> y+28
;           path -> y+26
;           node -> y+24
	.even
_PnutDirFindEntry:
	xcall push_arg4
	xcall push_xgsetF0FC
	sbiw R28,14
	ldd R12,y+30
	ldd R13,y+31
; }
; 
; /*!
;  * \brief Find a directory entry with a specified name.
;  *
;  * \param node  Node of the directory to search.
;  * \param path  Path name of the entry to retrieve.
;  * \param len   Number of characters in the path name to consider,
;  *              starting from the beginning. If the path name
;  *              contains more than one component, the caller must
;  *              set this parameter to the length of the first one.
;  * \param entry Receives the pointer to the directory entry.
;  *
;  * \return 0 on succes, otherwise an error code is returned.
;  */
; static int PnutDirFindEntry(PNUT_BLKNUM node, CONST char *path, size_t len, PNUT_DIRENTRY ** entry)
; {
;     int rc = ENOENT;
	ldi R24,2
	ldi R25,0
	movw R10,R24
;     u_long pos;
;     size_t size;
; 
;     /* Loop through the data blocks of this directory node. */
;     for (pos = 0; pos < BankNodePointer(node)->node_size; pos += PNUT_DIRENT_SIZE) {
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	std y+8,R20
	std y+9,R21
	std y+10,R22
	std y+11,R23
	xjmp L56
L53:
; 
;         /* Get the pointer to the next directory entry. */
;         if (PnutNodeGetDataPtr(node, pos, (void **) entry, &size, 0) || size == 0) {
	clr R2
	clr R3
	std y+7,R3
	std y+6,R2
	movw R24,R28
	adiw R24,12
	std y+5,R25
	std y+4,R24
	std y+3,R13
	std y+2,R12
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	std y+0,R4
	std y+1,R5
	movw R18,R2
	ldd R16,y+24
	ldd R17,y+25
	xcall _PnutNodeGetDataPtr
	cpi R16,0
	cpc R16,R17
	brne L59
X21:
	ldd R2,y+12
	ldd R3,y+13
	tst R2
	brne L57
	tst R3
	brne L57
X22:
L59:
;             /* No more entries. */
;             break;
	xjmp L55
L57:
;         }
; 
;         /* Compare this entry. */
;         if (size >= PNUT_DIRENT_SIZE) { /* Skip entries across block boundaries. */
	ldd R24,y+12
	ldd R25,y+13
	cpi R24,32
	ldi R30,0
	cpc R25,R30
	brlo L60
X23:
;             if ((*entry)->dir_inuse &&  /* Skip unused entries. */
	movw R30,R12
	ldd R4,z+0
	ldd R5,z+1
	movw R30,R4
	ldd R2,z+2
	tst R2
	breq L62
X24:
	movw R16,R30
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strlen
	movw R14,R16
	ldd R0,y+28
	ldd R1,y+29
	cp R16,R0
	cpc R17,R1
	brne L62
X25:
	std y+1,R1
	std y+0,R0
	ldd R18,y+26
	ldd R19,y+27
	movw R30,R12
	ldd R16,z+0
	ldd R17,z+1
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strncmp
	movw R14,R16
	cpi R16,0
	cpc R16,R17
	brne L62
X26:
;                 strlen((*entry)->dir_name) == len &&    /* Skip non-matching names. */
;                 strncmp((*entry)->dir_name, path, len) == 0) {
;                 /* Requested entry found */
;                 rc = 0;
	clr R10
	clr R11
;                 break;
	xjmp L55
L62:
;             }
;         }
L60:
;     }
L54:
	ldi R20,32
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
L56:
	ldd R16,y+24
	ldd R17,y+25
	xcall _BankNodePointer
	movw R14,R16
	movw R30,R14
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+8
	ldd R7,y+9
	ldd R8,y+10
	ldd R9,y+11
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh X28
	xjmp L53
X28:
X27:
L55:
;     return rc;
	movw R16,R10
L52:
	.dbline 0 ; func end
	adiw R28,14
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             cp -> R12,R13
;          entry -> y+4
;             rc -> R20,R21
;            len -> R14,R15
;         result -> R10,R11
;           path -> y+18
;           node -> y+16
	.even
_PnutDirFindPath:
	xcall push_arg4
	xcall push_xgsetF0FC
	sbiw R28,6
	ldd R10,y+20
	ldd R11,y+21
; }
; 
; /*
;  * \brief Find directory entry of a specified path.
;  *
;  * \param node   First node block to search.
;  * \param path   Path to find.
;  * \param result Search result.
;  *
;  * \return Error code.
;  */
; static int PnutDirFindPath(PNUT_BLKNUM node, CONST char *path, PNUT_FINDRESULT * result)
; {
;     int rc = 0;
	clr R20
	clr R21
;     size_t len;
;     PNUT_DIRENTRY *entry;
; 
;     result->fr_pnode = node;
	movw R30,R10
	ldd R0,y+16
	ldd R1,y+17
	std z+3,R1
	std z+2,R0
;     result->fr_node = -1;
	ldi R24,-1
	ldi R25,-1
	std z+1,R25
	std z+0,R24
;     result->fr_name = NULL;
	clr R2
	clr R3
	std z+5,R3
	std z+4,R2
	xjmp L66
L65:
; 
;     while (*path == '/') {
;         path++;
	ldd R24,y+18
	ldd R25,y+19
	adiw R24,1
	std y+19,R25
	std y+18,R24
;     }
L66:
	ldd R30,y+18
	ldd R31,y+19
	ldd R24,z+0
	cpi R24,47
	breq L65
X29:
; 
;     if (*path == 0) {
	ldd R2,z+0
	tst R2
	breq X39
	xjmp L71
X39:
X30:
;         path = ".";
	ldi R24,<L50
	ldi R25,>L50
	std y+19,R25
	std y+18,R24
;     }
	xjmp L71
L70:
; 
;     /* 
;      * Loop for each path component.
;      */
;     while (*path) {
;         CONST char *cp;
; 
;         /* Make sure that this is a directory node. */
;         if (BankNodePointer(node)->node_type != NODETYPE_DIR) {
	ldd R16,y+16
	ldd R17,y+17
	xcall _BankNodePointer
	movw R30,R16
	ldd R24,z+0
	cpi R24,1
	breq L73
X31:
;             rc = ENOTDIR;
	ldi R20,20
	ldi R21,0
;             break;
	xjmp L72
L73:
;         }
; 
;         /* Retrieve the length of first path component. */
;         for (len = 0, cp = path; *cp && *cp != '/'; cp++) {
	clr R14
	clr R15
	ldd R12,y+18
	ldd R13,y+19
	xjmp L78
L75:
;             len++;
	movw R24,R14
	adiw R24,1
	movw R14,R24
;         }
L76:
	movw R24,R12
	adiw R24,1
	movw R12,R24
L78:
	movw R30,R12
	ldd R2,z+0
	clr R3
	tst R2
	breq L79
X32:
	mov R24,R2
	cpi R24,47
	brne L75
X33:
L79:
; 
;         /* 
;          * If this component is last, we found all parents.
;          * Keep the last one stored in the result.
;          */
;         if (*cp == 0) {
	movw R30,R12
	ldd R2,z+0
	tst R2
	brne L80
X34:
;             result->fr_name = path;
	movw R30,R10
	ldd R0,y+18
	ldd R1,y+19
	std z+5,R1
	std z+4,R0
;         }
L80:
; 
;         /* Try to find this component. */
;         if ((rc = PnutDirFindEntry(node, path, len, &entry)) != 0) {
	movw R24,R28
	adiw R24,4
	std y+3,R25
	std y+2,R24
	std y+1,R15
	std y+0,R14
	ldd R18,y+18
	ldd R19,y+19
	ldd R16,y+16
	ldd R17,y+17
	xcall _PnutDirFindEntry
	movw R22,R16
	movw R20,R22
	cpi R16,0
	cpc R16,R17
	breq L82
X35:
;             rc = ENOENT;
	ldi R20,2
	ldi R21,0
;             break;
	xjmp L72
L82:
;         }
; 
;         /* Component found. Return if this was the last one. */
;         result->fr_node = entry->dir_node;
	ldd R30,y+4
	ldd R31,y+5
	ldd R2,z+0
	ldd R3,z+1
	movw R30,R10
	std z+1,R3
	std z+0,R2
;         if (*cp == 0) {
	movw R30,R12
	ldd R2,z+0
	tst R2
	brne L84
X36:
;             break;
	xjmp L72
L84:
;         }
; 
;         /* Not the last component. Store node as previous node. */
;         result->fr_pnode = result->fr_node;
	movw R30,R10
	ldd R2,z+0
	ldd R3,z+1
	std z+3,R3
	std z+2,R2
; 
;         /* Move forward to the next path component. */
;         node = result->fr_node;
	ldd R2,z+0
	ldd R3,z+1
	std y+17,R3
	std y+16,R2
;         path += len;
	ldd R0,y+18
	ldd R1,y+19
	add R0,R14
	adc R1,R15
	std y+19,R1
	std y+18,R0
;         if (*path == '/') {
	movw R30,R0
	ldd R24,z+0
	cpi R24,47
	brne L86
X37:
;             path++;
	movw R24,R30
	adiw R24,1
	std y+19,R25
	std y+18,R24
;         }
L86:
;     }
L71:
	ldd R30,y+18
	ldd R31,y+19
	ldd R2,z+0
	tst R2
	breq X40
	xjmp L70
X40:
X38:
L72:
;     return rc;
	movw R16,R20
L64:
	.dbline 0 ; func end
	adiw R28,6
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             fp -> R22,R23
;          found -> y+2
;             rc -> R20,R21
;            dir -> R10,R11
	.even
_PnutDirOpen:
	xcall push_xgsetF03C
	movw R10,R16
	sbiw R28,8
; }
; 
; 
; /*!
;  * \brief Open a directory.
;  *
;  * \param dir Pointer to the internal directory information structure.
;  *            The pathname of the directory to open must have been
;  *            copied to the data buffer prior calling this routine.
;  *
;  * \return 0 on success, -1 otherwise.
;  */
; static int PnutDirOpen(DIR * dir)
; {
;     PNUTFILE *fp;
;     PNUT_FINDRESULT found;
;     int rc;
; 
;     /* Locate the entry in our directory structure. */
;     if ((rc = PnutDirFindPath(root, dir->dd_buf, &found)) != 0) {
	movw R24,R28
	adiw R24,2
	std y+1,R25
	std y+0,R24
	movw R30,R10
	ldd R18,z+6
	ldd R19,z+7
	lds R16,_root
	lds R17,_root+1
	xcall _PnutDirFindPath
	movw R20,R16
	cpi R20,0
	cpc R20,R21
	breq L89
X41:
;         errno = rc;
	sts _errno+1,R21
	sts _errno,R20
;         rc = -1;
	ldi R20,-1
	ldi R21,-1
;     } else {
	xjmp L90
L89:
;         if (BankNodePointer(found.fr_node)->node_type != NODETYPE_DIR) {
	ldd R16,y+2
	ldd R17,y+3
	xcall _BankNodePointer
	movw R30,R16
	ldd R24,z+0
	cpi R24,1
	breq L91
X42:
;             errno = ENOTDIR;
	ldi R24,20
	ldi R25,0
	sts _errno+1,R25
	sts _errno,R24
;             rc = -1;
	ldi R20,-1
	ldi R21,-1
;         }
	xjmp L92
L91:
;         /* Allocate a PNUT file descriptor. */
;         else if ((fp = malloc(sizeof(PNUTFILE))) == 0) {
	ldi R16,8
	ldi R17,0
	xcall _malloc_M
	movw R12,R16
	movw R22,R12
	cpi R16,0
	cpc R16,R17
	brne L93
X43:
;             errno = ENOMEM;
	ldi R24,12
	ldi R25,0
	sts _errno+1,R25
	sts _errno,R24
;             rc = -1;
	ldi R20,-1
	ldi R21,-1
;         }
	xjmp L94
L93:
;         /* 
;          * Initialize the descriptor and store it in the directory
;          * information structure.
;          */
;         else {
;             fp->f_node = found.fr_node;
	ldd R2,y+2
	ldd R3,y+3
	movw R30,R22
	std z+1,R3
	std z+0,R2
;             fp->f_pos = 0;
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std z+2,R24
	std z+3,R25
	std z+4,R26
	std z+5,R27
;             dir->dd_fd.nf_fcb = fp;
	movw R30,R10
	std z+5,R23
	std z+4,R22
;             /* Keep track of the number of open calls. */
;             BankNodePointer(fp->f_node)->node_refs++;
	movw R30,R22
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R22,R16
	subi R22,255  ; offset = 1
	sbci R23,255
	movw R30,R22
	ldd R24,z+0
	subi R24,255    ; addi 1
	std z+0,R24
;         }
L94:
L92:
;     }
L90:
;     return rc;
	movw R16,R20
L88:
	.dbline 0 ; func end
	adiw R28,8
	xjmp pop_xgsetF03C
;             fp -> R10,R11
;            dir -> R20,R21
	.even
_PnutDirClose:
	xcall push_xgset300C
	movw R20,R16
; }
; 
; /*!
;  * \brief Close a directory.
;  */
; static int PnutDirClose(DIR * dir)
; {
;     PNUTFILE *fp = dir->dd_fd.nf_fcb;
	movw R30,R20
	ldd R10,z+4
	ldd R11,z+5
; 
;     if (fp) {
	tst R10
	brne X44
	tst R11
	breq L96
X44:
;         BankNodePointer(fp->f_node)->node_refs--;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R24,R16
	adiw R24,1
	movw R30,R24
	ldd R24,z+0
	subi R24,1
	std z+0,R24
;         free(fp);
	movw R16,R10
	xcall _free_M
;         return 0;
	clr R16
	clr R17
	xjmp L95
L96:
;     }
;     return EINVAL;
	ldi R16,22
	ldi R17,0
L95:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;            ent -> R12,R13
;             rc -> y+16
;             fp -> R14,R15
;           size -> y+14
;          entry -> y+12
;            pos -> y+8
;            dir -> R10,R11
	.even
_PnutDirRead:
	xcall push_xgsetF0FC
	movw R10,R16
	sbiw R28,18
; }
; 
; /*!
;  * \brief Read the next directory entry.
;  */
; static int PnutDirRead(DIR * dir)
; {
;     int rc = 0;
	clr R0
	clr R1
	std y+17,R1
	std y+16,R0
;     u_long pos;
;     PNUT_DIRENTRY *entry;
;     size_t size;
;     PNUTFILE *fp = dir->dd_fd.nf_fcb;
	movw R30,R10
	ldd R14,z+4
	ldd R15,z+5
;     struct dirent *ent = (struct dirent *) dir->dd_buf;
	ldd R12,z+6
	ldd R13,z+7
; 
;     ent->d_name[0] = 0;
	clr R2
	movw R30,R12
	std z+8,R2
;     for (pos = fp->f_pos; pos < BankNodePointer(fp->f_node)->node_size; pos += PNUT_DIRENT_SIZE) {
	movw R30,R14
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
	xjmp L102
L99:
;         /* Retrieve the entry at the current position. */
;         rc = PnutNodeGetDataPtr(fp->f_node, pos, (void *) &entry, &size, 0);
	clr R2
	clr R3
	std y+7,R3
	std y+6,R2
	movw R24,R28
	adiw R24,14
	std y+5,R25
	std y+4,R24
	movw R24,R28
	adiw R24,12
	std y+3,R25
	std y+2,R24
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	std y+0,R4
	std y+1,R5
	movw R18,R2
	movw R30,R14
	ldd R16,z+0
	ldd R17,z+1
	xcall _PnutNodeGetDataPtr
	std y+17,R17
	std y+16,R16
;         if (rc || size == 0) {
	movw R0,R16
	cpi R16,0
	cpc R16,R17
	brne L105
X45:
	ldd R2,y+14
	ldd R3,y+15
	tst R2
	brne L103
	tst R3
	brne L103
X46:
L105:
;             /* No more entries. */
;             break;
	xjmp L101
L103:
;         }
;         /* Skip entries across block boundaries and unused entires. */
;         if (size >= PNUT_DIRENT_SIZE && entry->dir_inuse) {
	ldd R24,y+14
	ldd R25,y+15
	cpi R24,32
	ldi R30,0
	cpc R25,R30
	brsh X50
	xjmp L106
X50:
X47:
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+2
	tst R2
	brne X51
	xjmp L106
X51:
X48:
;             memset(ent, 0, sizeof(struct dirent));
	ldi R24,264
	ldi R25,1
	std y+1,R25
	std y+0,R24
	clr R18
	clr R19
	movw R16,R12
	xcall _memset
;             ent->d_fileno = entry->dir_node;
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+0
	ldd R3,z+1
	clr R4
	sbrc R3,7
	com R4
	clr R5
	sbrc R4,7
	com R5
	movw R30,R12
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;             ent->d_namlen = (u_char) strlen(entry->dir_name);
	ldd R16,y+12
	ldd R17,y+13
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strlen
	movw R30,R12
	std z+7,R16
;             strcpy(ent->d_name, entry->dir_name);
	ldd R18,y+12
	ldd R19,y+13
	subi R18,253  ; offset = 3
	sbci R19,255
	movw R16,R30
	subi R16,248  ; offset = 8
	sbci R17,255
	xcall _strcpy
;             fp->f_pos = pos + PNUT_DIRENT_SIZE;
	ldi R20,32
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	movw R30,R14
	std z+2,R2
	std z+3,R3
	std z+4,R4
	std z+5,R5
;             break;
	xjmp L101
L106:
;         }
;     }
L100:
	ldi R20,32
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
L102:
	movw R30,R14
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R10,R16
	movw R30,R10
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+8
	ldd R7,y+9
	ldd R8,y+10
	ldd R9,y+11
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh X52
	xjmp L99
X52:
X49:
L101:
;     return rc;
	ldd R16,y+16
	ldd R17,y+17
L98:
	.dbline 0 ; func end
	adiw R28,18
	xjmp pop_xgsetF0FC
;             rc -> R10,R11
;             np -> R12,R13
;           size -> y+14
;          entry -> y+12
;            pos -> y+8
;          enode -> R14,R15
;           name -> R12,R13
;          dnode -> y+26
	.even
_PnutDirAddEntry:
	xcall push_arg4
	xcall push_xgsetF0FC
	movw R12,R18
	sbiw R28,16
	ldd R14,y+30
	ldd R15,y+31
; }
; 
; /*!
;  * \brief Add directory entry.
;  *
;  * \param dnode The entry is added to this directory node.
;  * \param name  The entry's name.
;  * \param enode  The entry's node.
;  *
;  * \return 0 if successful. Otherwise returns an error code.
;  */
; static int PnutDirAddEntry(PNUT_BLKNUM dnode, CONST char *name, PNUT_BLKNUM enode)
; {
;     int rc = 0;
	clr R10
	clr R11
;     u_long pos = 0;
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	std y+8,R20
	std y+9,R21
	std y+10,R22
	std y+11,R23
;     size_t size;
;     PNUT_DIRENTRY *entry;
;     PNUT_NODE *np;
; 
;     /*
;      * Loop through all directory entries until an unused one
;      * is found. If required, a new data block is allocated,
;      * so success is guaranteed unless we run out of free
;      * blocks.
;      */
;     for (;;) {
L109:
;         /* 
;          * Get the data pointer to the specified position. Automatically
;          * create a new block if we reached the end of the data.
;          */
;         if ((rc = PnutNodeGetDataPtr(dnode, pos, (void *) &entry, &size, 1)) != 0) {
	ldi R24,1
	ldi R25,0
	std y+7,R25
	std y+6,R24
	movw R24,R28
	adiw R24,14
	std y+5,R25
	std y+4,R24
	movw R24,R28
	adiw R24,12
	std y+3,R25
	std y+2,R24
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	std y+0,R4
	std y+1,R5
	movw R18,R2
	ldd R16,y+26
	ldd R17,y+27
	xcall _PnutNodeGetDataPtr
	movw R10,R16
	tst R10
	brne X53
	tst R11
	breq L113
X53:
;             break;
	xjmp L111
L113:
;         }
;         pos += PNUT_DIRENT_SIZE;
	ldi R20,32
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	add R2,R20
	adc R3,R21
	adc R4,R22
	adc R5,R23
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
; 
;         /* Process entry if it doesn't cross block boundaries. */
;         if (size >= PNUT_DIRENT_SIZE) {
	ldd R24,y+14
	ldd R25,y+15
	cpi R24,32
	ldi R30,0
	cpc R25,R30
	brlo L109
X54:
;             /* Did we find a previously released or a newly allocated entry? */
;             if (entry->dir_inuse == 0) {
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+2
	tst R2
	breq X57
	xjmp L109
X57:
X55:
;                 /* Update the entry. */
;                 entry->dir_node = enode;
	std z+1,R15
	std z+0,R14
;                 entry->dir_inuse = 1;
	ldi R24,1
	std z+2,R24
;                 strcpy(entry->dir_name, name);
	movw R18,R12
	movw R16,R30
	subi R16,253  ; offset = 3
	sbci R17,255
	xcall _strcpy
; 
;                 /* Update the directory node. */
;                 np = BankNodePointer(dnode);
	ldd R16,y+26
	ldd R17,y+27
	xcall _BankNodePointer
	movw R12,R16
;                 np->node_mtime = time(0);
	clr R16
	clr R17
	xcall _time
	movw R30,R12
	std z+8,R16
	std z+9,R17
	std z+10,R18
	std z+11,R19
;                 if (pos > np->node_size) {
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+8
	ldd R7,y+9
	ldd R8,y+10
	ldd R9,y+11
	cp R2,R6
	cpc R3,R7
	cpc R4,R8
	cpc R5,R9
	brsh L119
X56:
;                     np->node_size = pos;
	movw R2,R6
	movw R4,R8
	std z+4,R2
	std z+5,R3
	std z+6,R4
	std z+7,R5
;                 }
L119:
; 
;                 /* Update the entry's node. */
;                 np = BankNodePointer(enode);
	movw R16,R14
	xcall _BankNodePointer
	movw R14,R16
	movw R12,R14
;                 np->node_links++;
	movw R24,R12
	adiw R24,2
	movw R12,R24
	movw R30,R24
	ldd R24,z+0
	ldd R25,z+1
	adiw R24,1
	std z+1,R25
	std z+0,R24
;                 break;
;             }
;         }
;     }
L111:
;     return rc;
	movw R16,R10
L108:
	.dbline 0 ; func end
	adiw R28,16
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             rc -> R14,R15
;          rnode -> R14,R15
;          entry -> y+4
;            rnp -> R20,R21
;           name -> R12,R13
;           node -> R10,R11
	.even
_PnutDirDelEntry:
	xcall push_xgsetF0FC
	movw R12,R18
	movw R10,R16
	sbiw R28,6
; }
; 
; /*!
;  * \brief Remove a directory entry.
;  *
;  * Entries referring to nodes with active open calls are not deleted.
;  *
;  * Entries referring to directory nodes with more than two links left
;  * are not deleted.
;  *
;  * File nodes are released if the entry to remove was the last link
;  * to them.
;  *
;  * \param node Remove the entry from the directory node in this block.
;  * \param name Name of the entry to be removed.
;  *
;  * \return 0 if successful. Otherwise returns an error code.
;  */
; static int PnutDirDelEntry(PNUT_BLKNUM node, CONST char *name)
; {
;     int rc;
;     PNUT_DIRENTRY *entry;
;     PNUT_NODE *rnp;
;     PNUT_BLKNUM rnode;
; 
;     /* Make sure this entry exists. */
;     if ((rc = PnutDirFindEntry(node, name, strlen(name), &entry)) != 0) {
	movw R16,R12
	xcall _strlen
	movw R24,R28
	adiw R24,4
	std y+3,R25
	std y+2,R24
	std y+1,R17
	std y+0,R16
	movw R18,R12
	movw R16,R10
	xcall _PnutDirFindEntry
	movw R14,R16
	tst R14
	brne X58
	tst R15
	breq L122
X58:
;         return rc;
	xjmp L121
L122:
;     }
; 
;     /* The node to remove. */
;     rnode = entry->dir_node;
	ldd R30,y+4
	ldd R31,y+5
	ldd R14,z+0
	ldd R15,z+1
;     rnp = BankNodePointer(rnode);
	movw R16,R14
	xcall _BankNodePointer
	movw R20,R16
; 
;     /* Make sure that this node has no pending open calls. */
;     if (rnp->node_refs) {
	movw R30,R20
	ldd R2,z+1
	tst R2
	breq L124
X59:
;         return EACCES;
	ldi R16,13
	ldi R17,0
	xjmp L121
L124:
;     }
; 
;     /* We only remove empty directories. */
;     if (rnp->node_type == NODETYPE_DIR) {
	movw R30,R20
	ldd R24,z+0
	cpi R24,1
	brne L126
X60:
;         if (rnp->node_links > 2 || !PnutDirIsEmpty(rnode)) {
	ldi R24,2
	ldi R25,0
	ldd R2,z+2
	ldd R3,z+3
	cp R24,R2
	cpc R25,R3
	brlo L130
X61:
	movw R16,R14
	xcall _PnutDirIsEmpty
	movw R22,R16
	cpi R16,0
	cpc R16,R17
	brne L128
X62:
L130:
;             return EACCES;
	ldi R16,13
	ldi R17,0
	xjmp L121
L128:
;         }
;         rnp = BankNodePointer(node);
	movw R16,R10
	xcall _BankNodePointer
	movw R22,R16
	movw R20,R22
;         rnp->node_links--;
	subi R20,254  ; offset = 2
	sbci R21,255
	movw R30,R20
	ldd R24,z+0
	ldd R25,z+1
	sbiw R24,1
	std z+1,R25
	std z+0,R24
;         PnutNodeRelease(rnode);
	movw R16,R14
	xcall _PnutNodeRelease
;     }
	xjmp L127
L126:
; 
;     /* Remove a file. */
;     else {
;         PnutNodeRelease(rnode);
	movw R16,R14
	xcall _PnutNodeRelease
;     }
L127:
; 
;     /* Mark this entry unused. */
;     PnutDirFindEntry(node, name, strlen(name), &entry);
	movw R16,R12
	xcall _strlen
	movw R14,R16
	movw R24,R28
	adiw R24,4
	std y+3,R25
	std y+2,R24
	std y+1,R15
	std y+0,R14
	movw R18,R12
	movw R16,R10
	xcall _PnutDirFindEntry
;     entry->dir_inuse = 0;
	clr R2
	ldd R30,y+4
	ldd R31,y+5
	std z+2,R2
; 
;     return 0;
	clr R16
	clr R17
L121:
	.dbline 0 ; func end
	adiw R28,6
	xjmp pop_xgsetF0FC
;           node -> R10,R11
;          found -> y+2
;             ec -> R20,R21
;           path -> R10,R11
	.even
_PnutDirCreate:
	xcall push_xgset303C
	movw R10,R16
	sbiw R28,8
; }
; 
; /*!
;  * \brief Create a new subdirectory.
;  *
;  * \param path Full path to the directory.
;  *
;  * \return 0 on success. Otherwise -1 is returned.
;  */
; static int PnutDirCreate(CONST char *path)
; {
;     PNUT_BLKNUM node;
;     PNUT_FINDRESULT found;
;     int ec;
; 
;     /* Return an error if this entry already exists. */
;     if ((ec = PnutDirFindPath(root, path, &found)) == 0) {
	movw R24,R28
	adiw R24,2
	std y+1,R25
	std y+0,R24
	movw R18,R10
	lds R16,_root
	lds R17,_root+1
	xcall _PnutDirFindPath
	movw R20,R16
	cpi R20,0
	cpc R20,R21
	brne L132
X63:
;         ec = EEXIST;
	ldi R20,17
	ldi R21,0
;     }
	xjmp L133
L132:
; 
;     /* If this component does not exist, we can create it... */
;     else if (ec == ENOENT) {
	cpi R20,2
	ldi R30,0
	cpc R21,R30
	breq X71
	xjmp L134
X71:
X64:
;         /* ...provided that all parents had been found. */
;         if (found.fr_name) {
	ldd R2,y+6
	ldd R3,y+7
	tst R2
	brne X65
	tst R3
	brne X72
	xjmp L136
X72:
X65:
;             /* Allocate a new node block. */
;             if ((node = PnutNodeAlloc(NODETYPE_DIR)) < 0) {
	ldi R16,1
	xcall _PnutNodeAlloc
	movw R10,R16
	cpi R16,0
	ldi R30,0
	cpc R17,R30
	brge L139
X66:
;                 ec = ENOSPC;
	ldi R20,28
	ldi R21,0
;             }
	xjmp L140
L139:
;             /* Create a reference to itself. */
;             else if ((ec = PnutDirAddEntry(node, ".", node)) != 0) {
	std y+1,R11
	std y+0,R10
	ldi R18,<L50
	ldi R19,>L50
	movw R16,R10
	xcall _PnutDirAddEntry
	movw R12,R16
	movw R20,R12
	cpi R16,0
	cpc R16,R17
	breq L141
X67:
;                 PnutNodeRelease(node);
	movw R16,R10
	xcall _PnutNodeRelease
;             }
	xjmp L142
L141:
;             /* Create a reference to the parent. */
;             else if ((ec = PnutDirAddEntry(node, "..", found.fr_pnode)) != 0) {
	ldd R2,y+4
	ldd R3,y+5
	std y+1,R3
	std y+0,R2
	ldi R18,<L51
	ldi R19,>L51
	movw R16,R10
	xcall _PnutDirAddEntry
	movw R12,R16
	movw R20,R12
	cpi R16,0
	cpc R16,R17
	breq L143
X68:
;                 PnutNodeRelease(node);
	movw R16,R10
	xcall _PnutNodeRelease
;             }
	xjmp L144
L143:
;             /* Create a reference in the parent. */
;             else if ((ec = PnutDirAddEntry(found.fr_pnode, found.fr_name, node)) != 0) {
	std y+1,R11
	std y+0,R10
	ldd R18,y+6
	ldd R19,y+7
	ldd R16,y+4
	ldd R17,y+5
	xcall _PnutDirAddEntry
	movw R12,R16
	movw R20,R12
	cpi R16,0
	cpc R16,R17
	breq L146
X69:
;                 PnutNodeRelease(node);
	movw R16,R10
	xcall _PnutNodeRelease
;             }
L146:
L144:
L142:
L140:
;         }
L136:
;     }
L134:
L133:
; 
;     /* Something went wrong. */
;     if (ec) {
	cpi R20,0
	cpc R20,R21
	breq L150
X70:
;         errno = ec;
	sts _errno+1,R21
	sts _errno,R20
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L131
L150:
;     }
;     return 0;
	clr R16
	clr R17
L131:
	.dbline 0 ; func end
	adiw R28,8
	xjmp pop_xgset303C
;           file -> R10,R11
;          found -> y+2
;            nfp -> R20,R21
;             rc -> R10,R11
;           node -> R22,R23
;            acc -> y+28
;           mode -> R14,R15
;           path -> R10,R11
;            dev -> y+22
	.even
_PnutFileOpen:
	xcall push_arg4
	xcall push_xgsetF0FC
	movw R10,R18
	sbiw R28,12
	ldd R14,y+26
	ldd R15,y+27
; }
; 
; /* ------------------------ File routines ------------------------ */
; 
; /*!
;  * \brief Open a file.
;  *
;  * This function is called by the low level open routine of the C runtime 
;  * library, using the _NUTDEVICE::dev_open entry.
;  *
;  * \param dev  Pointer to the NUTDEVICE structure.
;  * \param path Pathname of the file to open.
;  * \param mode Operation mode. Any of the following values may be or-ed:
;  * - \ref _O_RDONLY
;  * - \ref _O_WRONLY
;  * - \ref _O_RDWR
;  * - \ref _O_APPEND
;  * - \ref _O_CREAT
;  * - \ref _O_TRUNC
;  * - \ref _O_EXCL
;  * - \ref _O_TEXT
;  * - \ref _O_BINARY
;  * \param acc Ignored, should be zero.
;  *
;  * \return Pointer to a NUTFILE structure if successful or NUTFILE_EOF otherwise.
;  */
; static NUTFILE *PnutFileOpen(NUTDEVICE * dev, CONST char *path, int mode, int acc)
; {
;     PNUT_BLKNUM node = -1;
	ldi R22,-1
	ldi R23,-1
;     PNUT_FINDRESULT found;
;     int rc;
;     PNUTFILE *file;
;     NUTFILE *nfp = NUTFILE_EOF;
	ldi R20,-1
	ldi R21,-1
; 
;     /* Try to find an exisiting file. */
;     if ((rc = PnutDirFindPath(root, path, &found)) == 0) {
	movw R24,R28
	adiw R24,2
	std y+1,R25
	std y+0,R24
	movw R18,R10
	lds R16,_root
	lds R17,_root+1
	xcall _PnutDirFindPath
	movw R10,R16
	tst R10
	brne L153
	tst R11
	brne L153
X73:
;         /* Return an error, if this is no regular file. */
;         if (BankNodePointer(found.fr_node)->node_type != NODETYPE_REG) {
	ldd R16,y+2
	ldd R17,y+3
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+0
	tst R2
	breq L155
X74:
;             errno = EISDIR;
	ldi R24,21
	ldi R25,0
	sts _errno+1,R25
	sts _errno,R24
;         }
	xjmp L154
L155:
; 
;         /*
;          * We return an error, if the file exists and _O_EXCL has been
;          * set with _O_CREAT.
;          */
;         else if ((mode & (_O_CREAT | _O_EXCL)) == (_O_CREAT | _O_EXCL)) {
	movw R24,R14
	andi R24,0
	andi R25,5
	cpi R24,0
	ldi R30,5
	cpc R25,R30
	brne L157
X75:
;             errno = EEXIST;
	ldi R24,17
	ldi R25,0
	sts _errno+1,R25
	sts _errno,R24
;         } else {
	xjmp L154
L157:
;             node = found.fr_node;
	ldd R22,y+2
	ldd R23,y+3
;             if (mode & _O_TRUNC) {
	sbrs R15,1
	rjmp L154
X76:
;                 PnutNodeDataRelease(node);
	movw R16,R22
	xcall _PnutNodeDataRelease
;             }
;         }
;     }
	xjmp L154
L153:
; 
;     else if (rc == ENOENT) {
	movw R24,R10
	cpi R24,2
	ldi R30,0
	cpc R25,R30
	brne L161
X77:
;         /*
;          * If the search failed on the last path component, then
;          * PnutDirFindPath() set fr_name. Only in this case we can
;          * create a new file.
;          */
;         if (found.fr_name && (mode & _O_CREAT)) {
	ldd R2,y+6
	ldd R3,y+7
	tst R2
	brne X78
	tst R3
	breq L163
X78:
	sbrs R15,0
	rjmp L163
X79:
;             node = PnutNodeAlloc(NODETYPE_REG);
	clr R16
	xcall _PnutNodeAlloc
	movw R12,R16
	movw R22,R12
; 
;             if (node < 0) {
	cpi R16,0
	ldi R30,0
	cpc R17,R30
	brge L166
X80:
;                 errno = ENOSPC;
	ldi R24,28
	ldi R25,0
	sts _errno+1,R25
	sts _errno,R24
;                 return NUTFILE_EOF;
	ldi R16,-1
	ldi R17,-1
	xjmp L152
L166:
;             }
; 
;             rc = PnutDirAddEntry(found.fr_pnode, found.fr_name, node);
	std y+1,R23
	std y+0,R22
	ldd R18,y+6
	ldd R19,y+7
	ldd R16,y+4
	ldd R17,y+5
	xcall _PnutDirAddEntry
	movw R12,R16
	movw R10,R12
; 
;             if (rc) {
	cpi R16,0
	cpc R16,R17
	breq L170
X81:
;                 PnutBlockRelease(node);
	movw R16,R22
	xcall _PnutBlockRelease
;             }
L170:
;         }
L163:
;     }
L161:
L154:
; 
;     if (rc == 0 && node >= 0) {
	tst R10
	breq X87
	xjmp L172
X87:
	tst R11
	breq X88
	xjmp L172
X88:
X82:
	cpi R22,0
	ldi R30,0
	cpc R23,R30
	brge X89
	xjmp L172
X89:
X83:
;         if ((file = malloc(sizeof(PNUTFILE))) != 0) {
	ldi R16,8
	ldi R17,0
	xcall _malloc_M
	movw R12,R16
	movw R10,R12
	cpi R16,0
	cpc R16,R17
	brne X90
	xjmp L174
X90:
X84:
;             file->f_flag |= mode & (_O_RDONLY | _O_WRONLY | _O_APPEND);
	movw R24,R10
	adiw R24,6
	movw R12,R24
	movw R24,R14
	andi R24,9
	andi R25,0
	movw R30,R12
	ldd R2,z+0
	ldd R3,z+1
	or R2,R24
	or R3,R25
	std z+1,R3
	std z+0,R2
;             file->f_pos = (mode & _O_APPEND) ? BankNodePointer(node)->node_size : 0;
	sbrs R14,3
	rjmp L176
X85:
	movw R16,R22
	xcall _BankNodePointer
	movw R14,R16
	movw R30,R14
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
	xjmp L177
L176:
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+8,R24
	std y+9,R25
	std y+10,R26
	std y+11,R27
L177:
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	movw R30,R10
	std z+2,R2
	std z+3,R3
	std z+4,R4
	std z+5,R5
;             file->f_node = node;
	std z+1,R23
	std z+0,R22
; 
;             if ((nfp = malloc(sizeof(NUTFILE))) == 0) {
	ldi R16,6
	ldi R17,0
	xcall _malloc_M
	movw R14,R16
	movw R20,R14
	cpi R16,0
	cpc R16,R17
	brne L178
X86:
;                 free(file);
	movw R16,R10
	xcall _free_M
;                 nfp = NUTFILE_EOF;
	ldi R20,-1
	ldi R21,-1
;             } else {
	xjmp L179
L178:
;                 nfp->nf_next = 0;
	clr R2
	clr R3
	movw R30,R20
	std z+1,R3
	std z+0,R2
;                 nfp->nf_dev = dev;
	ldd R0,y+22
	ldd R1,y+23
	std z+3,R1
	std z+2,R0
;                 nfp->nf_fcb = file;
	std z+5,R11
	std z+4,R10
;                 /* Keep track of the number of open calls. */
;                 BankNodePointer(node)->node_refs++;
	movw R16,R22
	xcall _BankNodePointer
	movw R22,R16
	subi R22,255  ; offset = 1
	sbci R23,255
	movw R30,R22
	ldd R24,z+0
	subi R24,255    ; addi 1
	std z+0,R24
;             }
L179:
;         }
L174:
;     }
L172:
;     return nfp;
	movw R16,R20
L152:
	.dbline 0 ; func end
	adiw R28,12
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             fp -> R10,R11
;            nfp -> R20,R21
	.even
_PnutFileClose:
	xcall push_xgset300C
	movw R20,R16
; }
; 
; /*!
;  * \brief Close a file.
;  */
; static int PnutFileClose(NUTFILE * nfp)
; {
;     if (nfp != NUTFILE_EOF) {
	cpi R20,255
	ldi R30,255
	cpc R21,R30
	breq L181
X91:
;         PNUTFILE *fp = nfp->nf_fcb;
	movw R30,R20
	ldd R10,z+4
	ldd R11,z+5
; 
;         if (fp) {
	tst R10
	brne X92
	tst R11
	breq L183
X92:
;             BankNodePointer(fp->f_node)->node_refs--;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R24,R16
	adiw R24,1
	movw R30,R24
	ldd R24,z+0
	subi R24,1
	std z+0,R24
;             free(fp);
	movw R16,R10
	xcall _free_M
;         }
L183:
;         free(nfp);
	movw R16,R20
	xcall _free_M
; 
;         return 0;
	clr R16
	clr R17
	xjmp L180
L181:
;     }
;     return EINVAL;
	ldi R16,22
	ldi R17,0
L180:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;          found -> y+2
;             ec -> R10,R11
;           path -> R10,R11
	.even
_PnutDelete:
	st -y,R10
	st -y,R11
	movw R10,R16
	sbiw R28,8
; }
; 
; /*!
;  * \brief Delete a file or directory.
;  *
;  * \param path Full path to the directory.
;  *
;  * \return 0 on success. Otherwise -1 is returned.
;  */
; static int PnutDelete(char *path)
; {
;     int ec;
;     PNUT_FINDRESULT found;
; 
;     if ((ec = PnutDirFindPath(root, path, &found)) == 0) {
	movw R24,R28
	adiw R24,2
	std y+1,R25
	std y+0,R24
	movw R18,R10
	lds R16,_root
	lds R17,_root+1
	xcall _PnutDirFindPath
	movw R10,R16
	tst R10
	brne L186
	tst R11
	brne L186
X93:
;         ec = PnutDirDelEntry(found.fr_pnode, found.fr_name);
	ldd R18,y+6
	ldd R19,y+7
	ldd R16,y+4
	ldd R17,y+5
	xcall _PnutDirDelEntry
	movw R10,R16
;     }
L186:
;     if (ec) {
	tst R10
	brne X94
	tst R11
	breq L190
X94:
;         errno = ec;
	sts _errno+1,R11
	sts _errno,R10
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L185
L190:
;     }
;     return 0;
	clr R16
	clr R17
L185:
	.dbline 0 ; func end
	adiw R28,8
	ld R11,y+
	ld R10,y+
	ret
;             rc -> R10,R11
;          found -> y+2
;         status -> R20,R21
;           path -> R10,R11
	.even
_PnutStatus:
	xcall push_xgsetF00C
	movw R20,R18
	movw R10,R16
	sbiw R28,8
; }
; 
; /*!
;  * \brief Return information about a specified file entry.
;  *
;  * \param path   Pointer to the full pathname of the file.
;  * \param status Pointer to a stat structure for the information returned.
;  *
;  * \return 0 on success. Otherwise -1 is returned.
;  */
; static int PnutStatus(char *path, struct stat *status)
; {
;     int rc;
;     PNUT_FINDRESULT found;
; 
;     if ((rc = PnutDirFindPath(root, path, &found)) == 0) {
	movw R24,R28
	adiw R24,2
	std y+1,R25
	std y+0,R24
	movw R18,R10
	lds R16,_root
	lds R17,_root+1
	xcall _PnutDirFindPath
	movw R10,R16
	tst R10
	breq X96
	xjmp L193
X96:
	tst R11
	breq X97
	xjmp L193
X97:
X95:
;         status->st_mode = BankNodePointer(found.fr_node)->node_type;
	ldd R16,y+2
	ldd R17,y+3
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+0
	movw R30,R20
	std z+2,R2
;         status->st_ino = found.fr_node;
	ldd R2,y+2
	ldd R3,y+3
	std z+1,R3
	std z+0,R2
;         status->st_nlink = BankNodePointer(found.fr_node)->node_links;
	movw R16,R2
	xcall _BankNodePointer
	movw R22,R16
	movw R30,R22
	ldd R2,z+2
	ldd R3,z+3
	movw R30,R20
	std z+4,R3
	std z+3,R2
;         status->st_size = BankNodePointer(found.fr_node)->node_size;
	ldd R16,y+2
	ldd R17,y+3
	xcall _BankNodePointer
	movw R22,R16
	movw R30,R22
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	movw R30,R20
	std z+5,R2
	std z+6,R3
	std z+7,R4
	std z+8,R5
;         status->st_mtime = BankNodePointer(found.fr_node)->node_mtime;
	ldd R16,y+2
	ldd R17,y+3
	xcall _BankNodePointer
	movw R22,R16
	movw R30,R22
	ldd R2,z+8
	ldd R3,z+9
	ldd R4,z+10
	ldd R5,z+11
	movw R30,R20
	std z+9,R2
	std z+10,R3
	std z+11,R4
	std z+12,R5
;     }
L193:
;     return rc;
	movw R16,R10
L192:
	.dbline 0 ; func end
	adiw R28,8
	xjmp pop_xgsetF00C
;         status -> R20,R21
;             fp -> R10,R11
	.even
_PnutFileStatus:
	xcall push_xgsetF00C
	movw R20,R18
	movw R10,R16
; }
; 
; /*!
;  * \brief Return information about a previously opened file.
;  *
;  * \param nfp    Pointer to a \ref NUTFILE structure, obtained by a previous 
;  *               call to PnutFileOpen().
;  * \param status Pointer to a stat structure for the information returned.
;  *
;  * \return 0 on success. Otherwise -1 is returned.
;  */
; static int PnutFileStatus(PNUTFILE * fp, struct stat *status)
; {
;     status->st_mode = BankNodePointer(fp->f_node)->node_type;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+0
	movw R30,R20
	std z+2,R2
;     status->st_ino = fp->f_node;
	movw R30,R10
	ldd R2,z+0
	ldd R3,z+1
	movw R30,R20
	std z+1,R3
	std z+0,R2
;     status->st_nlink = BankNodePointer(fp->f_node)->node_links;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+2
	ldd R3,z+3
	movw R30,R20
	std z+4,R3
	std z+3,R2
;     status->st_size = BankNodePointer(fp->f_node)->node_size;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R22,R16
	movw R30,R22
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	movw R30,R20
	std z+5,R2
	std z+6,R3
	std z+7,R4
	std z+8,R5
;     status->st_mtime = BankNodePointer(fp->f_node)->node_mtime;
	movw R30,R10
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R22,R16
	movw R30,R22
	ldd R2,z+8
	ldd R3,z+9
	ldd R4,z+10
	ldd R5,z+11
	movw R30,R20
	std z+9,R2
	std z+10,R3
	std z+11,R4
	std z+12,R5
; 
;     return 0;
	clr R16
	clr R17
L195:
	.dbline 0 ; func end
	xjmp pop_xgsetF00C
;             rc -> R10,R11
;           node -> R12,R13
;             ec -> R20,R21
;         blkptr -> y+10
;            buf -> R14,R15
;             fp -> R22,R23
;         blksiz -> y+8
;            len -> y+22
;         buffer -> R14,R15
;            nfp -> R20,R21
	.even
_PnutFileWrite:
	xcall push_xgsetF0FC
	movw R14,R18
	movw R20,R16
	sbiw R28,12
; }
; 
; /*!
;  * \brief Write data to a file.
;  *
;  * \param nfp    Pointer to a \ref NUTFILE structure, obtained by a previous 
;  *               call to PnutFileOpen().
;  * \param buffer Pointer to the data to be written. If zero, then the
;  *               output buffer will be flushed.
;  * \param len    Number of bytes to write.
;  *
;  * \return The number of bytes written. A return value of -1 indicates an 
;  *         error.
;  */
; static int PnutFileWrite(NUTFILE * nfp, CONST void *buffer, int len)
; {
;     PNUTFILE *fp = nfp->nf_fcb;
	movw R30,R20
	ldd R22,z+4
	ldd R23,z+5
;     int ec = 0;
	clr R20
	clr R21
;     int rc = 0;
	clr R10
	clr R11
;     PNUT_BLKNUM node = fp->f_node;
	movw R30,R22
	ldd R12,z+0
	ldd R13,z+1
;     u_char *blkptr;
;     size_t blksiz;
;     CONST char *buf = buffer;
	xjmp L198
L197:
; 
;     while (len) {
;         if ((ec = PnutNodeGetDataPtr(node, fp->f_pos, (void *) &blkptr, &blksiz, 1)) != 0) {
	ldi R24,1
	ldi R25,0
	std y+7,R25
	std y+6,R24
	movw R24,R28
	adiw R24,8
	std y+5,R25
	std y+4,R24
	movw R24,R28
	adiw R24,10
	std y+3,R25
	std y+2,R24
	movw R30,R22
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	std y+0,R4
	std y+1,R5
	movw R18,R2
	movw R16,R12
	xcall _PnutNodeGetDataPtr
	movw R20,R16
	cpi R20,0
	cpc R20,R21
	breq L200
X98:
;             break;
	xjmp L199
L200:
;         }
;         if (blksiz >= len) {
	ldd R2,y+8
	ldd R3,y+9
	ldd R0,y+22
	ldd R1,y+23
	cp R2,R0
	cpc R3,R1
	brlo L202
X99:
;             blksiz = len;
	std y+9,R1
	std y+8,R0
;             len = 0;
	clr R0
	clr R1
	std y+23,R1
	std y+22,R0
;         } else {
	xjmp L203
L202:
;             len -= blksiz;
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+22
	ldd R5,y+23
	sub R4,R2
	sbc R5,R3
	std y+23,R5
	std y+22,R4
;         }
L203:
;         memcpy(blkptr, buf, blksiz);
	ldd R2,y+8
	ldd R3,y+9
	std y+1,R3
	std y+0,R2
	movw R18,R14
	ldd R16,y+10
	ldd R17,y+11
	xcall _memcpy
;         rc += blksiz;
	ldd R2,y+8
	ldd R3,y+9
	add R10,R2
	adc R11,R3
;         buf += blksiz;
	add R14,R2
	adc R15,R3
;         fp->f_pos += blksiz;
	movw R24,R22
	adiw R24,2
	clr R4
	clr R5
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;     }
L198:
	ldd R0,y+22
	ldd R1,y+23
	tst R0
	breq X104
	xjmp L197
X104:
	tst R1
	breq X105
	xjmp L197
X105:
X100:
L199:
;     if (ec == 0 || ec == ENOSPC) {
	cpi R20,0
	cpc R20,R21
	breq L206
X101:
	cpi R20,28
	ldi R30,0
	cpc R21,R30
	breq X106
	xjmp L204
X106:
X102:
L206:
;         if (BankNodePointer(node)->node_size < fp->f_pos) {
	movw R16,R12
	xcall _BankNodePointer
	movw R20,R16
	movw R30,R22
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	movw R30,R20
	ldd R6,z+4
	ldd R7,z+5
	ldd R8,z+6
	ldd R9,z+7
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh L207
X103:
;             BankNodePointer(node)->node_size = fp->f_pos;
	movw R16,R12
	xcall _BankNodePointer
	movw R20,R16
	movw R30,R22
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	movw R30,R20
	std z+4,R2
	std z+5,R3
	std z+6,R4
	std z+7,R5
;         }
L207:
;         BankNodePointer(node)->node_mtime = time(0);
	movw R16,R12
	xcall _BankNodePointer
	movw R20,R16
	clr R16
	clr R17
	xcall _time
	movw R30,R20
	std z+8,R16
	std z+9,R17
	std z+10,R18
	std z+11,R19
;     }
L204:
;     return rc;
	movw R16,R10
L196:
	.dbline 0 ; func end
	adiw R28,12
	xjmp pop_xgsetF0FC
;            len -> y+0
;         buffer -> R18,R19
;            nfp -> R16,R17
	.even
_PnutFileWrite_P:
; }
; 
; #ifdef __HARVARD_ARCH__
; static int PnutFileWrite_P(NUTFILE * nfp, PGM_P buffer, int len)
; {
;     return -1;
	ldi R16,-1
	ldi R17,-1
L209:
	.dbline 0 ; func end
	ret
;             ec -> R20,R21
;             rc -> R10,R11
;           node -> R14,R15
;         blkptr -> y+10
;            buf -> R22,R23
;             fp -> y+12
;         blksiz -> y+8
;            len -> y+24
;         buffer -> R12,R13
;            nfp -> R20,R21
	.even
_PnutFileRead:
	xcall push_xgsetF0FC
	movw R12,R18
	movw R20,R16
	sbiw R28,14
; }
; #endif
; 
; /*!
;  * \brief Read data from a file.
;  *
;  * \param nfp    Pointer to a ::NUTFILE structure, obtained by a previous 
;  *               call to PnutFileOpen().
;  * \param buffer Pointer to the data buffer to fill.
;  * \param len    Maximum number of bytes to read.
;  *
;  * \return The number of bytes actually read. A return value of -1 indicates 
;  *         an error.
;  */
; static int PnutFileRead(NUTFILE * nfp, void *buffer, int len)
; {
;     PNUTFILE *fp = nfp->nf_fcb;
	movw R30,R20
	ldd R2,z+4
	ldd R3,z+5
	std y+13,R3
	std y+12,R2
;     int ec = 0;
	clr R20
	clr R21
;     int rc = 0;
	clr R10
	clr R11
;     PNUT_BLKNUM node = fp->f_node;
	movw R30,R2
	ldd R14,z+0
	ldd R15,z+1
;     u_char *blkptr;
;     size_t blksiz;
;     char *buf = buffer;
	movw R22,R12
; 
;     /* Respect end of file. */
;     if (len > BankNodePointer(node)->node_size - fp->f_pos) {
	movw R16,R14
	xcall _BankNodePointer
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	movw R30,R16
	ldd R6,z+4
	ldd R7,z+5
	ldd R8,z+6
	ldd R9,z+7
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	ldd R2,y+24
	ldd R3,y+25
	clr R4
	sbrc R3,7
	com R4
	clr R5
	sbrc R4,7
	com R5
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brlo X111
	xjmp L214
X111:
X107:
;         len = (size_t) (BankNodePointer(node)->node_size - fp->f_pos);
	movw R16,R14
	xcall _BankNodePointer
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+2
	ldd R3,z+3
	movw R30,R16
	ldd R4,z+4
	ldd R5,z+5
	sub R4,R2
	sbc R5,R3
	std y+25,R5
	std y+24,R4
;     }
	xjmp L214
L213:
;     while (len) {
;         if ((ec = PnutNodeGetDataPtr(node, fp->f_pos, (void *) &blkptr, &blksiz, 0)) != 0) {
	clr R2
	clr R3
	std y+7,R3
	std y+6,R2
	movw R24,R28
	adiw R24,8
	std y+5,R25
	std y+4,R24
	movw R24,R28
	adiw R24,10
	std y+3,R25
	std y+2,R24
	ldd R30,y+12
	ldd R31,y+13
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	std y+0,R4
	std y+1,R5
	movw R18,R2
	movw R16,R14
	xcall _PnutNodeGetDataPtr
	movw R12,R16
	movw R20,R12
	cpi R16,0
	cpc R16,R17
	breq L216
X108:
;             break;
	xjmp L215
L216:
;         }
;         if (blksiz >= len) {
	ldd R2,y+8
	ldd R3,y+9
	ldd R0,y+24
	ldd R1,y+25
	cp R2,R0
	cpc R3,R1
	brlo L218
X109:
;             blksiz = len;
	std y+9,R1
	std y+8,R0
;             len = 0;
	clr R0
	clr R1
	std y+25,R1
	std y+24,R0
;         } else {
	xjmp L219
L218:
;             len -= blksiz;
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+24
	ldd R5,y+25
	sub R4,R2
	sbc R5,R3
	std y+25,R5
	std y+24,R4
;         }
L219:
;         memcpy(buf, blkptr, blksiz);
	ldd R2,y+8
	ldd R3,y+9
	std y+1,R3
	std y+0,R2
	ldd R18,y+10
	ldd R19,y+11
	movw R16,R22
	xcall _memcpy
;         rc += blksiz;
	ldd R2,y+8
	ldd R3,y+9
	add R10,R2
	adc R11,R3
;         buf += blksiz;
	add R22,R2
	adc R23,R3
;         fp->f_pos += blksiz;
	ldd R24,y+12
	ldd R25,y+13
	adiw R24,2
	movw R12,R24
	clr R4
	clr R5
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;     }
L214:
	ldd R0,y+24
	ldd R1,y+25
	tst R0
	breq X112
	xjmp L213
X112:
	tst R1
	breq X113
	xjmp L213
X113:
X110:
L215:
;     return rc;
	movw R16,R10
L210:
	.dbline 0 ; func end
	adiw R28,14
	xjmp pop_xgsetF0FC
;             rc -> R10,R11
;           npos -> y+0
;         whence -> R14,R15
;            pos -> R12,R13
;             fp -> y+14
	.even
_PnutFileSeek:
	xcall push_arg4
	xcall push_xgsetF0FC
	movw R12,R18
	sbiw R28,4
	ldd R14,y+18
	ldd R15,y+19
; }
; 
; static int PnutFileSeek(PNUTFILE * fp, long *pos, int whence)
; {
;     int rc = 0;
	clr R10
	clr R11
;     long npos = *pos;
	movw R30,R12
	ldd R2,z+0
	ldd R3,z+1
	ldd R4,z+2
	ldd R5,z+3
	std y+0,R2
	std y+1,R3
	std y+2,R4
	std y+3,R5
; 
;     switch (whence) {
	movw R24,R14
	cpi R24,1
	ldi R30,0
	cpc R25,R30
	breq L223
X114:
	cpi R24,2
	ldi R30,0
	cpc R25,R30
	breq L224
X115:
	xjmp L221
L223:
;     case SEEK_CUR:
;         npos += fp->f_pos;
	ldd R30,y+14
	ldd R31,y+15
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std y+0,R6
	std y+1,R7
	std y+2,R8
	std y+3,R9
;         break;
	xjmp L222
L224:
;     case SEEK_END:
;         npos += BankNodePointer(fp->f_node)->node_size;
	ldd R30,y+14
	ldd R31,y+15
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std y+0,R6
	std y+1,R7
	std y+2,R8
	std y+3,R9
;         break;
L221:
L222:
;     }
; 
;     if (npos < 0 || npos > (long) BankNodePointer(fp->f_node)->node_size) {
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+0
	ldd R3,y+1
	ldd R4,y+2
	ldd R5,y+3
	cp R2,R20
	cpc R3,R21
	cpc R4,R22
	cpc R5,R23
	brlt L227
X116:
	ldd R30,y+14
	ldd R31,y+15
	ldd R16,z+0
	ldd R17,z+1
	xcall _BankNodePointer
	movw R30,R16
	ldd R2,z+4
	ldd R3,z+5
	ldd R4,z+6
	ldd R5,z+7
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	cp R2,R6
	cpc R3,R7
	cpc R4,R8
	cpc R5,R9
	brge L225
X117:
L227:
;         rc = EINVAL;
	ldi R24,22
	ldi R25,0
	movw R10,R24
;     } else {
	xjmp L226
L225:
;         fp->f_pos = npos;
	ldd R2,y+0
	ldd R3,y+1
	ldd R4,y+2
	ldd R5,y+3
	ldd R30,y+14
	ldd R31,y+15
	std z+2,R2
	std z+3,R3
	std z+4,R4
	std z+5,R5
;         *pos = npos;
	movw R30,R12
	std z+0,R2
	std z+1,R3
	std z+2,R4
	std z+3,R5
;     }
L226:
;     return rc;
	movw R16,R10
L220:
	.dbline 0 ; func end
	adiw R28,4
	xcall pop_xgsetF0FC
	adiw R28,4
	ret
;             rc -> R20,R21
;           conf -> R22,R23
;            req -> R10,R11
;            dev -> R12,R13
	.even
_PnutIOCtl::
	xcall push_xgsetF03C
	movw R10,R18
	movw R12,R16
	sbiw R28,2
	ldd R22,y+10
	ldd R23,y+11
; }
; 
; /* ------------------------ Special function interface ------------------------ */
; 
; /*!
;  * \brief Device specific functions.
;  */
; int PnutIOCtl(NUTDEVICE * dev, int req, void *conf)
; {
;     int rc = -1;
	ldi R20,-1
	ldi R21,-1
; 
;     switch (req) {
	movw R24,R10
	cpi R24,17
	ldi R30,17
	cpc R25,R30
	brne X128
	xjmp L232
X128:
X118:
	cpi R24,18
	ldi R30,17
	cpc R25,R30
	brne X129
	xjmp L233
X129:
X119:
	cpi R24,19
	ldi R30,17
	cpc R25,R30
	brne X130
	xjmp L234
X130:
X120:
	cpi R24,20
	ldi R30,17
	cpc R25,R30
	brne X131
	xjmp L235
X131:
X121:
	cpi R24,21
	ldi R30,17
	cpc R25,R30
	brne X132
	xjmp L236
X132:
X122:
	ldi R24,4373
	ldi R25,17
	cp R24,R10
	cpc R25,R11
	brlt L241
X123:
L240:
	movw R24,R10
	cpi R24,1
	ldi R30,17
	cpc R25,R30
	breq L231
X124:
	xjmp L229
L241:
	movw R24,R10
	cpi R24,33
	ldi R30,17
	cpc R25,R30
	brne X133
	xjmp L237
X133:
X125:
	cpi R24,34
	ldi R30,17
	cpc R25,R30
	brne X134
	xjmp L238
X134:
X126:
	cpi R24,35
	ldi R30,17
	cpc R25,R30
	brne X135
	xjmp L239
X135:
X127:
	xjmp L229
L231:
;     case FS_STATUS:
;         rc = PnutStatus((char *) ((IOCTL_ARG2 *) conf)->arg1,   /* */
	movw R30,R22
	ldd R18,z+2
	ldd R19,z+3
	ldd R16,z+0
	ldd R17,z+1
	xcall _PnutStatus
	movw R20,R16
;                         (struct stat *) ((IOCTL_ARG2 *) conf)->arg2);
;         break;
	xjmp L230
L232:
;     case FS_DIR_CREATE:
;         rc = PnutDirCreate((char *) conf);
	movw R16,R22
	xcall _PnutDirCreate
	movw R20,R16
;         break;
	xjmp L230
L233:
;     case FS_DIR_REMOVE:
;         rc = PnutDelete((char *) conf);
	movw R16,R22
	xcall _PnutDelete
	movw R10,R16
	movw R20,R10
;         break;
	xjmp L230
L234:
;     case FS_DIR_OPEN:
;         rc = PnutDirOpen((DIR *) conf);
	movw R16,R22
	xcall _PnutDirOpen
	movw R10,R16
	movw R20,R10
;         break;
	xjmp L230
L235:
;     case FS_DIR_CLOSE:
;         rc = PnutDirClose((DIR *) conf);
	movw R16,R22
	xcall _PnutDirClose
	movw R10,R16
	movw R20,R10
;         break;
	xjmp L230
L236:
;     case FS_DIR_READ:
;         rc = PnutDirRead((DIR *) conf);
	movw R16,R22
	xcall _PnutDirRead
	movw R10,R16
	movw R20,R10
;         break;
	xjmp L230
L237:
;     case FS_FILE_STATUS:
;         rc = PnutFileStatus((PNUTFILE *) ((IOCTL_ARG2 *) conf)->arg1,   /* */
	movw R30,R22
	ldd R18,z+2
	ldd R19,z+3
	ldd R16,z+0
	ldd R17,z+1
	xcall _PnutFileStatus
	movw R10,R16
	movw R20,R10
;                             (struct stat *) ((IOCTL_ARG2 *) conf)->arg2);
;         break;
	xjmp L230
L238:
;     case FS_FILE_DELETE:
;         rc = PnutDelete((char *) conf);
	movw R16,R22
	xcall _PnutDelete
	movw R10,R16
	movw R20,R10
;         break;
	xjmp L230
L239:
;     case FS_FILE_SEEK:
;         PnutFileSeek((PNUTFILE *) ((IOCTL_ARG3 *) conf)->arg1,  /* */
	movw R30,R22
	ldd R2,z+4
	ldd R3,z+5
	std y+1,R3
	std y+0,R2
	ldd R18,z+2
	ldd R19,z+3
	ldd R16,z+0
	ldd R17,z+1
	xcall _PnutFileSeek
;                      (long *) ((IOCTL_ARG3 *) conf)->arg2,      /* */
;                      (int) ((IOCTL_ARG3 *) conf)->arg3);
;         break;
L229:
L230:
;     }
;     return rc;
	movw R16,R20
L228:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgsetF03C
;             rc -> R20,R21
;              i -> R20,R21
;            dev -> R10,R11
	.even
_PnutInit:
	xcall push_xgset300C
	movw R10,R16
	sbiw R28,2
; }
; 
; /* ------------------------ Initialization ------------------------ */
; 
; /*!
;  * \brief Initialize the Peanut filesystem.
;  *
;  * Peanut is a volatile RAM filesystem. All contents is lost when
;  * power supply is removed. After initialization we start with
;  * a clean filesystem.
;  *
;  * \param Pointer to the device information structure.
;  *
;  * \return Zero on success. Otherwise an error code is returned.
;  */
; static int PnutInit(NUTDEVICE * dev)
; {
;     PNUT_BLKNUM i;
;     int rc;
; 
;     /* Add all blocks to the free list. */
;     for (i = 0; i < PNUT_TOTAL_BLOCKS; i++) {
	clr R20
	clr R21
L243:
;         PnutBlockRelease(i);
	movw R16,R20
	xcall _PnutBlockRelease
;     }
L244:
	subi R20,255  ; offset = 1
	sbci R21,255
	cpi R20,192
	ldi R30,3
	cpc R21,R30
	brlt L243
X136:
; 
;     /* Initialize the root directory. */
;     if ((root = PnutNodeAlloc(NODETYPE_DIR)) == -1) {
	ldi R16,1
	xcall _PnutNodeAlloc
	sts _root+1,R17
	sts _root,R16
	cpi R16,255
	ldi R30,255
	cpc R17,R30
	brne L247
X137:
;         rc = ENOSPC;
	ldi R20,28
	ldi R21,0
;     } else {
	xjmp L248
L247:
;         if ((rc = PnutDirAddEntry(root, ".", root)) == 0) {
	lds R2,_root
	lds R3,_root+1
	std y+1,R3
	std y+0,R2
	ldi R18,<L50
	ldi R19,>L50
	movw R16,R2
	xcall _PnutDirAddEntry
	movw R10,R16
	movw R20,R10
	cpi R16,0
	cpc R16,R17
	brne L249
X138:
;             rc = PnutDirAddEntry(root, "..", root);
	lds R10,_root
	lds R11,_root+1
	std y+1,R11
	std y+0,R10
	ldi R18,<L51
	ldi R19,>L51
	movw R16,R10
	xcall _PnutDirAddEntry
	movw R10,R16
	movw R20,R10
;         }
L249:
;         if (rc) {
	cpi R20,0
	cpc R20,R21
	breq L251
X139:
;             PnutBlockRelease(root);
	lds R16,_root
	lds R17,_root+1
	xcall _PnutBlockRelease
;         }
L251:
;     }
L248:
;     return rc;
	movw R16,R20
L242:
	.dbline 0 ; func end
	adiw R28,2
	xjmp pop_xgset300C
	.area data(ram, con, rel)
_devPnut::
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 'P,'N
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 'U,'T
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.byte 0,0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 1
	.area idata
	.byte 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutInit
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutIOCtl
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutFileRead
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutFileWrite
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutFileWrite_P
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutFileOpen
	.area data(ram, con, rel)
	.blkw 1
	.area idata
	.word PL_PnutFileClose
	.area data(ram, con, rel)
	.blkb 2
	.area idata
	.word 0
	.area data(ram, con, rel)
	.area bss(ram, con, rel)
_root:
	.blkb 2
	.area data(ram, con, rel)
L51:
	.blkb 3
	.area idata
	.byte 46,46,0
	.area data(ram, con, rel)
L50:
	.blkb 2
	.area idata
	.byte 46,0
	.area data(ram, con, rel)
; }
; 
; /*!
;  * \brief Peanut device information structure.
;  */
; NUTDEVICE devPnut = {
;     0,                          /*!< Pointer to next device, dev_next. */
;     {'P', 'N', 'U', 'T', 0, 0, 0, 0, 0}
;     ,                           /*!< Unique device name, dev_name. */
;     IFTYP_RAM,                  /*!< Type of device, dev_type. Obsolete. */
;     0,                          /*!< Base address, dev_base. Unused. */
;     0,                          /*!< First interrupt number, dev_irq. Unused. */
;     0,                          /*!< Interface control block, dev_icb. Unused. */
;     0,                          /*!< Driver control block, dev_dcb. Unused. */
;     PnutInit,                   /*!< Driver initialization routine, dev_init. */
;     PnutIOCtl,                  /*!< Driver specific control function, dev_ioctl. */
;     PnutFileRead,               /*!< Read data from a file, dev_read. */
;     PnutFileWrite,              /*!< Write data to a file, dev_write. */
; #ifdef __HARVARD_ARCH__
;     PnutFileWrite_P,            /*!< Write data from program space to a file, dev_write_P. */
; #endif
;     PnutFileOpen,               /*!< Open a file, dev_open. */
;     PnutFileClose,              /*!< Close a file, dev_close. */
;     0                           /*!< Return file size, dev_size. */
; };
; 
; /*@}*/
	.area func_lit
PL_PnutFileClose:	.word `_PnutFileClose
PL_PnutFileOpen:	.word `_PnutFileOpen
PL_PnutFileWrite_P:	.word `_PnutFileWrite_P
PL_PnutFileWrite:	.word `_PnutFileWrite
PL_PnutFileRead:	.word `_PnutFileRead
PL_PnutIOCtl:	.word `_PnutIOCtl
PL_PnutInit:	.word `_PnutInit
