	.module mutex.c
	.area text(rom, con, rel)
;          mutex -> R16,R17
	.even
_NutMutexInit::
; /*
;  * Copyright (C) 2000-2004 by ETH Zurich
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
;  * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  */
; 
; /* mutex.c - a nut/os implementation of recursive mutex functions
;  *
;  * 2004.05.06 Matthias Ringwald <matthias.ringwald@inf.ethz.ch>
;  *
;  */
; /*
;  * $Log: mutex.c,v $
;  * Revision 1.4  2005/08/02 17:47:04  haraldkipp
;  * Major API documentation update.
;  *
;  * Revision 1.3  2004/07/19 16:24:23  freckle
;  * Code contained same bug as os/semaphore.c.
;  * Unfortunately I didn't understand the other fix, but anyway made
;  * NutMutexLock more robust.
;  * Fixed wrong indention of all functions
;  *
;  * Revision 1.2  2004/05/18 18:38:42  drsung
;  * Added $Log keyword for CVS.
;  *
;  */
; 
; /*!
;  * \addtogroup xgMutex
;  */
; /*@{*/
; 
; #ifdef __cplusplus
; extern "C" {
; #endif
; 
; #include <sys/mutex.h>
; #include <sys/event.h>
; 
; /*!
;  * \brief Create a mutex
;  *
;  * The type for the mutex is recursive
;  */
;  
; void NutMutexInit(MUTEX * mutex) {
;     mutex->thread = 0;
	clr R2
	clr R3
	movw R30,R16
	std z+3,R3
	std z+2,R2
;     mutex->count = 0;
	std z+5,R3
	std z+4,R2
;     mutex->qhp = 0;
	std z+1,R3
	std z+0,R2
L1:
	.dbline 0 ; func end
	ret
;          mutex -> R20,R21
	.even
_NutMutexLock::
	st -y,R20
	st -y,R21
	movw R20,R16
	sbiw R28,2
; }
; 
; /*!
;  * \brief Lock a mutex
;  *
;  * If the mutex is already locked by another thread,
;  * the thread will block until the mutex becomes available
;  *
;  * \Note: Should not be called from interrupt context
;  */
;  
; void NutMutexLock(MUTEX * mutex) {
;     if (mutex->thread != runningThread) {
	lds R2,_runningThread
	lds R3,_runningThread+1
	movw R30,R20
	ldd R4,z+2
	ldd R5,z+3
	cp R4,R2
	cpc R5,R3
	breq L3
X0:
	xjmp L6
L5:
;         while( mutex->count != 0)
;             NutEventWaitNext(&mutex->qhp, NUT_WAIT_INFINITE);
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	movw R16,R20
	xcall _NutEventWaitNext
L6:
	movw R30,R20
	ldd R2,z+4
	ldd R3,z+5
	tst R2
	brne L5
	tst R3
	brne L5
X1:
;     }
L3:
;     mutex->thread = runningThread;
	lds R2,_runningThread
	lds R3,_runningThread+1
	movw R30,R20
	std z+3,R3
	std z+2,R2
;     mutex->count++;
	movw R24,R30
	adiw R24,4
	movw R30,R24
	ldd R24,z+0
	ldd R25,z+1
	adiw R24,1
	std z+1,R25
	std z+0,R24
L2:
	.dbline 0 ; func end
	adiw R28,2
	ld R21,y+
	ld R20,y+
	ret
;          mutex -> R20,R21
	.even
_NutMutexTrylock::
	st -y,R20
	st -y,R21
	movw R20,R16
; }
; 
; /*!
;  * \brief Attempt to lock a mutex without blocking
;  *
;  * Return zero, if successful, otherwise the mutex is already locked
;  * by another thread
;  * \Note: Should not be called from interrupt context
;  */
;  
; int NutMutexTrylock(MUTEX * mutex) {
;     if ((mutex->count != 0) && (mutex->thread != runningThread))
	movw R30,R20
	ldd R2,z+4
	ldd R3,z+5
	tst R2
	brne X2
	tst R3
	breq L9
X2:
	lds R2,_runningThread
	lds R3,_runningThread+1
	ldd R4,z+2
	ldd R5,z+3
	cp R4,R2
	cpc R5,R3
	breq L9
X3:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L8
L9:
;     NutMutexLock(mutex);
	movw R16,R20
	xcall _NutMutexLock
;     return 0;
	clr R16
	clr R17
L8:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;          mutex -> R20,R21
	.even
_NutMutexUnlock::
	st -y,R20
	st -y,R21
	movw R20,R16
; }
; 
; /*!
;  * \brief Unlock a mutex.
;  *
;  * Return zero, if successful, otherwise the current thread does not
;  * hold a lock on mutex.
;  * \Note: Should not be called from interrupt context
;  */
;  
; int NutMutexUnlock(MUTEX * mutex) {
;     if (mutex->thread != runningThread)
	lds R2,_runningThread
	lds R3,_runningThread+1
	movw R30,R20
	ldd R4,z+2
	ldd R5,z+3
	cp R4,R2
	cpc R5,R3
	breq L12
X4:
;         return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L11
L12:
;     if (--mutex->count == 0) {
	movw R24,R20
	adiw R24,4
	movw R30,R24
	ldd R24,z+0
	ldd R25,z+1
	sbiw R24,1
	std z+1,R25
	std z+0,R24
	cpi R24,0
	cpc R24,R25
	brne L14
X5:
;         NutEventPost(&mutex->qhp);
	movw R16,R20
	xcall _NutEventPost
;     }
L14:
;     return 0;
	clr R16
	clr R17
L11:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;          mutex -> R16,R17
	.even
_NutMutexDestroy::
; }
; 
; /*!
;  * \brief Free resources allocated for a mutex
;  *
;  * Return zero, if successful, otherwise the mutex is locked
;  * by another thread
;  */
; 
; int NutMutexDestroy(MUTEX * mutex) {
;     if (mutex->count == 0)
	movw R30,R16
	ldd R2,z+4
	ldd R3,z+5
	tst R2
	brne L17
	tst R3
	brne L17
X6:
;         return 0;
	clr R16
	clr R17
	xjmp L16
L17:
;     if (mutex->thread == runningThread)
	lds R2,_runningThread
	lds R3,_runningThread+1
	movw R30,R16
	ldd R4,z+2
	ldd R5,z+3
	cp R4,R2
	cpc R5,R3
	brne L19
X7:
;         return 0;
	clr R16
	clr R17
	xjmp L16
L19:
;     return -1;
	ldi R16,-1
	ldi R17,-1
L16:
	.dbline 0 ; func end
	ret
; }
; 
; #ifdef __cplusplus
; }
; #endif
; 
; /*@}*/
