	.module semaphore.c
	.area text(rom, con, rel)
;          value -> R18,R19
;            sem -> R16,R17
	.even
_NutSemInit::
; /*
;  * Copyright (C) 2000-2004 by ETH Zurich
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
;  * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  */
; 
; /* semaphore.c - a nut/os implementation of semaphore functions
;  *
;  * 2004.05.06 Matthias Ringwald <matthias.ringwald@inf.ethz.ch>
;  *
;  */
; /*
;  * $Log: semaphore.c,v $
;  * Revision 1.6  2005/08/02 17:47:04  haraldkipp
;  * Major API documentation update.
;  *
;  * Revision 1.5  2004/06/03 08:44:50  olereinhardt
;  * According to a hint from oliver I changed NutEventWait to NutEventWaitNext
;  *
;  * Revision 1.4  2004/06/03 08:24:21  olereinhardt
;  * Changed semaphore behavior in NutSemTryWait too.
;  *
;  * Revision 1.3  2004/06/02 16:42:53  olereinhardt
;  * fixed bug (integer overflow) in semaphore implementation.
;  *
;  * Revision 1.2  2004/05/18 18:38:42  drsung
;  * Added $Log keyword for CVS.
;  *
;  */
; 
; /*!
;  * \addtogroup xgSemaphore
;  */
; /*@{*/
; 
; #ifdef __cplusplus
; extern "C" {
; #endif
; 
; #include <sys/semaphore.h>
; #include <sys/event.h>
; 
; /*!
;  * \brief Initialize an unnamed semaphore to value
;  */
;     void NutSemInit(SEM * sem, short value) {
;         sem->qhp = 0;
	clr R2
	clr R3
	movw R30,R16
	std z+1,R3
	std z+0,R2
;         sem->value = value;
	std z+3,R19
	std z+2,R18
L1:
	.dbline 0 ; func end
	ret
;            sem -> R20,R21
	.even
_NutSemWait::
	st -y,R20
	st -y,R21
	movw R20,R16
	sbiw R28,2
;     }
; /*!
;  * \brief Lock a semaphore
;  *
;  *  If the semaphore value is currently zero, then the calling thread will not
;  *  return from the call to sem_wait() the semaphore becomes available
;  *
;  * \Note: Should not be called from interrupt context
;  */ void NutSemWait(SEM * sem) {
;         sem->value--;
	movw R24,R20
	adiw R24,2
	movw R30,R24
	ldd R24,z+0
	ldd R25,z+1
	sbiw R24,1
	std z+1,R25
	std z+0,R24
;         if (sem->value < 0)
	movw R30,R20
	ldd R24,z+2
	ldd R25,z+3
	cpi R24,0
	ldi R26,0
	cpc R25,R26
	brge L3
X0:
;         {
;             NutEventWaitNext(&sem->qhp, NUT_WAIT_INFINITE);
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std y+0,R26
	std y+1,R27
	movw R18,R24
	movw R16,R30
	xcall _NutEventWaitNext
;         }
L3:
L2:
	.dbline 0 ; func end
	adiw R28,2
	ld R21,y+
	ld R20,y+
	ret
;            sem -> R20,R21
	.even
_NutSemPost::
	st -y,R20
	st -y,R21
	movw R20,R16
;     }
; 
; /*!
;  * \brief Unlock a sempahore.
;  *
;  * \Note: Should not be called from interrupt context
;  */
;     void NutSemPost(SEM * sem) {
;         sem->value++;
	movw R24,R20
	adiw R24,2
	movw R30,R24
	ldd R24,z+0
	ldd R25,z+1
	adiw R24,1
	std z+1,R25
	std z+0,R24
;         if (sem->value <= 0)
	clr R2
	clr R3
	movw R30,R20
	ldd R4,z+2
	ldd R5,z+3
	cp R2,R4
	cpc R3,R5
	brlt L6
X1:
;         {
;             NutEventPost(&sem->qhp);
	movw R16,R30
	xcall _NutEventPost
;         }
L6:
L5:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;            sem -> R20,R21
	.even
_NutSemTryWait::
	st -y,R20
	st -y,R21
	movw R20,R16
;     }
; 
; /*!
;  * \brief Attempt to lock a semaphore without blocking
;  *
;  * Return zero, if successful, otherwise the sempahore is already locked
;  * \Note: Should not be called from interrupt context
;  */
; 
;     int NutSemTryWait(SEM * sem) {
;         if (sem->value < 0)
	movw R30,R20
	ldd R24,z+2
	ldd R25,z+3
	cpi R24,0
	ldi R30,0
	cpc R25,R30
	brge L9
X2:
;             return -1;
	ldi R16,-1
	ldi R17,-1
	xjmp L8
L9:
;         else
;             NutSemWait(sem);
	movw R16,R20
	xcall _NutSemWait
;         return 0;
	clr R16
	clr R17
L8:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;            sem -> R16,R17
	.even
_NutSemDestroy::
;     }
; 
; /*!
;  * \brief Free resources allocated for a semaphore
;  *
;  * Return zero, if successful, otherwise there are threads blocked on the
;  * sempahore
;  */
; 
;     int NutSemDestroy(SEM * sem) {
;         if (sem->qhp == SIGNALED)
	movw R30,R16
	ldd R24,z+0
	ldd R25,z+1
	cpi R24,255
	ldi R30,255
	cpc R25,R30
	brne L12
X3:
;             return 0;
	clr R16
	clr R17
	xjmp L11
L12:
;         if (sem->qhp == 0)
	movw R30,R16
	ldd R2,z+0
	ldd R3,z+1
	tst R2
	brne L14
	tst R3
	brne L14
X4:
;             return 0;
	clr R16
	clr R17
	xjmp L11
L14:
;         return -1;
	ldi R16,-1
	ldi R17,-1
L11:
	.dbline 0 ; func end
	ret
;     }
; 
; #ifdef __cplusplus
; }
; #endif
; 
; /*@}*/
