	.module thread.c
	.area text(rom, con, rel)
;            tqp -> R20,R21
;           tqpp -> R18,R19
;             td -> R16,R17
	.even
_NutThreadAddPriQueue::
	st -y,R20
	st -y,R21
; /*
;  * Copyright (C) 2001-2005 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  * -
;  * Portions Copyright (C) 2000 David J. Hudson <dave@humbug.demon.co.uk>
;  *
;  * This file is distributed in the hope that it will be useful, but WITHOUT
;  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;  * FITNESS FOR A PARTICULAR PURPOSE.
;  *
;  * You can redistribute this file and/or modify it under the terms of the GNU
;  * General Public License (GPL) as published by the Free Software Foundation;
;  * either version 2 of the License, or (at your discretion) any later version.
;  * See the accompanying file "copying-gpl.txt" for more details.
;  *
;  * As a special exception to the GPL, permission is granted for additional
;  * uses of the text contained in this file.  See the accompanying file
;  * "copying-liquorice.txt" for details.
;  */
; 
; /*
;  * $Log: thread.c,v $
;  * Revision 1.22  2005/10/04 05:14:49  hwmaier
;  * Added support for separating stack and conventional heap as required by AT09CAN128 MCUs
;  *
;  * Revision 1.21  2005/07/26 16:00:16  haraldkipp
;  * Cygwin added.
;  * Platform dependent part is now located in arch/(target)/context.c
;  *
;  * Revision 1.20  2005/07/14 08:57:54  freckle
;  * Rewrote CS in NutThreadSetPriority
;  *
;  * Revision 1.19  2005/07/12 13:57:56  freckle
;  * Fixed bug in thread not waking up on NutSleep
;  *
;  * Revision 1.18  2005/06/12 16:56:20  haraldkipp
;  * Calls new function to process elapsed timers.
;  *
;  * Revision 1.17  2005/05/27 17:30:26  drsung
;  * Platform dependant files were moved to /arch directory.
;  *
;  * Revision 1.16  2005/04/19 14:15:28  phblum
;  * Moved trace macro to avoid jamming the buffer with yields that do not switch to another thread (e.g. idle thread does so all the time).
;  *
;  * Revision 1.15  2005/02/21 12:37:57  phblum
;  * Removed tabs and added semicolons after NUTTRACER macros
;  *
;  * Revision 1.14  2005/02/21 00:53:57  hwmaier
;  * Changes to incorporate support for AT90CAN128 CPU
;  *
;  * Revision 1.13  2005/02/17 14:44:39  phblum
;  * Removed volatile declarations of runQueue since it is not touched from interrupt context. Same for nutThreadList, runningThread and killedThreads.
;  * As runQueue is not modified by interrupts, runningThread == runQueue always. Therefore removed obsolete comparison in NutThreadYield().
;  * Removed unnecessary critical section in NutThreadYield.
;  * Put NutThreadSwitch call of NutThreadResume into critical section.
;  *
;  * Revision 1.12  2005/02/16 19:53:17  haraldkipp
;  * Ready-to-run queue handling removed from interrupt context.
;  *
;  * Revision 1.11  2005/01/24 22:34:36  freckle
;  * Added new tracer by Phlipp Blum <blum@tik.ee.ethz.ch>
;  *
;  * Revision 1.10  2005/01/02 10:07:10  haraldkipp
;  * Replaced platform dependant formats in debug outputs.
;  *
;  * Revision 1.9  2004/08/05 12:13:57  freckle
;  * Added unix emulation hook in NutThreadYield to safely process
;  * NutPostEventAsync calls occuring in non Nut/OS threads.
;  * Rewrote the unix read function again using the new unix NutThreadYield hook
;  * to call the NutPostEventAsync function safely (fast & correct).
;  * _write(nf, 0, 0) aka fflush is ignored on unix emulation.
;  *
;  * Revision 1.8  2004/04/07 12:13:58  haraldkipp
;  * Matthias Ringwald's *nix emulation added
;  *
;  * Revision 1.7  2004/03/19 09:05:12  jdubiec
;  * Fixed format strings declarations for AVR.
;  *
;  * Revision 1.6  2004/03/16 16:48:45  haraldkipp
;  * Added Jan Dubiec's H8/300 port.
;  *
;  * Revision 1.5  2004/02/01 18:49:48  haraldkipp
;  * Added CPU family support
;  *
;  * Revision 1.4  2003/12/15 19:31:46  haraldkipp
;  * Ralph Mason's thread termination and reduced stack frame.
;  *
;  * Revision 1.3  2003/11/19 12:06:23  drsung
;  * CR/LF issue corrected
;  *
;  * Revision 1.2  2003/11/18 22:18:40  drsung
;  * thread name with 8 characters fixed.
;  *
;  * Revision 1.1.1.1  2003/05/09 14:41:54  haraldkipp
;  * Initial using 3.2.1
;  *
;  * Revision 1.23  2003/05/06 18:56:24  harald
;  * Avoid inlining with GCC optimizations
;  *
;  * Revision 1.22  2003/03/31 14:39:43  harald
;  * *** empty log message ***
;  *
;  * Revision 1.21  2003/02/04 18:15:57  harald
;  * Version 3 released
;  *
;  * Revision 1.20  2002/10/29 15:36:59  harald
;  * New function retrieves thread by name
;  *
;  * Revision 1.19  2002/08/08 17:31:47  harald
;  * Imagecraft support by Klaus Ummenhofer
;  *
;  * Revision 1.18  2002/07/03 16:45:42  harald
;  * Using GCC 3.2
;  *
;  * Revision 1.17  2002/06/26 17:29:44  harald
;  * First pre-release with 2.4 stack
;  *
;  */
; 
; #include <cfg/os.h>
; #include <cfg/memory.h>
; 
; #include <string.h>
; 
; #ifdef __IMAGECRAFT__
; #include <iom128v.h>
; #endif
; #include <sys/types.h>
; #include <sys/heap.h>
; #include <sys/atom.h>
; #include <sys/timer.h>
; #include <sys/thread.h>
; 
; #ifdef NUTDEBUG
; #include <sys/osdebug.h>
; #endif
; 
; #ifdef NUTTRACER
; #include <sys/tracer.h>
; #endif
; 
; /*!
;  * \addtogroup xgThread
;  */
; /*@{*/
; 
; #if defined(__linux__) || defined(__APPLE__) || defined(__CYGWIN__)
; #include "../arch/unix/os/thread.c"
; #endif
; 
; /*!
;  * \brief Currently running thread.
;  *
;  * Pointer to the NUTTHREADINFO structure of the currently
;  * running thread.
;  */
; NUTTHREADINFO * runningThread;
; 
; /*!
;  * \brief Thread to be killed.
;  *
;  * Pointer to the NUTTHREADINFO structure of the latest
;  * killed thread.
;  */
; NUTTHREADINFO * killedThread;
; 
; /*!
;  * \brief List of all created threads.
;  *
;  * Linked list of NUTTHREADINFO structures of all threads.
;  * New threads are put in front. This list contains at
;  * least two threads, the main application thread followed
;  * by the idle thread.
;  */
; NUTTHREADINFO * nutThreadList;
; 
; /*!
;  * \brief List of threads to resume.
;  *
;  * Unordered linked list of NUTTHREADINFO structures of all threads which
;  * are ready to run but haven't been added to the runQueue yet.
;  */
; NUTTHREADINFO *volatile readyQueue;
; 
; /*!
;  * \brief List of ready-to-run threads.
;  *
;  * Priority ordered linked list of NUTTHREADINFO structures
;  * of all threads which are ready to run.
;  */
; NUTTHREADINFO * runQueue;
; 
; 
; 
; /*!
;  * \brief Add a thread to a prioritiy ordered queue.
;  *
;  * Insert the thread into a specified queue behind
;  * the last thread with lower or equal priority.
;  *
;  * \note Depending on the given queue, CPU interrupts must have
;  * been disabled before calling this function.
;  *
;  * \param td   Pointer to NUTTHREADINFO of the thread to be
;  *             inserted in the queue.
;  * \param tqpp Pointer to the root of the queue.
;  */
; void NutThreadAddPriQueue(NUTTHREADINFO * td, NUTTHREADINFO ** tqpp)
; {
;     NUTTHREADINFO *tqp;
; #ifdef NUTDEBUG
;     NUTTHREADINFO **dump = tqpp;
; #endif
; 
;     td->td_queue = tqpp;
	movw R30,R16
	std z+22,R19
	std z+21,R18
;     tqp = *tqpp;
	movw R30,R18
	ldd R20,z+0
	ldd R21,z+1
	xjmp L3
L2:
;     while (tqp && (tqp->td_priority <= td->td_priority)) {
;         tqpp = &tqp->td_qnxt;
	movw R18,R20
	subi R18,254  ; offset = 2
	sbci R19,255
;         tqp = tqp->td_qnxt;
	movw R30,R20
	ldd R20,z+2
	ldd R21,z+3
;     }
L3:
	cpi R20,0
	cpc R20,R21
	breq L5
X0:
	movw R30,R16
	ldd R2,z+16
	clr R3
	movw R30,R20
	ldd R4,z+16
	clr R5
	cp R2,R4
	cpc R3,R5
	brge L2
X1:
L5:
;     td->td_qnxt = tqp;
	movw R30,R16
	std z+3,R21
	std z+2,R20
;     *tqpp = td;
	movw R30,R18
	std z+1,R17
	std z+0,R16
L1:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;            tqp -> R20,R21
;           tqpp -> R18,R19
;             td -> R16,R17
	.even
_NutThreadRemoveQueue::
	st -y,R20
	st -y,R21
; #ifdef NUTDEBUG
;     if (__os_trf)
;         NutDumpThreadQueue(__os_trs, *dump);
; #endif
; }
; 
; /*!
;  * \brief Remove a thread from a specified queue.
;  *
;  * \note Depending on the given queue, CPU interrupts must have
;  * been disabled before calling this function.
;  *
;  * \param td   Pointer to NUTTHREADINFO of the thread to be
;  *             removed from the queue.
;  * \param tqpp Pointer to the root of the queue.
;  */
; void NutThreadRemoveQueue(NUTTHREADINFO * td, NUTTHREADINFO * volatile *tqpp)
; {
;     NUTTHREADINFO *tqp;
; #ifdef NUTDEBUG
;     NUTTHREADINFO *volatile *dump = tqpp;
; #endif
; 
;     tqp = *tqpp;
	movw R30,R18
	ldd R20,z+0
	ldd R21,z+1
	xjmp L8
L7:
;     while (tqp) {
;         if (tqp == td) {
	cp R20,R16
	cpc R21,R17
	brne L10
X2:
;             *tqpp = td->td_qnxt;
	movw R30,R16
	ldd R2,z+2
	ldd R3,z+3
	movw R30,R18
	std z+1,R3
	std z+0,R2
;             td->td_qnxt = 0;
	clr R2
	clr R3
	movw R30,R16
	std z+3,R3
	std z+2,R2
;             td->td_queue = 0;
	std z+22,R3
	std z+21,R2
;             break;
	xjmp L9
L10:
;         }
;         tqpp = &tqp->td_qnxt;
	movw R18,R20
	subi R18,254  ; offset = 2
	sbci R19,255
;         tqp = tqp->td_qnxt;
	movw R30,R20
	ldd R20,z+2
	ldd R21,z+3
;     }
L8:
	cpi R20,0
	cpc R20,R21
	brne L7
X3:
L9:
L6:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;             td -> R20,R21
	.even
_NutThreadResume::
	st -y,R20
	st -y,R21
; #ifdef NUTDEBUG
;     if (__os_trf)
;         NutDumpThreadQueue(__os_trs, *dump);
; #endif
; }
; 
; /*!
;  * \brief Continue with the highest priority thread, which is ready to run.
;  *
;  * All threads, which had been woken up in interrupt context, will be
;  * inserted into the priority ordered queue of ready-to-run threads.
;  * If the currently running thread lost its top position in the queue
;  * of ready-to-run threads, then the context will be switched.
;  */
; void NutThreadResume(void)
; {
;     NUTTHREADINFO *td;
; 
;     /*
;      * Firste process elapsed timers
;      */
;     NutTimerProcessElapsed();
	xcall _NutTimerProcessElapsed
L13:
; 
;     /*
;      * Then process all entries of the readyQueue.
;      */
;     do {
;         /* Protect access to the readyQueue. */
;         NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
; 
;         /* Remove the top entry from the woken up threads. */
;         if ((td = readyQueue) != 0) {
	lds R2,_readyQueue
	lds R3,_readyQueue+1
	movw R20,R2
	cpi R20,0
	cpc R20,R21
	breq L16
X4:
;             readyQueue = td->td_qnxt;
	movw R30,R2
	ldd R2,z+2
	ldd R3,z+3
	sts _readyQueue+1,R3
	sts _readyQueue,R2
;         }
L16:
; 
;         NutExitCritical();
	pop R0
	out 0x3F, R0
	
; 
;         /* Add the thread to the runQueue. */
;         if (td) {
	cpi R20,0
	cpc R20,R21
	breq L18
X5:
;             td->td_state = TDS_READY;
	ldi R24,2
	movw R30,R20
	std z+13,R24
;             NutThreadAddPriQueue(td, (NUTTHREADINFO **) & runQueue);
	ldi R18,<_runQueue
	ldi R19,>_runQueue
	movw R16,R30
	xcall _NutThreadAddPriQueue
;         }
L18:
;     } while (td);
L14:
	cpi R20,0
	cpc R20,R21
	brne L13
X6:
; 
; 
;     /* Check for context switch. */
;     if (runningThread != runQueue) {
	lds R2,_runQueue
	lds R3,_runQueue+1
	lds R4,_runningThread
	lds R5,_runningThread+1
	cp R4,R2
	cpc R5,R3
	breq L20
X7:
; #ifdef NUTDEBUG
;         if (__os_trf) {
;             static prog_char fmt2[] = "SW<%p %p>";
;             fprintf_P(__os_trs, fmt2, runningThread, runQueue);
;         }
; #endif
; #ifdef NUTTRACER
;         TRACE_ADD_ITEM(TRACE_TAG_THREAD_YIELD,(int)runningThread);
; #endif
; 
;         if (runningThread->td_state == TDS_RUNNING) {
	movw R30,R4
	ldd R24,z+13
	cpi R24,1
	brne L22
X8:
;             runningThread->td_state = TDS_READY;
	ldi R24,2
	std z+13,R24
;         }
L22:
;         NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;         NutThreadSwitch();
	xcall _NutThreadSwitch
;         NutExitCritical();
	pop R0
	out 0x3F, R0
	
;     }
L20:
L12:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;             th -> R18,R19
;          timer -> R16,R17
	.even
_NutThreadWake::
; }
; 
; /*!
;  * \brief Resume a previously suspended thread.
;  *
;  * This routine is called by the system when a
;  * sleep timer elapses.
;  *
;  * \note This routine is used as a timer callback
;  *       for NutSleep implementation
;  *       Applications typically do not call this
;  *       function.
;  *
;  * \param timer Handle of the elapsed timer.
;  * \param th    Handle of the thread to wake up.
;  */
; void NutThreadWake(HANDLE timer, HANDLE th)
; {
;     /* clear pointer on timer and waiting queue */
;     ((NUTTHREADINFO *) th)->td_timer = 0;
	clr R2
	clr R3
	movw R30,R18
	std z+20,R3
	std z+19,R2
;     ((NUTTHREADINFO *) th)->td_queue = 0;
	std z+22,R3
	std z+21,R2
; 
;     /* protected move into readyQueue */
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;     ((NUTTHREADINFO *) th)->td_qnxt = readyQueue;
	lds R2,_readyQueue
	lds R3,_readyQueue+1
	std z+3,R3
	std z+2,R2
;     readyQueue = (NUTTHREADINFO *)th;
	sts _readyQueue+1,R19
	sts _readyQueue,R18
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
L24:
	.dbline 0 ; func end
	ret
	.even
_NutThreadYield::
; }
; 
; /*!
;  * \brief Give up the CPU.
;  *
;  * If another thread within the same or higher priority
;  * is ready to run, then the current thread is stopped
;  * and the other one is started.
;  *
;  */
; void NutThreadYield(void)
; {
; 
; #if defined(__linux__) || defined(__APPLE__) || defined(__CYGWIN__)
;     NutEnterCritical();
;     NutUnixThreadYieldHook();
;     NutExitCritical();
; #endif
; 
;     /*
;      * Remove current thread from runQueue and reinsert it.
;      * The idle thread is the last one in the queue and will
;      * never be removed.
;      */
;     if (runningThread->td_qnxt) {
	lds R30,_runningThread
	lds R31,_runningThread+1
	ldd R2,z+2
	ldd R3,z+3
	tst R2
	brne X9
	tst R3
	breq L26
X9:
; #ifdef NUTDEBUG
;         if (__os_trf) {
;             static prog_char fmt1[] = "Yld<%p>";
;             fprintf_P(__os_trs, fmt1, runningThread);
;         }
; #endif
;         runQueue = runningThread->td_qnxt;
	ldd R2,z+2
	ldd R3,z+3
	sts _runQueue+1,R3
	sts _runQueue,R2
;         runningThread->td_qnxt = 0;
	clr R2
	clr R3
	std z+3,R3
	std z+2,R2
;         NutThreadAddPriQueue(runningThread, (NUTTHREADINFO **) & runQueue);
	ldi R18,<_runQueue
	ldi R19,>_runQueue
	movw R16,R30
	xcall _NutThreadAddPriQueue
; #ifdef NUTDEBUG
;         if (__os_trf)
;             NutDumpThreadList(__os_trs);
; #endif
;     }
L26:
; 
;     /* Continue with the highest priority thread, which is ready to run. */
;     NutThreadResume();
	xcall _NutThreadResume
L25:
	.dbline 0 ; func end
	ret
;           last -> R10
;          level -> R20
	.even
_NutThreadSetPriority::
	xcall push_xgset300C
	mov R20,R16
; }
; 
; /*!
;  * \brief Set the current thread's priority.
;  *
;  * The priority of newly created threads is set to 64,
;  * but may be changed when the thread starts running.
;  *
;  * Changing the priority level to 255 will kill the
;  * calling thread.
;  *
;  * When another thread with a higher or equal priority
;  * is ready to run, the current thread will be stopped
;  * and control of the CPU is passed to the other thread.
;  *
;  * The function returns the old priority, which makes it
;  * easy to temporarily switch to another priority and
;  * later set back the old one.
;  *
;  * \param level New priority level or 255 to kill the thread.
;  *
;  * \return The old priority of this thread.
;  */
; u_char NutThreadSetPriority(u_char level)
; {
;     u_char last = runningThread->td_priority;
	lds R30,_runningThread
	lds R31,_runningThread+1
	ldd R10,z+16
; 
; #ifdef NUTDEBUG
;     if (__os_trf) {
;         static prog_char fmt1[] = "Pri%u<%p>";
;         fprintf_P(__os_trs, fmt1, level, runningThread);
;     }
; #endif
; 
;     /*
;      * Remove the thread from the run queue and re-insert it with a new
;      * priority, if this new priority level is below 255. A priotity of
;      * 255 will kill the thread.
;      */
;     NutThreadRemoveQueue(runningThread, &runQueue);
	ldi R18,<_runQueue
	ldi R19,>_runQueue
	movw R16,R30
	xcall _NutThreadRemoveQueue
;     runningThread->td_priority = level;
	lds R30,_runningThread
	lds R31,_runningThread+1
	std z+16,R20
;     if (level < 255) {
	cpi R20,255
	brsh L29
X10:
;         NutThreadAddPriQueue(runningThread, (NUTTHREADINFO **) & runQueue);
	ldi R18,<_runQueue
	ldi R19,>_runQueue
	lds R16,_runningThread
	lds R17,_runningThread+1
	xcall _NutThreadAddPriQueue
;     } else {
	xjmp L30
L29:
;         NutThreadKill();
	xcall _NutThreadKill
;     }
L30:
; 
; #ifdef NUTDEBUG
;     if (__os_trf) {
;         NutDumpThreadList(__os_trs);
;         NutDumpThreadQueue(__os_trs, runQueue);
;     }
; #endif
; 
;     /*
;      * Are we still on top of the queue? If yes, then change our status
;      * back to running, otherwise do a context switch.
;      */
;     if (runningThread == runQueue) {
	lds R2,_runQueue
	lds R3,_runQueue+1
	lds R4,_runningThread
	lds R5,_runningThread+1
	cp R4,R2
	cpc R5,R3
	brne L31
X11:
;         runningThread->td_state = TDS_RUNNING;
	ldi R24,1
	movw R30,R4
	std z+13,R24
;     } else {
	xjmp L32
L31:
;         runningThread->td_state = TDS_READY;
	ldi R24,2
	lds R30,_runningThread
	lds R31,_runningThread+1
	std z+13,R24
; #ifdef NUTDEBUG
;         if (__os_trf) {
;             static prog_char fmt2[] = "SWC<%p %p>";
;             fprintf_P(__os_trs, fmt2, runningThread, runQueue);
;         }
; #endif
; #ifdef NUTTRACER
;         TRACE_ADD_ITEM(TRACE_TAG_THREAD_SETPRIO,(int)runningThread);
; #endif
; 
;         NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;         NutThreadSwitch();
	xcall _NutThreadSwitch
;         NutExitCritical();
	pop R0
	out 0x3F, R0
	
;     }
L32:
; 
;     return last;
	mov R16,R10
L28:
	.dbline 0 ; func end
	xjmp pop_xgset300C
	.even
_NutThreadExit::
; }
; 
; /*!
;  * \brief End the current thread
;  *
;  * Terminates the current thread, in due course the memory associated
;  * with the thread will be released back to the OS this is done by the
;  * idle thread.
;  */
; void NutThreadExit(void)
; {
;     NutThreadSetPriority(255);
	ldi R16,255
	xcall _NutThreadSetPriority
L33:
	.dbline 0 ; func end
	ret
	.even
_NutThreadDestroy::
; }
; 
; /*!
;  * \brief Free a thread that was previously killed and release memory
;  *        back to the OS.
;  *
;  * Called when another thread is killed and by the idle thread.
;  *
;  * Applications generally do not call this function, however you could
;  * call it to try to reclaim memory.
;  */
; void NutThreadDestroy(void)
; {
;     if (killedThread) {
	lds R2,_killedThread
	lds R3,_killedThread+1
	tst R2
	brne X12
	tst R3
	breq L35
X12:
; #if defined (NUTMEM_STACKHEAP) /* Stack resides in internal memory */
;         NutStackFree(killedThread->td_memory);
; #else
;         NutHeapFree(killedThread->td_memory);
	movw R30,R2
	ldd R16,z+17
	ldd R17,z+18
	xcall _NutHeapFree
; #endif
;         killedThread = 0;
	clr R2
	clr R3
	sts _killedThread+1,R3
	sts _killedThread,R2
;     }
L35:
L34:
	.dbline 0 ; func end
	ret
;             pp -> R20,R21
;           pCur -> R10,R11
	.even
_NutThreadKill::
	xcall push_xgset300C
; }
; 
; /*!
;  * \brief Kill the running thread.
;  *
;  * The thread is moved from the schedule que and
;  *
;  * Applications generally do not call this function
;  */
; void NutThreadKill(void)
; {
; 
;     NUTTHREADINFO *pCur = nutThreadList;
	lds R10,_nutThreadList
	lds R11,_nutThreadList+1
;     NUTTHREADINFO **pp = (NUTTHREADINFO **) & nutThreadList;
	ldi R20,<_nutThreadList
	ldi R21,>_nutThreadList
; 
;     /* Free up any unfinished already killed threads. */
;     NutThreadDestroy();
	xcall _NutThreadDestroy
	xjmp L39
L38:
; 
;     /* Remove from the thread list. */
;     while (pCur) {
;         if (pCur == runningThread) {
	lds R2,_runningThread
	lds R3,_runningThread+1
	cp R10,R2
	cpc R11,R3
	brne L41
X13:
;             *pp = pCur->td_next;
	movw R30,R10
	ldd R2,z+0
	ldd R3,z+1
	movw R30,R20
	std z+1,R3
	std z+0,R2
;             break;
	xjmp L40
L41:
;         }
; 
;         pp = (NUTTHREADINFO **) & pCur->td_next;
	movw R20,R10
;         pCur = pCur->td_next;
	movw R30,R10
	ldd R10,z+0
	ldd R11,z+1
;     }
L39:
	tst R10
	brne L38
	tst R11
	brne L38
X14:
L40:
; 
;     /* Schedule for cleanup. */
;     killedThread = runningThread;
	lds R2,_runningThread
	lds R3,_runningThread+1
	sts _killedThread+1,R3
	sts _killedThread,R2
L37:
	.dbline 0 ; func end
	xjmp pop_xgset300C
;            tdp -> R20,R21
;           name -> R10,R11
	.even
_GetThreadByName::
	xcall push_xgset300C
	movw R10,R16
; }
; 
; HANDLE GetThreadByName(u_char * name)
; {
;     NUTTHREADINFO *tdp;
; 
;     for (tdp = nutThreadList; tdp; tdp = tdp->td_next) {
	lds R20,_nutThreadList
	lds R21,_nutThreadList+1
	xjmp L47
L44:
;         if (strcmp(tdp->td_name, name) == 0)
	movw R18,R10
	movw R16,R20
	subi R16,252  ; offset = 4
	sbci R17,255
	xcall _strcmp
	cpi R16,0
	cpc R16,R17
	brne L48
X15:
;             return tdp;
	movw R16,R20
	xjmp L43
L48:
;     }
L45:
	movw R30,R20
	ldd R20,z+0
	ldd R21,z+1
L47:
	cpi R20,0
	cpc R20,R21
	brne L44
X16:
;     return NULL;
	clr R16
	clr R17
L43:
	.dbline 0 ; func end
	xjmp pop_xgset300C
	.area bss(ram, con, rel)
_killedThread::
	.blkb 2
_runQueue::
	.blkb 2
_readyQueue::
	.blkb 2
_nutThreadList::
	.blkb 2
_runningThread::
	.blkb 2
; }
; 
; 
; #if defined (NUTMEM_STACKHEAP) /* Stack resides in internal memory */
; /*
;  * The following routines are wrappers around the standard heap
;  * allocation routines.  These wrappers tweak the free heap pointer to point
;  * to a second heap which is kept in internal memory and used only for a
;  * thread's stack.
;  */
; 
; HEAPNODE* volatile stackHeapFreeList = 0; /* for special stack heap */
; 
; void *NutStackAlloc(size_t size)
; {
;     void * result;
;     HEAPNODE* savedHeapNode;
; 
;     savedHeapNode = heapFreeList;
;     heapFreeList = stackHeapFreeList;
;     result = NutHeapAlloc(size);
;     stackHeapFreeList = heapFreeList;
;     heapFreeList = savedHeapNode;
;     return result;
; }
; 
; int NutStackFree(void *block)
; {
;     int result;
;     HEAPNODE* savedHeapNode;
; 
;     savedHeapNode = heapFreeList;
;     heapFreeList = stackHeapFreeList;
;     result = NutHeapFree(block);
;     stackHeapFreeList = heapFreeList;
;     heapFreeList = savedHeapNode;
;     return result;
; }
; 
; void NutStackAdd(void *addr, size_t size)
; {
;    HEAPNODE* savedHeapNode;
; 
;    savedHeapNode = heapFreeList;
;    heapFreeList = stackHeapFreeList;
;    NutHeapAdd(addr, size);
;    stackHeapFreeList = heapFreeList;
;    heapFreeList = savedHeapNode;
; }
; 
; #endif /* defined(NUTMEM_STACKHEAP) */
; 
; /*@}*/
