	.module timer.c
	.area text(rom, con, rel)
;            arg -> R16,R17
	.even
_NutTimerIntr:
; /*
;  * Copyright (C) 2001-2005 by egnite Software GmbH. All rights reserved.
;  *
;  * Redistribution and use in source and binary forms, with or without
;  * modification, are permitted provided that the following conditions
;  * are met:
;  *
;  * 1. Redistributions of source code must retain the above copyright
;  *    notice, this list of conditions and the following disclaimer.
;  * 2. Redistributions in binary form must reproduce the above copyright
;  *    notice, this list of conditions and the following disclaimer in the
;  *    documentation and/or other materials provided with the distribution.
;  * 3. Neither the name of the copyright holders nor the names of
;  *    contributors may be used to endorse or promote products derived
;  *    from this software without specific prior written permission.
;  *
;  * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
;  * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
;  * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
;  * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
;  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
;  * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
;  * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
;  * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
;  * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;  * SUCH DAMAGE.
;  *
;  * For additional information see http://www.ethernut.de/
;  *
;  * -
;  * Portions Copyright (C) 2000 David J. Hudson <dave@humbug.demon.co.uk>
;  *
;  * This file is distributed in the hope that it will be useful, but WITHOUT
;  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;  * FITNESS FOR A PARTICULAR PURPOSE.
;  *
;  * You can redistribute this file and/or modify it under the terms of the GNU
;  * General Public License (GPL) as published by the Free Software Foundation;
;  * either version 2 of the License, or (at your discretion) any later version.
;  * See the accompanying file "copying-gpl.txt" for more details.
;  *
;  * As a special exception to the GPL, permission is granted for additional
;  * uses of the text contained in this file.  See the accompanying file
;  * "copying-liquorice.txt" for details.
;  */
; 
; /*
;  * $Log: timer.c,v $
;  * Revision 1.28  2005/08/03 14:43:54  freckle
;  * made nut_tick public
;  *
;  * Revision 1.27  2005/07/26 16:01:19  haraldkipp
;  * Cygwin added.
;  * Platform dependent code is now located in arch/(target)/ostimer.c.
;  *
;  * Revision 1.26  2005/07/21 15:31:04  freckle
;  * rewrote timer handling. timer interrupt handler only updates nut_ticks now
;  *
;  * Revision 1.25  2005/07/20 09:26:34  haraldkipp
;  * Use native heap calls to avoid dependencies
;  *
;  * Revision 1.24  2005/07/15 14:44:46  freckle
;  * corrected NutGetMillis, update comments
;  *
;  * Revision 1.23  2005/07/12 16:37:02  freckle
;  * made NutTimerInsert public
;  *
;  * Revision 1.22  2005/07/12 16:26:04  freckle
;  * extracted timer creation from NutTimerStartTicks into NutTimerCreate
;  *
;  * Revision 1.21  2005/07/12 15:23:40  olereinhardt
;  * Added NULL pointer checks in NutTimerProcessElapsed(void)
;  *
;  * Revision 1.20  2005/07/12 15:10:43  freckle
;  * removed critical section in NutSleep
;  *
;  * Revision 1.19  2005/07/07 12:43:49  freckle
;  * removed stopped timers in NutTimerProcessElapsed
;  *
;  * Revision 1.18  2005/06/14 19:04:44  freckle
;  * Fixed NutTimerStopAsync(): it now removes callback ptr => timer stop
;  *
;  * Revision 1.17  2005/06/12 16:53:01  haraldkipp
;  * Major redesing to reduce interrupt latency.
;  * Much better seperation of hardware dependent and independent parts.
;  *
;  * Revision 1.16  2005/05/27 17:30:27  drsung
;  * Platform dependant files were moved to /arch directory.
;  *
;  * Revision 1.15  2005/02/21 12:37:57  phblum
;  * Removed tabs and added semicolons after NUTTRACER macros
;  *
;  * Revision 1.14  2005/02/16 19:53:27  haraldkipp
;  * Ready-to-run queue handling removed from interrupt context.
;  *
;  * Revision 1.13  2005/02/10 07:06:51  hwmaier
;  * Changes to incorporate support for AT90CAN128 CPU
;  *
;  * Revision 1.12  2005/01/24 22:34:40  freckle
;  * Added new tracer by Phlipp Blum <blum@tik.ee.ethz.ch>
;  *
;  * Revision 1.11  2005/01/19 17:59:43  freckle
;  * Improved interrupt performance by reducing some critical section
;  *
;  * Revision 1.10  2005/01/13 18:51:23  haraldkipp
;  * Moved ms62_5 counter to nutinit.c to make sure this is located in internal
;  * RAM (AVR platforms). This fixes the wrong baudrate bug for applications
;  * occupying all internal RAM.
;  *
;  * Optimized some timer handling.
;  *
;  * Revision 1.9  2005/01/02 10:07:10  haraldkipp
;  * Replaced platform dependant formats in debug outputs.
;  *
;  * Revision 1.8  2004/09/08 10:53:25  haraldkipp
;  * os/timer.c
;  *
;  * Revision 1.7  2004/04/07 12:13:58  haraldkipp
;  * Matthias Ringwald's *nix emulation added
;  *
;  * Revision 1.6  2004/03/19 09:05:12  jdubiec
;  * Fixed format strings declarations for AVR.
;  *
;  * Revision 1.5  2004/03/16 16:48:45  haraldkipp
;  * Added Jan Dubiec's H8/300 port.
;  *
;  * Revision 1.4  2004/02/01 18:49:48  haraldkipp
;  * Added CPU family support
;  *
;  * Revision 1.3  2003/11/03 16:34:21  haraldkipp
;  * New API returns system uptime in miliseconds
;  *
;  * Revision 1.2  2003/10/13 10:17:11  haraldkipp
;  * Seconds counter added
;  *
;  * Revision 1.1.1.1  2003/05/09 14:41:55  haraldkipp
;  * Initial using 3.2.1
;  *
;  * Revision 1.24  2003/05/06 18:56:58  harald
;  * Allow compilation without clock crystal
;  *
;  * Revision 1.23  2003/04/21 17:09:19  harald
;  * *** empty log message ***
;  *
;  * Revision 1.22  2003/03/31 14:39:54  harald
;  * *** empty log message ***
;  *
;  * Revision 1.21  2003/02/04 18:15:57  harald
;  * Version 3 released
;  *
;  * Revision 1.20  2003/01/14 16:58:41  harald
;  * Tick counter added.
;  *
;  * Revision 1.19  2002/10/29 15:37:55  harald
;  * Some compilers refuse fixed crystal source
;  *
;  * Revision 1.18  2002/09/15 17:08:07  harald
;  * *** empty log message ***
;  *
;  * Revision 1.17  2002/08/08 17:32:44  harald
;  * Imagecraft support by Klaus Ummenhofer
;  *
;  * Revision 1.16  2002/06/26 17:29:45  harald
;  * First pre-release with 2.4 stack
;  *
;  */
; 
; #include <cfg/os.h>
; #include <dev/irqreg.h>
; 
; #include <sys/types.h>
; #include <sys/atom.h>
; #include <sys/heap.h>
; #include <sys/thread.h>
; #include <sys/timer.h>
; 
; #ifdef NUTDEBUG
; #include <sys/osdebug.h>
; #endif
; 
; #ifdef NUTTRACER
; #include <sys/tracer.h>
; #endif
; 
; /*!
;  * \addtogroup xgTimer
;  */
; /*@{*/
; 
; /*!
;  * \brief Linked list of all system timers.
;  */
; NUTTIMERINFO *volatile nutTimerList;
; 
; /*!
; * \brief Flag to indicate that a timer was stopped.
;  */
; u_char volatile nutTimerStopped;
; 
; /*!
;  * \brief Pointer to the timer, which is currently processed by the timer
;  *        interrupt handler.
;  */
; // NUTTIMERINFO *volatile runningTimer;
; 
; /*!
; *  \brief System tick counter
;  */
; volatile u_long nut_ticks;
; 
; /*!
;  *  \brief Time of last NutTimerProcessElapsed execution
;  */
; static u_long nut_ticks_resume;
; 
; /*!
;  * \brief System timer interrupt handler.
;  */
; #ifdef USE_TIMER
; SIGNAL( SIG_TIMER ) 
; #else
; static void NutTimerIntr(void *arg)
; #endif
; {
;     nut_ticks++;
	ldi R24,1
	ldi R25,0
	ldi R26,0
	ldi R27,0
	lds R4,_nut_ticks+2
	lds R5,_nut_ticks+2+1
	lds R2,_nut_ticks
	lds R3,_nut_ticks+1
	add R2,R24
	adc R3,R25
	adc R4,R26
	adc R5,R27
	sts _nut_ticks+1,R3
	sts _nut_ticks,R2
	sts _nut_ticks+2+1,R5
	sts _nut_ticks+2,R4
L2:
	.dbline 0 ; func end
	ret
	.even
_NutTimerInit::
; }
; 
; 
; /*!
;  * \brief Initialize system timer.
;  *
;  * This function is automatically called by Nut/OS during system 
;  * initialization. It calls the hardware dependent layer to initialze
;  * the timer hardware and register a timer interrupt handler.
;  */
; void NutTimerInit(void)
; {
;     NutRegisterTimer(NutTimerIntr);
	ldi R16,<PL_NutTimerIntr
	ldi R17,>PL_NutTimerIntr
	xcall _NutRegisterTimer
;     NutEnableTimerIrq();
	in R24,0x37
	ori R24,2
	out 0x37,R24
L3:
	.dbline 0 ; func end
	ret
;           tnpp -> R20,R21
;            tnp -> R22,R23
;             tn -> R16,R17
	.even
_NutTimerInsert::
	xcall push_xgsetF000
; }
; 
; 
; /*!
;  * \brief Insert a new timer in the global timer list.
;  */
; void NutTimerInsert(NUTTIMERINFO * tn)
; {
;     NUTTIMERINFO **tnpp;
;     NUTTIMERINFO *tnp;
; 
; #ifdef NUTDEBUG
;     if (__os_trf) {
;         static prog_char fmt[] = "InsTmr<%p>\n";
;         fprintf_P(__os_trs, fmt, tn);
;     }
; #endif
; 
;     tnpp = (NUTTIMERINFO **) & nutTimerList;
	ldi R20,<_nutTimerList
	ldi R21,>_nutTimerList
;     for (tnp = nutTimerList; tnp; tnp = tnp->tn_next) {
	lds R22,_nutTimerList
	lds R23,_nutTimerList+1
	xjmp L8
L5:
;         if (tn->tn_ticks_left < tnp->tn_ticks_left) {
	movw R30,R22
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	movw R30,R16
	ldd R6,z+6
	ldd R7,z+7
	ldd R8,z+8
	ldd R9,z+9
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh L9
X0:
;             tnp->tn_ticks_left -= tn->tn_ticks_left;
	movw R24,R22
	adiw R24,6
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;             break;
	xjmp L7
L9:
;         }
;         tn->tn_ticks_left -= tnp->tn_ticks_left;
	movw R24,R16
	adiw R24,6
	movw R30,R22
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;         tnpp = &tnp->tn_next;
	movw R20,R22
;     }
L6:
	movw R30,R22
	ldd R22,z+0
	ldd R23,z+1
L8:
	cpi R22,0
	cpc R22,R23
	breq X2
	xjmp L5
X2:
X1:
L7:
;     tn->tn_next = tnp;
	movw R30,R16
	std z+1,R23
	std z+0,R22
;     *tnpp = tn;
	movw R30,R20
	std z+1,R17
	std z+0,R16
L4:
	.dbline 0 ; func end
	xjmp pop_xgsetF000
; tempTimerStopped -> R12
;          ticks -> y+4
;         nexttn -> R12,R13
;      ticks_new -> y+0
;             tn -> R10,R11
	.even
_NutTimerProcessElapsed::
	xcall push_xgsetF03C
	sbiw R28,12
; 
; #ifdef NUTDEBUG
;     if (__os_trf)
;         NutDumpTimerList(__os_trs);
; #endif
; }
; 
; void NutTimerProcessElapsed(void)
; {
;     NUTTIMERINFO *tn, *nexttn;
;     u_char tempTimerStopped;
;     u_long ticks;
;     u_long ticks_new;
;     
;     // calculate ticks since last call
;     ticks = NutGetTickCount();
	xcall _NutGetTickCount
	std y+4,R16
	std y+5,R17
	std y+6,R18
	std y+7,R19
;     ticks_new = ticks - nut_ticks_resume;
	lds R4,_nut_ticks_resume+2
	lds R5,_nut_ticks_resume+2+1
	lds R2,_nut_ticks_resume
	lds R3,_nut_ticks_resume+1
	ldd R6,y+4
	ldd R7,y+5
	ldd R8,y+6
	ldd R9,y+7
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	std y+0,R6
	std y+1,R7
	std y+2,R8
	std y+3,R9
;     nut_ticks_resume = ticks;
	ldd R2,y+4
	ldd R3,y+5
	ldd R4,y+6
	ldd R5,y+7
	sts _nut_ticks_resume+1,R3
	sts _nut_ticks_resume,R2
	sts _nut_ticks_resume+2+1,R5
	sts _nut_ticks_resume+2,R4
	xjmp L13
L12:
;     
;     // process timers
;     while (nutTimerList && ticks_new){
;         
;         tn = nutTimerList;
	lds R10,_nutTimerList
	lds R11,_nutTimerList+1
; 
;         // subtract time
;         if (ticks_new < tn->tn_ticks_left) {
	movw R30,R10
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	cp R6,R2
	cpc R7,R3
	cpc R8,R4
	cpc R9,R5
	brsh L15
X3:
;             tn->tn_ticks_left -= ticks_new;
	movw R24,R30
	adiw R24,6
	movw R2,R6
	movw R4,R8
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;             ticks_new = 0;
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	std y+0,R20
	std y+1,R21
	std y+2,R22
	std y+3,R23
;         } else {
	xjmp L16
L15:
;             ticks_new -= tn->tn_ticks_left;
	movw R30,R10
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	sub R6,R2
	sbc R7,R3
	sbc R8,R4
	sbc R9,R5
	std y+0,R6
	std y+1,R7
	std y+2,R8
	std y+3,R9
;             tn->tn_ticks_left = 0;
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	std z+6,R20
	std z+7,R21
	std z+8,R22
	std z+9,R23
;         }
L16:
;         
;         // elapsed
;         if (tn->tn_ticks_left == 0){
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	movw R30,R10
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	cp R2,R20
	cpc R3,R21
	cpc R4,R22
	cpc R5,R23
	breq X15
	xjmp L17
X15:
X4:
; 
;             // remove from list
;             nutTimerList = nutTimerList->tn_next;
	lds R30,_nutTimerList
	lds R31,_nutTimerList+1
	ldd R2,z+0
	ldd R3,z+1
	sts _nutTimerList+1,R3
	sts _nutTimerList,R2
; 
;             // callback
;             if (tn->tn_callback) {
	movw R30,R10
	ldd R2,z+10
	ldd R3,z+11
	tst R2
	brne X5
	tst R3
	breq L19
X5:
;                 (*tn->tn_callback) (tn, (void *) tn->tn_arg);
	ldd R18,z+12
	ldd R19,z+13
	movw R16,R30
	ldd R26,z+10
	ldd R27,z+11
	movw R30,R26
	xcall  xicall
;             }
L19:
;             // re-insert
;             if ((tn->tn_ticks_left = tn->tn_ticks) != 0) {
	movw R30,R10
	ldd R2,z+2
	ldd R3,z+3
	ldd R4,z+4
	ldd R5,z+5
	std y+8,R2
	std y+9,R3
	std y+10,R4
	std y+11,R5
	ldd R2,y+8
	ldd R3,y+9
	ldd R4,y+10
	ldd R5,y+11
	std z+6,R2
	std z+7,R3
	std z+8,R4
	std z+9,R5
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	cp R2,R20
	cpc R3,R21
	cpc R4,R22
	cpc R5,R23
	breq L21
X6:
;                 NutTimerInsert(tn);
	movw R16,R30
	xcall _NutTimerInsert
;             } else {
	xjmp L22
L21:
;                 NutHeapFree(tn);
	movw R16,R10
	xcall _NutHeapFree
;             }
L22:
;         }
L17:
;     }
L13:
	lds R2,_nutTimerList
	lds R3,_nutTimerList+1
	tst R2
	brne X7
	tst R3
	breq L23
X7:
	ldi R20,0
	ldi R21,0
	ldi R22,0
	ldi R23,0
	ldd R2,y+0
	ldd R3,y+1
	ldd R4,y+2
	ldd R5,y+3
	cp R2,R20
	cpc R3,R21
	cpc R4,R22
	cpc R5,R23
	breq X16
	xjmp L12
X16:
X8:
L23:
; 
;     // some timers stopped by NutTimerStopAsync?
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;     tempTimerStopped = nutTimerStopped;
	lds R12,_nutTimerStopped
;     nutTimerStopped  = 0;
	clr R2
	sts _nutTimerStopped,R2
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
; 
;     // remove stopped timers, if nutTimerStopped
;     if ( (nutTimerList && tempTimerStopped) ){
	lds R2,_nutTimerList
	lds R3,_nutTimerList+1
	tst R2
	brne X9
	tst R3
	brne X17
	xjmp L24
X17:
X9:
	tst R12
	brne X18
	xjmp L24
X18:
X10:
;         tn = nutTimerList;
	lds R10,_nutTimerList
	lds R11,_nutTimerList+1
	xjmp L27
L26:
;         while ((tn) && (tn->tn_next)) {
;             nexttn = tn->tn_next;
	movw R30,R10
	ldd R12,z+0
	ldd R13,z+1
;             if (nexttn->tn_callback == 0) {
	movw R30,R12
	ldd R2,z+10
	ldd R3,z+11
	tst R2
	breq X19
	xjmp L29
X19:
	tst R3
	breq X20
	xjmp L29
X20:
X11:
;                 // remove entry from linked list and update ticks
;                 if (nexttn->tn_next) {
	ldd R2,z+0
	ldd R3,z+1
	tst R2
	brne X12
	tst R3
	breq L31
X12:
;                     nexttn->tn_next->tn_ticks_left += nexttn->tn_ticks_left;
	ldd R24,z+0
	ldd R25,z+1
	adiw R24,6
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	movw R30,R24
	ldd R6,z+0
	ldd R7,z+1
	ldd R8,z+2
	ldd R9,z+3
	add R6,R2
	adc R7,R3
	adc R8,R4
	adc R9,R5
	std z+0,R6
	std z+1,R7
	std z+2,R8
	std z+3,R9
;                 }
L31:
;                 tn->tn_next = nexttn->tn_next;
	movw R30,R12
	ldd R2,z+0
	ldd R3,z+1
	movw R30,R10
	std z+1,R3
	std z+0,R2
;                 NutHeapFree(nexttn);
	movw R16,R12
	xcall _NutHeapFree
;             }
L29:
;             tn = tn->tn_next;
	movw R30,R10
	ldd R10,z+0
	ldd R11,z+1
;         }
L27:
	tst R10
	brne X13
	tst R11
	breq L33
X13:
	movw R30,R10
	ldd R2,z+0
	ldd R3,z+1
	tst R2
	breq X21
	xjmp L26
X21:
	tst R3
	breq X22
	xjmp L26
X22:
X14:
L33:
;     }
L24:
L11:
	.dbline 0 ; func end
	adiw R28,12
	xjmp pop_xgsetF03C
;             tn -> R20,R21
;          flags -> y+10
;            arg -> y+8
;       callback -> y+6
;          ticks -> y+2
	.even
_NutTimerCreate::
	xcall push_arg4
	st -y,R20
	st -y,R21
; }
; 
; 
; /*!
; * \brief Create a new timer.
;  */
; NUTTIMERINFO * NutTimerCreate(u_long ticks, void (*callback) (HANDLE, void *), void *arg, u_char flags)
; {
;     NUTTIMERINFO *tn;
;     
;     tn = NutHeapAlloc(sizeof(NUTTIMERINFO));
	ldi R16,14
	ldi R17,0
	xcall _NutHeapAlloc
	movw R20,R16
;     if (tn) {
	cpi R16,0
	cpc R16,R17
	brne X25
	xjmp L35
X25:
X23:
;         tn->tn_ticks_left = ticks;
	ldd R2,y+2
	ldd R3,y+3
	ldd R4,y+4
	ldd R5,y+5
	movw R30,R20
	std z+6,R2
	std z+7,R3
	std z+8,R4
	std z+9,R5
;         
;         /*
;          * Periodic timers will reload the tick counter on each timer 
;          * intervall.
;          */
;         if (flags & TM_ONESHOT) {
	ldd R0,y+10
	sbrs R0,0
	rjmp L37
X24:
;             tn->tn_ticks = 0;
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	std z+2,R24
	std z+3,R25
	std z+4,R26
	std z+5,R27
;         } else {
	xjmp L38
L37:
;             tn->tn_ticks = tn->tn_ticks_left;
	movw R30,R20
	ldd R2,z+6
	ldd R3,z+7
	ldd R4,z+8
	ldd R5,z+9
	std z+2,R2
	std z+3,R3
	std z+4,R4
	std z+5,R5
;         }
L38:
;         
;         /* Set callback and callback argument. */
;         tn->tn_callback = callback;
	movw R30,R20
	ldd R0,y+6
	ldd R1,y+7
	std z+11,R1
	std z+10,R0
;         tn->tn_arg = arg;
	ldd R0,y+8
	ldd R1,y+9
	std z+13,R1
	std z+12,R0
;     }
L35:
;     return tn;    
	movw R16,R20
L34:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	adiw R28,4
	ret
;             tn -> R10,R11
;          flags -> y+15
;            arg -> y+13
;       callback -> y+11
;          ticks -> y+7
	.even
_NutTimerStartTicks::
	xcall push_arg4
	st -y,R10
	st -y,R11
	sbiw R28,5
; }
; 
; 
; HANDLE NutTimerStartTicks(u_long ticks, void (*callback) (HANDLE, void *), void *arg, u_char flags)
; {
;     NUTTIMERINFO *tn;
; 
;     tn = NutTimerCreate( ticks, callback, arg, flags);
	ldd R0,y+15
	std y+4,R0
	ldd R0,y+13
	ldd R1,y+14
	std y+3,R1
	std y+2,R0
	ldd R0,y+11
	ldd R1,y+12
	std y+1,R1
	std y+0,R0
	ldd R16,y+7
	ldd R17,y+8
	ldd R18,y+9
	ldd R19,y+10
	xcall _NutTimerCreate
	movw R10,R16
;     if (tn) {
	cpi R16,0
	cpc R16,R17
	breq L40
X26:
;         /* Add the timer to the list. */
;         NutTimerInsert(tn);
	xcall _NutTimerInsert
;     }
L40:
;     return tn;
	movw R16,R10
L39:
	.dbline 0 ; func end
	adiw R28,5
	ld R11,y+
	ld R10,y+
	adiw R28,4
	ret
;          flags -> y+13
;            arg -> y+11
;       callback -> y+9
;             ms -> y+5
	.even
_NutTimerStart::
	xcall push_arg4
	sbiw R28,5
; }
; 
; /*!
;  * \brief Create an asynchronous timer.
;  *
;  * The function returns immediately, while the timer runs
;  * asynchronously in the background.
;  *
;  * The timer counts for a specified number of milliseconds,
;  * then calls the callback routine with a given argument.
;  *
;  * The callback function is executed in interrupt context
;  * at a very high priority. It can call only a limited set
;  * of functions and must return as soon as possible.
;  *
;  * \param ms       Specifies the timer interval in milliseconds.
;  * \param callback Identifies the function to be called on each
;  *                 timer interval.
;  * \param arg      The argument passed to the callback function.
;  * \param flags    If set to TM_ONESHOT, the timer will be stopped
;  *                 after the first interval. Set to 0 for periodic
;  *                 timers.
;  *
;  * \return Timer handle if successfull, 0 otherwise. The handle
;  *         may be used to stop the timer by calling TimerStop.
;  */
; HANDLE NutTimerStart(u_long ms, void (*callback) (HANDLE, void *), void *arg, u_char flags)
; {
; 	return NutTimerStartTicks(NutTimerMillisToTicks(ms), callback, arg, flags);
	ldd R16,y+5
	ldd R17,y+6
	ldd R18,y+7
	ldd R19,y+8
	xcall _NutTimerMillisToTicks
	ldd R0,y+13
	std y+4,R0
	ldd R0,y+11
	ldd R1,y+12
	std y+3,R1
	std y+2,R0
	ldd R0,y+9
	ldd R1,y+10
	std y+1,R1
	std y+0,R0
	xcall _NutTimerStartTicks
L42:
	.dbline 0 ; func end
	adiw R28,9
	ret
;             ms -> y+5
	.even
_NutSleep::
	xcall push_arg4
	sbiw R28,5
; }
; 
; /*!
;  * \brief Temporarily suspends the current thread.
;  *
;  * Causes the current thread to wait for a specified interval or,
;  * if the specified interval is zero, to give up the CPU for
;  * another thread with higher or same priority.
;  *
;  * This function may switch to another application thread, that
;  * got the same or a higher priority and is ready to run.
;  *
;  * \note Threads may sleep longer than the specified number of
;  *       milliseconds, depending on the number of threads
;  *       with higher or equal priority, which are ready to run.
;  *
;  * \param ms Milliseconds to sleep. If 0, the current thread will not 
;  *           sleep, but may give up the CPU.
;  */
; void NutSleep(u_long ms)
; {
;     if (ms) {
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	ldd R2,y+5
	ldd R3,y+6
	ldd R4,y+7
	ldd R5,y+8
	cp R2,R24
	cpc R3,R25
	cpc R4,R26
	cpc R5,R27
	brne X29
	xjmp L44
X29:
X27:
; 
;         /* remove running thread from runQueue */
;         NutThreadRemoveQueue(runningThread, &runQueue);
	ldi R18,<_runQueue
	ldi R19,>_runQueue
	lds R16,_runningThread
	lds R17,_runningThread+1
	xcall _NutThreadRemoveQueue
;         runningThread->td_state = TDS_SLEEP;
	ldi R24,3
	lds R30,_runningThread
	lds R31,_runningThread+1
	std z+13,R24
; 
;         if ((runningThread->td_timer = NutTimerStart(ms, NutThreadWake, runningThread, TM_ONESHOT)) != 0) {
	ldi R24,1
	std y+4,R24
	lds R2,_runningThread
	lds R3,_runningThread+1
	std y+3,R3
	std y+2,R2
	ldi R24,<PL_NutThreadWake
	ldi R25,>PL_NutThreadWake
	std y+1,R25
	std y+0,R24
	ldd R16,y+5
	ldd R17,y+6
	ldd R18,y+7
	ldd R19,y+8
	push R2
	push R3
	xcall _NutTimerStart
	pop R31
	pop R30
	std z+20,R17
	std z+19,R16
	cpi R16,0
	cpc R16,R17
	breq L46
X28:
; #ifdef NUTDEBUG
;             if (__os_trf) {
;                 static prog_char fmt1[] = "Rem<%p>\n";
;                 fprintf_P(__os_trs, fmt1, runningThread);
;             }
; #endif
; #ifdef NUTDEBUG
;             if (__os_trf) {
;                 static prog_char fmt2[] = "SWS<%p %p>\n";
;                 NutDumpThreadList(__os_trs);
;                 //NutDumpThreadQueue(__os_trs, runQueue);
;                 fprintf_P(__os_trs, fmt2, runningThread, runQueue);
;             }
; #endif
; #ifdef NUTTRACER
;             TRACE_ADD_ITEM(TRACE_TAG_THREAD_SLEEP,(int)runningThread);
; #endif
;             NutThreadResume();
	xcall _NutThreadResume
;         } else
	xjmp L45
L46:
;         {
;             /* timer creation failed, restore queues */
;             runningThread->td_queue = &runQueue;
	ldi R24,<_runQueue
	ldi R25,>_runQueue
	lds R30,_runningThread
	lds R31,_runningThread+1
	std z+22,R25
	std z+21,R24
;             runningThread->td_qnxt  = runQueue;
	lds R2,_runQueue
	lds R3,_runQueue+1
	std z+3,R3
	std z+2,R2
;             runningThread->td_state = TDS_RUNNING;
	ldi R24,1
	std z+13,R24
;             runQueue = runningThread;
	lds R2,_runningThread
	lds R3,_runningThread+1
	sts _runQueue+1,R3
	sts _runQueue,R2
;         }
;     } else
	xjmp L45
L44:
;         NutThreadYield();
	xcall _NutThreadYield
L45:
L43:
	.dbline 0 ; func end
	adiw R28,9
	ret
;             tn -> R20,R21
;         handle -> R16,R17
	.even
_NutTimerStopAsync::
	st -y,R20
	st -y,R21
; }
; 
; /*!
;  * \brief Asynchronously stop a specified timer.
;  *
;  * Stops one-shot and periodic timers.
;  *
;  * \note It is save to call this function from within an interrupt
;  *       handler.
;  *
;  * \param handle Identifies the timer to be stopped. This handle
;  *               must have been created by calling NutTimerStart().
;  *
;  */
; void NutTimerStopAsync(HANDLE handle)
; {
;     NUTTIMERINFO *tn = (NUTTIMERINFO *)handle;
	movw R20,R16
; 
; #ifdef NUTDEBUG
;     if (__os_trf) {
;         static prog_char fmt[] = "StpTmr<%p>\r\n";
;         fprintf_P(__os_trs, fmt, (uptr_t) handle);
;     }
; #endif
; 
;     /* MR (20050707) moved this commented code to NutTimerProcessElapsed
;      if (tn->tn_next) {
;         tn->tn_next->tn_ticks_left += tn->tn_ticks_left;
;         tn->tn_ticks_left = 0;
;     }
;     */
; 
;     tn->tn_ticks = 0;
	ldi R24,0
	ldi R25,0
	ldi R26,0
	ldi R27,0
	movw R30,R20
	std z+2,R24
	std z+3,R25
	std z+4,R26
	std z+5,R27
;     tn->tn_callback = 0;
	clr R2
	clr R3
	std z+11,R3
	std z+10,R2
;     nutTimerStopped = 1; // set flag for NutTimerProcessElapsed
	ldi R24,1
	sts _nutTimerStopped,R24
L48:
	.dbline 0 ; func end
	ld R21,y+
	ld R20,y+
	ret
;         handle -> R10,R11
	.even
_NutTimerStop::
	st -y,R10
	st -y,R11
	movw R10,R16
; 
; #ifdef NUTDEBUG
;     if (__os_trf)
;         NutDumpTimerList(__os_trs);
; #endif
; }
; 
; /*!
;  * \brief Stop a specified timer.
;  *
;  * Only periodic timers need to be stopped. One-shot timers
;  * are automatically stopped by the timer management after
;  * ther first timer interval. Anyway, long running one-shot
;  * timers may be stopped to release the occupied memory.
;  *
;  * \param handle Identifies the timer to be stopped. This handle
;  *               must have been created by calling NutTimerStart().
;  */
; void NutTimerStop(HANDLE handle)
; {
;     NutTimerStopAsync(handle);
	movw R16,R10
	xcall _NutTimerStopAsync
L49:
	.dbline 0 ; func end
	ld R11,y+
	ld R10,y+
	ret
;             rc -> y+0
	.even
_NutGetTickCount::
	sbiw R28,4
; }
; 
; /*!
;  * \brief Return the number of timer ticks.
;  *
;  * This function returns the TickCount since the system was started. It
;  * is limited to the resolution of the system timer.
;  *
;  * \return Number of ticks.
;  */
; u_long NutGetTickCount(void)
; {
;     u_long rc;
; 
;     NutEnterCritical();
	in R0, 0x3F
	cli
	push R0
	
;     rc = nut_ticks;
	lds R4,_nut_ticks+2
	lds R5,_nut_ticks+2+1
	lds R2,_nut_ticks
	lds R3,_nut_ticks+1
	std y+0,R2
	std y+1,R3
	std y+2,R4
	std y+3,R5
;     NutExitCritical();
	pop R0
	out 0x3F, R0
	
; 
;     return rc;
	ldd R16,y+0
	ldd R17,y+1
	ldd R18,y+2
	ldd R19,y+3
L50:
	.dbline 0 ; func end
	adiw R28,4
	ret
	.even
_NutGetSeconds::
; }
; 
; /*!
;  * \brief Return the seconds counter value.
;  *
;  * This function returns the value of a counter, which is incremented
;  * every second. During system start, the counter is cleared to zero.
;  *
;  * \note There is intentionally no provision to modify the seconds counter.
;  *       Callers can rely on a continuous update and use this value for
;  *       system tick independend timeout calculations. Applications,
;  *       which want to use this counter for date and time functions,
;  *       should use an offset value.
;  *
;  * \return Value of the seconds counter.
;  */
; u_long NutGetSeconds(void)
; {
;     return NutGetTickCount() / NutGetTickClock();
	xcall _NutGetTickCount
	push R16
	push R17
	push R18
	push R19
	xcall _NutGetTickClock
	pop R5
	pop R4
	pop R3
	pop R2
	st -y,R19
	st -y,R18
	st -y,R17
	st -y,R16
	movw R16,R2
	movw R18,R4
	xcall div32u
L51:
	.dbline 0 ; func end
	ret
;        seconds -> y+4
;          ticks -> y+0
	.even
_NutGetMillis::
	xcall push_xgsetF000
	sbiw R28,8
; }
; 
; /*!
;  * \brief Return the milliseconds counter value.
;  *
;  * This function returns the value of a counter, which is incremented
;  * every system timer tick. During system start, the counter is cleared
;  * to zero and will overflow with the 64 bit tick counter (4294967296).
;  * With the default 1024 ticks/s this will happen after 7.9 years.
;  * The resolution is also given by the system ticks.
;  *
;  * \note There is intentionally no provision to modify the seconds counter.
;  *       Callers can rely on a continuous update and use this value for
;  *       system tick independend timeout calculations.
;  *       Depending on 
;  *
;  * \return Value of the seconds counter.
;  */
; u_long NutGetMillis(void)
; {
;     // carefully stay within 64 bit values
;     u_long ticks   = NutGetTickCount();
	xcall _NutGetTickCount
	std y+0,R16
	std y+1,R17
	std y+2,R18
	std y+3,R19
;     u_long seconds = ticks / NutGetTickClock();
	xcall _NutGetTickClock
	movw R2,R16
	movw R4,R18
	ldd R16,y+0
	ldd R17,y+1
	ldd R18,y+2
	ldd R19,y+3
	st -y,R5
	st -y,R4
	st -y,R3
	st -y,R2
	xcall div32u
	std y+4,R16
	std y+5,R17
	std y+6,R18
	std y+7,R19
;     ticks         -= seconds * NutGetTickClock();
	xcall _NutGetTickClock
	movw R2,R16
	movw R4,R18
	ldd R16,y+4
	ldd R17,y+5
	ldd R18,y+6
	ldd R19,y+7
	st -y,R5
	st -y,R4
	st -y,R3
	st -y,R2
	xcall empy32u
	ldd R2,y+0
	ldd R3,y+1
	ldd R4,y+2
	ldd R5,y+3
	sub R2,R16
	sbc R3,R17
	sbc R4,R18
	sbc R5,R19
	std y+0,R2
	std y+1,R3
	std y+2,R4
	std y+3,R5
;     return seconds * 1000 + (ticks * 1000 ) / NutGetTickClock();
	xcall _NutGetTickClock
	movw R2,R16
	movw R4,R18
	ldd R6,y+0
	ldd R7,y+1
	ldd R8,y+2
	ldd R9,y+3
	ldi R20,232
	ldi R21,3
	ldi R22,0
	ldi R23,0
	st -y,R9
	st -y,R8
	st -y,R7
	st -y,R6
	movw R16,R20
	movw R18,R22
	xcall empy32u
	st -y,R5
	st -y,R4
	st -y,R3
	st -y,R2
	xcall div32u
	movw R2,R16
	movw R4,R18
	ldd R6,y+4
	ldd R7,y+5
	ldd R8,y+6
	ldd R9,y+7
	ldi R20,232
	ldi R21,3
	ldi R22,0
	ldi R23,0
	st -y,R9
	st -y,R8
	st -y,R7
	st -y,R6
	movw R16,R20
	movw R18,R22
	xcall empy32u
	add R16,R2
	adc R17,R3
	adc R18,R4
	adc R19,R5
L52:
	.dbline 0 ; func end
	adiw R28,8
	xjmp pop_xgsetF000
	.area bss(ram, con, rel)
_nut_ticks_resume:
	.blkb 4
_nut_ticks::
	.blkb 4
_nutTimerStopped::
	.blkb 1
_nutTimerList::
	.blkb 2
; }
; 
; /*@}*/
	.area func_lit
PL_NutThreadWake:	.word `_NutThreadWake
PL_NutTimerIntr:	.word `_NutTimerIntr
